﻿{ ------------------------------------------------------------------------------ }
{ }
{ N Package }
{ by Kambiz R. Khojasteh }
{ }
{ kambiz@delphiarea.com }
{ http://www.delphiarea.com }
{ }
{ ------------------------------------------------------------------------------ }

/// This unit defines some functions, data types, and constants regarding the
/// Unicode characters.
Unit NUnicode;

{$I DELPHIAREA.INC}

Interface

Const

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Left-to-right mark (LMR)</summary>
  /// <remarks>
  /// Acts exactly like left-to-right characters, except that it is not displayed.
  /// LRM does not have any other semantic effect.</remarks>
{$ENDREGION}
  UCC_LMR = #$200E;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Right-to-left mark (RLM)</summary>
  /// <remarks>
  /// Acts exactly like right-to-left characters, except that it is not displayed.
  /// RLM does not have any other semantic effect.</remarks>
{$ENDREGION}
  UCC_RLM = #$200F;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Zero width joiner (ZWJ)</summary>
  /// <remarks>
  /// Prevents the joining of consecutive characters on output.</remarks>
{$ENDREGION}
  UCC_ZWJ = #$200D;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Zero width non-joiner (ZWNJ)</summary>
  /// <remarks>
  /// Adds a non-joiner between two characters to prevent them from forming a
  /// cursive connection with each other when rendered.</remarks>
{$ENDREGION}
  UCC_ZWNJ = #$200C;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Start of left-to-right embedding (LRE)</summary>
  /// <remarks>
  /// Signals that a piece of text is to be treated as embedded left-to-right.
  /// For example, an English quotation in the middle of an Arabic sentence could
  /// be marked as being embedded left-to-right text.
  /// NOTE: LRE affects word order, not character order.</remarks>
{$ENDREGION}
  UCC_LRE = #$202A;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Start of right-to-left embedding (RLE)</summary>
  /// <remarks>
  /// Signals that a piece of text is to be treated as embedded right-to-left.
  /// For example, a Hebrew phrase in the middle of an English quotation could be
  /// marked as being embedded right-to-left.
  /// NOTE: RLE affects word order, not character order.</remarks>
{$ENDREGION}
  UCC_RLE = #$202B;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Start of left-to-right override (LRO)</summary>
  /// <remarks>
  /// Overrides bidirectional character types when required for special cases, such
  /// as for part numbers. LRO forces characters to be treated as strong left-to-right
  /// characters.</remarks>
{$ENDREGION}
  UCC_LRO = #$202D;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Start of right-to-left override (RLO)</summary>
  /// <remarks>
  /// Overrides bidirectional character types when required for special cases, such
  /// as for part numbers. RLO forces characters to be treated as strong right-to-left
  /// characters.</remarks>
{$ENDREGION}
  UCC_RLO = #$202E;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Pop directional formatting (PDF)</summary>
  /// <remarks>
  /// Terminates the effects of the last explicit code (either embedding or override)
  /// and restores the bidirectional state to what it was before the last LRE, RLE,
  /// RLO, or LRO control characters.</remarks>
{$ENDREGION}
  UCC_PDF = #$202C;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: National digit shapes substitution (NADS)</summary>
  /// <remarks>
  /// Uses national digit shapes to display the U+0030-U+0039 (ASCII numerals). The
  /// national digit shapes are determined by the current user locale.</remarks>
{$ENDREGION}
  UCC_NADS = #$206E;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Nominal (European) digit shapes (NODS)</summary>
  /// <remarks>
  /// Uses the nominal digit shapes for the digits U+0030-U+0039 (ASCII digits).
  /// Nominal digit shapes are Western digits.</remarks>
{$ENDREGION}
  UCC_NODS = #$206F;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Activate symmetric swapping (ASS)</summary>
  /// <remarks>
  /// Indicates whether the term LEFT or RIGHT in pairs of character names, such
  /// as parentheses, should be interpreted as meaning opening or closing, respectively.
  /// The default state is activated.</remarks>
{$ENDREGION}
  UCC_ASS = #$206B;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Inhibit symmetric swapping (ISS)</summary>
  /// <remarks>
  /// Turns off the symmetric swapping of characters such as parenthesis, so that left
  /// and right continue to mean left and right facing, as opposed to opening and closing,
  /// when symmetric swapping is on.</remarks>
{$ENDREGION}
  UCC_ISS = #$206A;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Activate Arabic form shaping (AAFS)</summary>
  /// <remarks>
  /// Controls the shaping behavior of the Arabic compatibility characters. During the
  /// presentation process, certain letterforms may be joined together in cursive
  /// connection or ligatures. The shaping selector codes indicate that the character
  /// shape determination (glyph selection) process used to achieve this presentation
  /// effect is to be either activated or inhibited. The default state is inhibited.</remarks>
{$ENDREGION}
  UCC_AAFS = #$206D;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Inhibit Arabic form shaping (IAFS)</summary>
  /// <remarks>
  /// Inhibits the character shaping determination process, so that characters are not
  /// shaped based on position.</remarks>
{$ENDREGION}
  UCC_IAFS = #$206C;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Record Separator (Block separator) (RS)</summary>
  /// <remarks>
  /// Begins a new line after each line separator.</remarks>
{$ENDREGION}
  UCC_RS = #$241E;

{$REGION 'xmldoc'}
  /// <summary>
  /// Unicode Control Character: Unit Separator (Segment separator) (US)</summary>
  /// <remarks>
  /// Unicode Control Character: Begins a new paragraph after each paragraph separator.</remarks>
{$ENDREGION}
  UCC_US = #$241F;

Type

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the different type of scripts.</summary>
{$ENDREGION}
  TUnicodeScript = (
{$REGION 'xmldoc'}
    /// Arabic
{$ENDREGION}
    UsArabic,
{$REGION 'xmldoc'}
    /// Imperial Aramaic
{$ENDREGION}
    UsImperialAramaic,
{$REGION 'xmldoc'}
    /// Armenian
{$ENDREGION}
    UsArmenian,
{$REGION 'xmldoc'}
    /// Avestan
{$ENDREGION}
    UsAvestan,
{$REGION 'xmldoc'}
    /// Balinese
{$ENDREGION}
    UsBalinese,
{$REGION 'xmldoc'}
    /// Bamum
{$ENDREGION}
    UsBamum,
{$REGION 'xmldoc'}
    /// Bassa Vah
{$ENDREGION}
    UsBassaVah,
{$REGION 'xmldoc'}
    /// Batak
{$ENDREGION}
    UsBatak,
{$REGION 'xmldoc'}
    /// Bengali
{$ENDREGION}
    UsBengali,
{$REGION 'xmldoc'}
    /// Blissymbols
{$ENDREGION}
    UsBlissymbols,
{$REGION 'xmldoc'}
    /// Bopomofo
{$ENDREGION}
    UsBopomofo,
{$REGION 'xmldoc'}
    /// Brahmi
{$ENDREGION}
    UsBrahmi,
{$REGION 'xmldoc'}
    /// Braille
{$ENDREGION}
    UsBraille,
{$REGION 'xmldoc'}
    /// Buginese
{$ENDREGION}
    UsBuginese,
{$REGION 'xmldoc'}
    /// Buhid
{$ENDREGION}
    UsBuhid,
{$REGION 'xmldoc'}
    /// Chakma
{$ENDREGION}
    UsChakma,
{$REGION 'xmldoc'}
    /// Unified Canadian Aboriginal Syllabics
{$ENDREGION}
    UsUnifiedCanadianAboriginalSyllabics,
{$REGION 'xmldoc'}
    /// Carian
{$ENDREGION}
    UsCarian,
{$REGION 'xmldoc'}
    /// Cham
{$ENDREGION}
    UsCham,
{$REGION 'xmldoc'}
    /// Cherokee
{$ENDREGION}
    UsCherokee,
{$REGION 'xmldoc'}
    /// Cirth
{$ENDREGION}
    UsCirth,
{$REGION 'xmldoc'}
    /// Coptic
{$ENDREGION}
    UsCoptic,
{$REGION 'xmldoc'}
    /// Cypriot
{$ENDREGION}
    UsCypriot,
{$REGION 'xmldoc'}
    /// Cyrillic
{$ENDREGION}
    UsCyrillic,
{$REGION 'xmldoc'}
    /// Cyrillic (Old Church Slavonic variant)
{$ENDREGION}
    UsCyrillic_OldChurchSlavonic,
{$REGION 'xmldoc'}
    /// Devanagari (Nagari)
{$ENDREGION}
    UsDevanagari,
{$REGION 'xmldoc'}
    /// Deseret (Mormon)
{$ENDREGION}
    UsDeseret,
{$REGION 'xmldoc'}
    /// Duployan shorthand, Duployan stenography
{$ENDREGION}
    UsDuployan,
{$REGION 'xmldoc'}
    /// Egyptian demotic
{$ENDREGION}
    UsEgyptianDemotic,
{$REGION 'xmldoc'}
    /// Egyptian hieratic
{$ENDREGION}
    UsEgyptianHieratic,
{$REGION 'xmldoc'}
    /// Egyptian hieroglyphs
{$ENDREGION}
    UsEgyptianHieroglyphs,
{$REGION 'xmldoc'}
    /// Elbasan
{$ENDREGION}
    UsElbasan,
{$REGION 'xmldoc'}
    /// Ethiopic
{$ENDREGION}
    UsEthiopic,
{$REGION 'xmldoc'}
    /// Khutsuri
{$ENDREGION}
    UsKhutsuri,
{$REGION 'xmldoc'}
    /// Georgian (Mkhedruli)
{$ENDREGION}
    UsGeorgian,
{$REGION 'xmldoc'}
    /// Glagolitic
{$ENDREGION}
    UsGlagolitic,
{$REGION 'xmldoc'}
    /// Gothic
{$ENDREGION}
    UsGothic,
{$REGION 'xmldoc'}
    /// Grantha
{$ENDREGION}
    UsGrantha,
{$REGION 'xmldoc'}
    /// Greek
{$ENDREGION}
    UsGreek,
{$REGION 'xmldoc'}
    /// Gujarati
{$ENDREGION}
    UsGujarati,
{$REGION 'xmldoc'}
    /// Gurmukhi
{$ENDREGION}
    UsGurmukhi,
{$REGION 'xmldoc'}
    /// Hangul (Hangŭl, Hangeul)
{$ENDREGION}
    UsHangul,
{$REGION 'xmldoc'}
    /// Han (Hanzi, Kanji, Hanja)
{$ENDREGION}
    UsHan,
{$REGION 'xmldoc'}
    /// Hanunoo (Hanunóo)
{$ENDREGION}
    UsHanunoo,
{$REGION 'xmldoc'}
    /// Han (Simplified variant)
{$ENDREGION}
    UsHan_Simplified,
{$REGION 'xmldoc'}
    /// Han (Traditional variant)
{$ENDREGION}
    UsHan_Traditional,
{$REGION 'xmldoc'}
    /// Hebrew
{$ENDREGION}
    UsHebrew,
{$REGION 'xmldoc'}
    /// Hiragana
{$ENDREGION}
    UsHiragana,
{$REGION 'xmldoc'}
    /// Pahawh Hmong
{$ENDREGION}
    UsPahawhHmong,
{$REGION 'xmldoc'}
    /// (alias for Hiragana + Katakana)
{$ENDREGION}
    UsHiraganaKatakana,
{$REGION 'xmldoc'}
    /// Old Hungarian
{$ENDREGION}
    UsOldHungarian,
{$REGION 'xmldoc'}
    /// Indus (Harappan)
{$ENDREGION}
    UsIndus,
{$REGION 'xmldoc'}
    /// Old Italic (Etruscan, Oscan, etc.)
{$ENDREGION}
    UsOldItalic,
{$REGION 'xmldoc'}
    /// Javanese
{$ENDREGION}
    UsJavanese,
{$REGION 'xmldoc'}
    /// Japanese (alias for Han + Hiragana + Katakana)
{$ENDREGION}
    UsJapanese,
{$REGION 'xmldoc'}
    /// Kayah Li
{$ENDREGION}
    UsKayahLi,
{$REGION 'xmldoc'}
    /// Katakana
{$ENDREGION}
    UsKatakana,
{$REGION 'xmldoc'}
    /// Kharoshthi
{$ENDREGION}
    UsKharoshthi,
{$REGION 'xmldoc'}
    /// Khmer
{$ENDREGION}
    UsKhmer,
{$REGION 'xmldoc'}
    /// Kannada
{$ENDREGION}
    UsKannada,
{$REGION 'xmldoc'}
    /// Korean (alias for Hangul + Han)
{$ENDREGION}
    UsKorean,
{$REGION 'xmldoc'}
    /// Kpelle
{$ENDREGION}
    UsKpelle,
{$REGION 'xmldoc'}
    /// Kaithi
{$ENDREGION}
    UsKaithi,
{$REGION 'xmldoc'}
    /// Tai Tham (Lanna)
{$ENDREGION}
    UsTaiTham,
{$REGION 'xmldoc'}
    /// Lao
{$ENDREGION}
    UsLao,
{$REGION 'xmldoc'}
    /// Latin (Fraktur variant)
{$ENDREGION}
    UsLatin_Fraktur,
{$REGION 'xmldoc'}
    /// Latin (Gaelic variant)
{$ENDREGION}
    UsLatin_Gaelic,
{$REGION 'xmldoc'}
    /// Latin
{$ENDREGION}
    UsLatin,
{$REGION 'xmldoc'}
    /// Lepcha (Róng)
{$ENDREGION}
    UsLepcha,
{$REGION 'xmldoc'}
    /// Limbu
{$ENDREGION}
    UsLimbu,
{$REGION 'xmldoc'}
    /// Linear A
{$ENDREGION}
    UsLinearA,
{$REGION 'xmldoc'}
    /// Linear B
{$ENDREGION}
    UsLinearB,
{$REGION 'xmldoc'}
    /// Lisu (Fraser)
{$ENDREGION}
    UsLisu,
{$REGION 'xmldoc'}
    /// Loma
{$ENDREGION}
    UsLoma,
{$REGION 'xmldoc'}
    /// Lycian
{$ENDREGION}
    UsLycian,
{$REGION 'xmldoc'}
    /// Lydian
{$ENDREGION}
    UsLydian,
{$REGION 'xmldoc'}
    /// Mandaic
{$ENDREGION}
    UsMandaic,
{$REGION 'xmldoc'}
    /// Manichaean
{$ENDREGION}
    UsManichaean,
{$REGION 'xmldoc'}
    /// Mayan hieroglyphs
{$ENDREGION}
    UsMayanHieroglyphs,
{$REGION 'xmldoc'}
    /// Mende
{$ENDREGION}
    UsMende,
{$REGION 'xmldoc'}
    /// Meroitic Cursive
{$ENDREGION}
    UsMeroiticCursive,
{$REGION 'xmldoc'}
    /// Meroitic Hieroglyphs
{$ENDREGION}
    UsMeroiticHieroglyphs,
{$REGION 'xmldoc'}
    /// Malayalam
{$ENDREGION}
    UsMalayalam,
{$REGION 'xmldoc'}
    /// Mongolian
{$ENDREGION}
    UsMongolian,
{$REGION 'xmldoc'}
    /// Moon (Moon code, Moon script, Moon type)
{$ENDREGION}
    UsMoon,
{$REGION 'xmldoc'}
    /// Meitei Mayek (Meithei, Meetei)
{$ENDREGION}
    UsMeiteiMayek,
{$REGION 'xmldoc'}
    /// Myanmar (Burmese)
{$ENDREGION}
    UsMyanmar,
{$REGION 'xmldoc'}
    /// Old North Arabian (Ancient North Arabian)
{$ENDREGION}
    UsOldNorthArabian,
{$REGION 'xmldoc'}
    /// Nabataean
{$ENDREGION}
    UsNabataean,
{$REGION 'xmldoc'}
    /// Nakhi Geba ('Na-'Khi ²Ggŏ-¹baw, Naxi Geba)
{$ENDREGION}
    UsNakhiGeba,
{$REGION 'xmldoc'}
    /// N’Ko
{$ENDREGION}
    UsNKo,
{$REGION 'xmldoc'}
    /// Ogham
{$ENDREGION}
    UsOgham,
{$REGION 'xmldoc'}
    /// Ol Chiki (Ol Cemet’, Ol, Santali)
{$ENDREGION}
    UsOlChiki,
{$REGION 'xmldoc'}
    /// Old Turkic, Orkhon Runic
{$ENDREGION}
    UsOldTurkicOrkhonRunic,
{$REGION 'xmldoc'}
    /// Oriya
{$ENDREGION}
    UsOriya,
{$REGION 'xmldoc'}
    /// Osmanya
{$ENDREGION}
    UsOsmanya,
{$REGION 'xmldoc'}
    /// Palmyrene
{$ENDREGION}
    UsPalmyrene,
{$REGION 'xmldoc'}
    /// Old Permic
{$ENDREGION}
    UsOldPermic,
{$REGION 'xmldoc'}
    /// Phags-pa
{$ENDREGION}
    UsPhagsPa,
{$REGION 'xmldoc'}
    /// Inscriptional Pahlavi
{$ENDREGION}
    UsInscriptionalPahlavi,
{$REGION 'xmldoc'}
    /// Psalter Pahlavi
{$ENDREGION}
    UsPsalterPahlavi,
{$REGION 'xmldoc'}
    /// Book Pahlavi
{$ENDREGION}
    UsBookPahlavi,
{$REGION 'xmldoc'}
    /// Phoenician
{$ENDREGION}
    UsPhoenician,
{$REGION 'xmldoc'}
    /// Miao (Pollard)
{$ENDREGION}
    UsMiao,
{$REGION 'xmldoc'}
    /// Inscriptional Parthian
{$ENDREGION}
    UsInscriptionalParthian,
{$REGION 'xmldoc'}
    /// Reserved for private use (start)
{$ENDREGION}
    UsReservedStart,
{$REGION 'xmldoc'}
    /// Reserved for private use (end)
{$ENDREGION}
    UsReservedEnd,
{$REGION 'xmldoc'}
    /// Rejang (Redjang, Kaganga)
{$ENDREGION}
    UsRejang,
{$REGION 'xmldoc'}
    /// Rongorongo
{$ENDREGION}
    UsRongorongo,
{$REGION 'xmldoc'}
    /// Runic
{$ENDREGION}
    UsRunic,
{$REGION 'xmldoc'}
    /// Samaritan
{$ENDREGION}
    UsSamaritan,
{$REGION 'xmldoc'}
    /// Sarati
{$ENDREGION}
    UsSarati,
{$REGION 'xmldoc'}
    /// Old South Arabian
{$ENDREGION}
    UsOldSouthArabian,
{$REGION 'xmldoc'}
    /// Saurashtra
{$ENDREGION}
    UsSaurashtra,
{$REGION 'xmldoc'}
    /// SignWriting
{$ENDREGION}
    UsSignWriting,
{$REGION 'xmldoc'}
    /// Shavian (Shaw)
{$ENDREGION}
    UsShavian,
{$REGION 'xmldoc'}
    /// Sindhi
{$ENDREGION}
    UsSindhi,
{$REGION 'xmldoc'}
    /// Sinhala
{$ENDREGION}
    UsSinhala,
{$REGION 'xmldoc'}
    /// Sundanese
{$ENDREGION}
    UsSundanese,
{$REGION 'xmldoc'}
    /// Syloti Nagri
{$ENDREGION}
    UsSylotiNagri,
{$REGION 'xmldoc'}
    /// Syriac
{$ENDREGION}
    UsSyriac,
{$REGION 'xmldoc'}
    /// Syriac (Estrangelo variant)
{$ENDREGION}
    UsSyriac_Estrangelo,
{$REGION 'xmldoc'}
    /// Syriac (Western variant)
{$ENDREGION}
    UsSyriac_Western,
{$REGION 'xmldoc'}
    /// Syriac (Eastern variant)
{$ENDREGION}
    UsSyriac_Eastern,
{$REGION 'xmldoc'}
    /// Tagbanwa
{$ENDREGION}
    UsTagbanwa,
{$REGION 'xmldoc'}
    /// Tai
{$ENDREGION}
    UsTai,
{$REGION 'xmldoc'}
    /// New Tai Lue
{$ENDREGION}
    UsNewTaiLue,
{$REGION 'xmldoc'}
    /// Tamil
{$ENDREGION}
    UsTamil,
{$REGION 'xmldoc'}
    /// Tai Viet
{$ENDREGION}
    UsTaiViet,
{$REGION 'xmldoc'}
    /// Telugu
{$ENDREGION}
    UsTelugu,
{$REGION 'xmldoc'}
    /// Tengwar
{$ENDREGION}
    UsTengwar,
{$REGION 'xmldoc'}
    /// Tifinagh (Berber)
{$ENDREGION}
    UsTifinagh,
{$REGION 'xmldoc'}
    /// Tagalog (Baybayin, Alibata)
{$ENDREGION}
    UsTagalog,
{$REGION 'xmldoc'}
    /// Thaana
{$ENDREGION}
    UsThaana,
{$REGION 'xmldoc'}
    /// Thai
{$ENDREGION}
    UsThai,
{$REGION 'xmldoc'}
    /// Tibetan
{$ENDREGION}
    UsTibetan,
{$REGION 'xmldoc'}
    /// Ugaritic
{$ENDREGION}
    UsUgaritic,
{$REGION 'xmldoc'}
    /// Vai
{$ENDREGION}
    UsVai,
{$REGION 'xmldoc'}
    /// Visible Speech
{$ENDREGION}
    UsVisibleSpeech,
{$REGION 'xmldoc'}
    /// Warang Citi (Varang Kshiti)
{$ENDREGION}
    UsWarangCiti,
{$REGION 'xmldoc'}
    /// Old Persian
{$ENDREGION}
    UsOldPersian,
{$REGION 'xmldoc'}
    /// Cuneiform, Sumero-Akkadian
{$ENDREGION}
    UsCuneiformSumeroAkkadian,
{$REGION 'xmldoc'}
    /// Yi
{$ENDREGION}
    UsYi,
{$REGION 'xmldoc'}
    /// Code for inherited script
{$ENDREGION}
    UsInherited,
{$REGION 'xmldoc'}
    /// Mathematical notation
{$ENDREGION}
    UsMathematical,
{$REGION 'xmldoc'}
    /// Symbols
{$ENDREGION}
    UsSymbols,
{$REGION 'xmldoc'}
    /// Unwritten documents
{$ENDREGION}
    UsUnwritten,
{$REGION 'xmldoc'}
    /// Undetermined script
{$ENDREGION}
    UsUndetermined,
{$REGION 'xmldoc'}
    /// Uncoded script
{$ENDREGION}
    UsUncoded);

{$REGION 'xmldoc'}
  /// <summary>
  /// This data type represents a set of <see cref="TUnicodeScript"/> values.</summary>
{$ENDREGION}
  TUnicodeScripts = Set Of TUnicodeScript;

{$REGION 'xmldoc'}
  /// <summary>
  /// Converts a semicolon separated list of the four characters ISO 15924 script
  /// codes to a set of <see cref="TUnicodeScript"/> values.</summary>
  /// <param name="Codes">
  /// The semicolon separated list of the four-character codes.</param>
  /// <returns>
  /// The set of <see cref="TUnicodeScript"/> values.</returns>
  /// <seealso cref="ScriptsToScriptCodes"/>
{$ENDREGION}
Function ScriptCodesToScripts(Const Codes: String): TUnicodeScripts;

{$REGION 'xmldoc'}
/// <summary>
/// Converts a set of <see cref="TUnicodeScript"/> values to a semicolon separated
/// list of the four characters ISO 15924 script codes.</summary>
/// <param name="Scripts">
/// The set of <see cref="TUnicodeScript"/> values.</param>
/// <returns>
/// The semicolon separated list of the four-character codes.</returns>
/// <seealso cref="ScriptCodesToScripts"/>
{$ENDREGION}
Function ScriptsToScriptCodes(Scripts: TUnicodeScripts): String;

Implementation

Uses
  SysUtils, Types;

Const
  // ISO 15924 script codes in alphabetical order
  ScriptCodes: Array [TUnicodeScript] Of PChar = ('Arab', 'Armi', 'Armn',
    'Avst', 'Bali', 'Bamu', 'Bass', 'Batk', 'Beng', 'Blis', 'Bopo', 'Brah',
    'Brai', 'Bugi', 'Buhd', 'Cakm', 'Cans', 'Cari', 'Cham', 'Cher', 'Cirt',
    'Copt', 'Cprt', 'Cyrl', 'Cyrs', 'Deva', 'Dsrt', 'Dupl', 'Egyd', 'Egyh',
    'Egyp', 'Elba', 'Ethi', 'Geok', 'Geor', 'Glag', 'Goth', 'Gran', 'Grek',
    'Gujr', 'Guru', 'Hang', 'Hani', 'Hano', 'Hans', 'Hant', 'Hebr', 'Hira',
    'Hmng', 'Hrkt', 'Hung', 'Inds', 'Ital', 'Java', 'Jpan', 'Kali', 'Kana',
    'Khar', 'Khmr', 'Knda', 'Kore', 'Kpel', 'Kthi', 'Lana', 'Laoo', 'Latf',
    'Latg', 'Latn', 'Lepc', 'Limb', 'Lina', 'Linb', 'Lisu', 'Loma', 'Lyci',
    'Lydi', 'Mand', 'Mani', 'Maya', 'Mend', 'Merc', 'Mero', 'Mlym', 'Mong',
    'Moon', 'Mtei', 'Mymr', 'Narb', 'Nbat', 'Nkgb', 'Nkoo', 'Ogam', 'Olck',
    'Orkh', 'Orya', 'Osma', 'Palm', 'Perm', 'Phag', 'Phli', 'Phlp', 'Phlv',
    'Phnx', 'Plrd', 'Prti', 'Qaaa', 'Qabx', 'Rjng', 'Roro', 'Runr', 'Samr',
    'Sara', 'Sarb', 'Saur', 'Sgnw', 'Shaw', 'Sind', 'Sinh', 'Sund', 'Sylo',
    'Syrc', 'Syre', 'Syrj', 'Syrn', 'Tagb', 'Tale', 'Talu', 'Taml', 'Tavt',
    'Telu', 'Teng', 'Tfng', 'Tglg', 'Thaa', 'Thai', 'Tibt', 'Ugar', 'Vaii',
    'Visp', 'Wara', 'Xpeo', 'Xsux', 'Yiii', 'Zinh', 'Zmth', 'Zsym', 'Zxxx',
    'Zyyy', 'Zzzz');

Function FindScript(PCode: PChar; Out Script: TUnicodeScript): Boolean;
Var
  L, H: TUnicodeScript;
  C: Integer;
Begin
  L := Low(TUnicodeScript);
  H := High(TUnicodeScript);
  Repeat
    Script := TUnicodeScript((Ord(L) + Ord(H)) Shr 1);
    C := StrLIComp(ScriptCodes[Script], PCode, 4);
    If C < 0 Then
      L := Succ(Script)
    Else If C > 0 Then
      H := Pred(Script)
    Else
    Begin
      Result := True;
      Exit;
    End;
  Until L > H;
  Result := False;
End;

Function ScriptCodesToScripts(Const Codes: String): TUnicodeScripts;
Var
  Script: TUnicodeScript;
  S, E: PChar;
Begin
  Result := [];
  S := PChar(Codes);
  E := StrScan(S, ';');
  While E <> Nil Do
  Begin
    If (E - S = 4) And FindScript(S, Script) Then
      Include(Result, Script);
    S := E + 1;
    If S^ = #0 Then
      Exit;
    E := StrScan(S, ';');
  End;
End;

Function ScriptsToScriptCodes(Scripts: TUnicodeScripts): String;
Var
  Script: TUnicodeScript;
Begin
  Result := '';
  For Script := Low(TUnicodeScript) To High(TUnicodeScript) Do
    If Script In Scripts Then
      Result := Result + ScriptCodes[Script] + ';';
End;

End.
