{ ------------------------------------------------------------------------------ }
{ }
{ i18n Package }
{ by Kambiz R. Khojasteh }
{ }
{ kambiz@delphiarea.com }
{ http://www.delphiarea.com }
{ }
{ ------------------------------------------------------------------------------ }

/// This unit implements a class to evaluate expressions in C language syntax.
Unit i18nCExpr;

{$I DELPHIAREA.INC}

Interface

Uses
  SysUtils, Classes, Types;

Type

{$REGION 'xmldoc'}
  /// <summary>
  /// ECExpressionError is the exception class for problems that occur while parsing
  /// a expression in C language syntax.</summary>
  /// <remarks>
  /// ECExpressionError exception occurs when an expression in C language syntax has
  /// a syntax error.</remarks>
  /// <seealso cref="TCExpression"/>
{$ENDREGION}
  ECExpressionError = Class(Exception)
  Private
    FExpression: String;
    FOffset: Integer;
    FToken: String;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Sets the additional information about the error.</summary>
    /// <param name="AExpression">
    /// The expression that caused the exception to occur.</param>
    /// <param name="AOffset">
    /// The offset from the start of the expression where the parser encountered
    /// the problem.</param>
    /// <param name="AToken">
    /// The last token that the parser was processing when the problem occured.</param>
{$ENDREGION}
    Procedure SetErrorDetails(Const AExpression: String; AOffset: Integer; Const AToken: String);
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the expression that caused the execption to occur.</summary>
{$ENDREGION}
    Property Expression: String Read FExpression;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the offset from the start of the expression where the parser encountered
    /// the problem.</summary>
{$ENDREGION}
    Property Offset: Integer Read FOffset;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the last token that the parser was processing when the problem occured.</summary>
{$ENDREGION}
    Property Token: String Read FToken;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprVar represents an integer variable in an expression.</summary>
  /// <seealso cref="TCExpression"/>
{$ENDREGION}

  TCExprVar = Class(TObject)
  Private
    FName: String;
    FValue: Integer;
    FRefCount: Integer;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Increments the reference count of the variable.</summary>
    /// <returns>
    /// The object itself.</returns>
    /// <seealso cref="RefCount"/>
{$ENDREGION}
    Function AddRef: TCExprVar; Inline;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for a specified variable name.</summary>
    /// <param name="AName">
    /// The name of the variable.</param>
{$ENDREGION}
    Constructor Create(Const AName: String);
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the name of the variable.</summary>
{$ENDREGION}
    Property Name: String Read FName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the number of times that the variable is referenced in the expression.</summary>
{$ENDREGION}
    Property RefCount: Integer Read FRefCount;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets value of the variable.</summary>
{$ENDREGION}
    Property Value: Integer Read FValue Write FValue;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprNode is the base class for sub-expressions of an expression.</summary>
{$ENDREGION}

  TCExprNode = Class Abstract(TObject)
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the value of the sub-expression.</summary>
    /// <returns>
    /// The value of the sub-expression after evaluation.</returns>
    /// <seealso cref="Value"/>
{$ENDREGION}
    Function GetValue: Integer; Virtual; Abstract;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-expression as a string.</summary>
    /// <returns>
    /// The sub-expression as string.</returns>
    /// <seealso cref="AsString"/>
{$ENDREGION}
    Function GetAsString: String; Virtual; Abstract;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of sub-expression depends on the value of one
    /// or more variables.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the value of sub-expression depends on
    /// variable. Otherwise, returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Virtual; Abstract;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the sub-expression as a string.</summary>
{$ENDREGION}
    Property AsString: String Read GetAsString;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets The value of the sub-expression after evaluation.</summary>
{$ENDREGION}
    Property Value: Integer Read GetValue;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprVarNode represents a reference to an integer variable in an expression.</summary>
{$ENDREGION}

  TCExprVarNode = Class(TCExprNode)
  Private
    FVariable: TCExprVar;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the value of the sub-expression.</summary>
    /// <returns>
    /// The value of the variable.</returns>
    /// <seealso cref="Value"/>
{$ENDREGION}
    Function GetValue: Integer; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-expression as a string.</summary>
    /// <returns>
    /// The name of the variable.</returns>
    /// <seealso cref="AsString"/>
{$ENDREGION}
    Function GetAsString: String; Override;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for a specified variable.</summary>
    /// <param name="AVariable">
    /// The variable that this instance refrences it.</param>
{$ENDREGION}
    Constructor Create(Const AVariable: TCExprVar);
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of sub-expression depends on the value of one
    /// or more variables.</summary>
    /// <returns>
    /// Returns always <see langword="true"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the variable which this sub-expression is a reference to it.</summary>
{$ENDREGION}
    Property Variable: TCExprVar Read FVariable;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprVarNode represents an integer constant in an expression.</summary>
{$ENDREGION}

  TCExprConstNode = Class(TCExprNode)
  Private
    FValue: Integer;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the value of the sub-expression.</summary>
    /// <returns>
    /// The value of the constant.</returns>
    /// <seealso cref="Value"/>
{$ENDREGION}
    Function GetValue: Integer; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-expression as a string.</summary>
    /// <returns>
    /// The value of constant as a string.</returns>
    /// <seealso cref="AsString"/>
{$ENDREGION}
    Function GetAsString: String; Override;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for a specified integer value.</summary>
    /// <param name="AValue">
    /// The value of constant that this instance represents it.</param>
{$ENDREGION}
    Constructor Create(AValue: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of sub-expression depends on the value of one
    /// or more variables.</summary>
    /// <returns>
    /// Returns always <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Override;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the unary operators in a C language expression.</summary>
{$ENDREGION}

  TCUnaryOp = (
{$REGION 'xmldoc'}
    /// unary plus (+)
{$ENDREGION}
    OpNone,
{$REGION 'xmldoc'}
    /// unary minus (-)
{$ENDREGION}
    OpNegate,
{$REGION 'xmldoc'}
    /// logical not (!)
{$ENDREGION}
    OpLogNot,
{$REGION 'xmldoc'}
    /// bitwise not (~)
{$ENDREGION}
    OpBitNot);

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprUnaryNode represents a unary operation in an expression.</summary>
{$ENDREGION}

  TCExprUnaryNode = Class(TCExprNode)
  Private
    FOp: TCUnaryOp;
    FOperand: TCExprNode;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the value of the sub-expression.</summary>
    /// <returns>
    /// The value of the sub-expression after evaluation.</returns>
    /// <seealso cref="Value"/>
{$ENDREGION}
    Function GetValue: Integer; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-expression as a string.</summary>
    /// <returns>
    /// The sub-expression as string.</returns>
    /// <seealso cref="AsString"/>
{$ENDREGION}
    Function GetAsString: String; Override;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class.</summary>
    /// <param name="AOp">
    /// The unary operator.</param>
    /// <param name="AOperand">
    /// The operand of the unary operation.</param>
{$ENDREGION}
    Constructor Create(AOp: TCUnaryOp; AOperand: TCExprNode);
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of sub-expression depends on the value of one
    /// or more variables.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the value of sub-expression depends on
    /// variable. Otherwise, returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the operator of the unary operation.</summary>
{$ENDREGION}
    Property Op: TCUnaryOp Read FOp;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the operand of the unary operation.</summary>
{$ENDREGION}
    Property Operand: TCExprNode Read FOperand;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the binary operators in a C language expression.</summary>
{$ENDREGION}

  TCBinaryOp = (
{$REGION 'xmldoc'}
    /// addition (+)
{$ENDREGION}
    OpAdd,
{$REGION 'xmldoc'}
    /// subtraction (-)
{$ENDREGION}
    OpSubtract,
{$REGION 'xmldoc'}
    /// multiplication (*)
{$ENDREGION}
    OpMultiply,
{$REGION 'xmldoc'}
    /// integer division (/)
{$ENDREGION}
    OpDivide,
{$REGION 'xmldoc'}
    /// reminder of integer devision (%)
{$ENDREGION}
    OpModulus,
{$REGION 'xmldoc'}
    /// bitwise and (&amp;)
{$ENDREGION}
    OpBitAnd,
{$REGION 'xmldoc'}
    /// bitwise or (|)
{$ENDREGION}
    OpBitOr,
{$REGION 'xmldoc'}
    /// bitwise xor (^)
{$ENDREGION}
    OpBitXor,
{$REGION 'xmldoc'}
    /// shift left (&lt;&lt;)
{$ENDREGION}
    OpShiftLeft,
{$REGION 'xmldoc'}
    /// shift right (&gt;&gt;)
{$ENDREGION}
    OpShiftRight,
{$REGION 'xmldoc'}
    /// equality (==)
{$ENDREGION}
    OpEqual,
{$REGION 'xmldoc'}
    /// not eqaul (!=)
{$ENDREGION}
    OpNotEqual,
{$REGION 'xmldoc'}
    /// greater than (&gt;)
{$ENDREGION}
    OpGreaterThan,
{$REGION 'xmldoc'}
    /// greater than or equal (&gt;=)
{$ENDREGION}
    OpGreaterThanOrEqual,
{$REGION 'xmldoc'}
    /// less than (&lt;)
{$ENDREGION}
    OpLessThan,
{$REGION 'xmldoc'}
    /// less than or equal (&lt;=)
{$ENDREGION}
    OpLessThanOrEqual,
{$REGION 'xmldoc'}
    /// logical and (&amp;&amp;)
{$ENDREGION}
    OpLogAnd,
{$REGION 'xmldoc'}
    /// logical or (||)
{$ENDREGION}
    OpLogOr);

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprBinaryNode represents a binary operation in an expression.</summary>
{$ENDREGION}

  TCExprBinaryNode = Class(TCExprNode)
  Private
    FOp: TCBinaryOp;
    FOperandLeft: TCExprNode;
    FOperandRight: TCExprNode;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the value of the sub-expression.</summary>
    /// <returns>
    /// The value of the sub-expression after evaluation.</returns>
    /// <seealso cref="Value"/>
{$ENDREGION}
    Function GetValue: Integer; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-expression as a string.</summary>
    /// <returns>
    /// The sub-expression as string.</returns>
    /// <seealso cref="AsString"/>
{$ENDREGION}
    Function GetAsString: String; Override;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class.</summary>
    /// <param name="AOp">
    /// The binary operator.</param>
    /// <param name="AOperandLeft">
    /// The left operand of the binary operation.</param>
    /// <param name="AOperandRight">
    /// The right operand of the binary operation.</param>
{$ENDREGION}
    Constructor Create(AOp: TCBinaryOp; AOperandLeft, AOperandRight: TCExprNode);
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of sub-expression depends on the value of one
    /// or more variables.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the value of sub-expression depends on
    /// variable. Otherwise, returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the operator of the binary operation.</summary>
{$ENDREGION}
    Property Op: TCBinaryOp Read FOp;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the left operand of the binary operation.</summary>
{$ENDREGION}
    Property OperandLeft: TCExprNode Read FOperandLeft;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the right operand of the binary operation.</summary>
{$ENDREGION}
    Property OperandRight: TCExprNode Read FOperandRight;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExprBinaryNode represents a ternary conditional operation in an expression.</summary>
{$ENDREGION}

  TCExprTernaryNode = Class(TCExprNode)
  Private
    FCondition: TCExprNode;
    FOperandTrue: TCExprNode;
    FOperandFalse: TCExprNode;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the value of the sub-expression.</summary>
    /// <returns>
    /// The value of the sub-expression after evaluation.</returns>
    /// <seealso cref="Value"/>
{$ENDREGION}
    Function GetValue: Integer; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-expression as a string.</summary>
    /// <returns>
    /// The sub-expression as string.</returns>
    /// <seealso cref="AsString"/>
{$ENDREGION}
    Function GetAsString: String; Override;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
    /// <param name="ACondition">
    /// The condition operand.</param>
    /// <param name="AOperandTrue">
    /// The operand that evaluates when the condition evaluates to true.</param>
    /// <param name="AOperandFalse">
    /// The operand that evaluates when the condition evaluates to false.</param>
{$ENDREGION}
    Constructor Create(ACondition, AOperandTrue, AOperandFalse: TCExprNode);
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of sub-expression depends on the value of one
    /// or more variables.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the value of sub-expression depends on
    /// variable. Otherwise, returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the condition operand of the ternary operation.</summary>
{$ENDREGION}
    Property Condition: TCExprNode Read FCondition;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the true phrase operand of the ternary operation.</summary>
{$ENDREGION}
    Property OperandTrue: TCExprNode Read FOperandTrue;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the false phrase operand of the ternary operation.</summary>
{$ENDREGION}
    Property OperandFalse: TCExprNode Read FOperandFalse;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCExpression is parser and evaluator for expressions in C language syntax.</summary>
  /// <remarks>
  /// TCExpression parses and evaluates an expression (specified as a string) which is using
  /// the C language syntax. Exceptions are that numbers must be decimal and no assignment,
  /// incerement/decrement, or comma operators are allowed. The expression can use user-defined
  /// variables.
  ///
  /// TCExpression supports comparison and logical operations as well as arithmetic
  /// and bitwaise ones. The comparison and logical operations always evaluate to
  /// either zero (false) or one (true).
  ///
  /// TCExpression parses the expression once to build the parse tree. Then, each time
  /// you call the <see cref="Evaluate"/> method, it uses the parse tree and the value
  /// of the variables listed by the <see cref="Vars"/> property to calculate the result
  /// of the expression. For this reason, TCExpression is very usful and suitable when
  /// you have one expression that will be evaluated many times with different variable
  /// values.
  ///
  /// The following table lists C operators that are supported by TCExpression. The
  /// operators are listed in order of precedence (highest to lowest). Their associativity
  /// indicates in what order operators of equal precedence in an expression are applied.
  ///
  /// <list type="table">
  /// <listheader>
  /// <term>Operator</term>
  /// <description>Description</description>
  /// <description>Associativity</description>
  /// </listheader>
  /// <item>
  /// <term>+ - ! ~</term>
  /// <description>Unary plus &bull; Unary minus &bull; Logical negation &bull;
  /// Bitwise complement</description>
  /// <description>right-to-left</description>
  /// </item>
  /// <item>
  /// <term>* / %</term>
  /// <description>Multiplication &bull; Division &bull; Modulus</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>+ -</term>
  /// <description>Addition &bull; Subtraction</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>&lt;&lt; &gt;&gt;</term>
  /// <description>Bitwise shift left &bull; Bitwise shift right</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>&lt; &lt;= &gt; &gt;=</term>
  /// <description>Relational less than &bull; less than or equal to &bull;
  /// greater than &bull; greater than or equal to</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>== !=</term>
  /// <description>Relational is equal to &bull; is not equal to</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>&amp;</term>
  /// <description>Bitwise AND</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>^</term>
  /// <description>Bitwise exclusive OR</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>|</term>
  /// <description>Bitwise inclusive OR</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>&amp;&amp;</term>
  /// <description>Logical AND</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>||</term>
  /// <description>Logical OR</description>
  /// <description>left-to-right</description>
  /// </item>
  /// <item>
  /// <term>?:</term>
  /// <description>Ternary conditional</description>
  /// <description>left-to-right</description>
  /// </item>
  /// </list>
  ///
  /// Parenthesis are also used to group sub-expressions to force a different
  /// precedence; such parenthetical expressions can be nested and are evaluated
  /// from inner to outer.</remarks>
{$ENDREGION}

  TCExpression = Class(TObject)
  Private
    FSource: String;
    FVars: TStringList;
    Expr: TCExprNode;
    Function GetFormattedSource: String; Inline;
    Function GetVarCount: Integer; Inline;
    Function GetVarNames(Index: Integer): String; Inline;
    Function GetVars(Const Name: String): TCExprVar;
    Function Parse(Const Expression: String): TCExprNode;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the expression evaluator for a specified expression in
    /// C language syntax.</summary>
    /// <param name="Expression">
    /// The expression to parse and evaluate.</param>
    /// <exception cref="ECExpressionError">
    /// Occurs when the expression has a syntax error.</exception>
{$ENDREGION}
    Constructor Create(Const Expression: String);
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Evaluates the expression for the variables specified by the <see cref="Vars"/>
    /// property.</summary>
    /// <returns>
    /// Returns the result of the expression.</returns>
{$ENDREGION}
    Function Evaluate: Integer; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the value of expression depends on value of any variables.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the value of the expression depends on
    /// one or more variables. Otherwise, returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsVariable: Boolean; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the source expression.</summary>
    /// <seealso cref="FormattedSource"/>
{$ENDREGION}
    Property Source: String Read FSource;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the source expression as formatted.</summary>
    /// <seealso cref="Source"/>
{$ENDREGION}
    Property FormattedSource: String Read GetFormattedSource;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the numnber of variables in the expression.</summary>
{$ENDREGION}
    Property VarCount: Integer Read GetVarCount;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists the name of variables in the expression.</summary>
{$ENDREGION}
    Property VarNames[Index: Integer]: String Read GetVarNames;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists the variables in the expression.</summary>
    /// <remarks>
    /// Use the Vars property to get or set values of variables in the expression.
    ///
    /// TCExpression does not prevent you to set value for variables that are not used by
    /// the expression. This helps you to manage the variables within your code independent
    /// of the expression being used. You can examine the value of <see cref="TCExprVar.RefCount"/>
    /// property to discover whether a veriable is referenced in the expression or not.</remarks>
    /// <seealso cref="TCExprVar"/>
{$ENDREGION}
    Property Vars[Const Name: String]: TCExprVar Read GetVars;
  End;

Implementation

Resourcestring
  SUnexpectedEndOfExpressionError = 'Unexpected end of expression';
  SUnexpectedSymbolError = 'Unexpected symbol ''%s''';
  SMissingExpectedSymbolError = 'Missing symbol ''%s''';

Const
  UnaryOpSymbols: Array [TCUnaryOp] Of String = ('+', '-', '!', '~');
  BinaryOpSymbols: Array [TCBinaryOp] Of String = ('+', '-', '*', '/', '%', '&', '|', '^', '<<', '>>', '==',
    '!=', '>', '>=', '<', '<=', '&&', '||');

  { ECExpressionError }

Procedure ECExpressionError.SetErrorDetails(Const AExpression: String; AOffset: Integer; Const AToken: String);
Begin
  FExpression := AExpression;
  FOffset := AOffset;
  FToken := AToken;
End;

{ TCExprVar }

Constructor TCExprVar.Create(Const AName: String);
Begin
  FName := AName;
End;

Function TCExprVar.AddRef: TCExprVar;
Begin
  Inc(FRefCount);
  Result := Self;
End;

{ TCExprVarNode }

Constructor TCExprVarNode.Create(Const AVariable: TCExprVar);
Begin
  FVariable := AVariable.AddRef;
End;

Function TCExprVarNode.GetAsString: String;
Begin
  Result := Variable.Name;
End;

Function TCExprVarNode.GetValue: Integer;
Begin
  Result := Variable.Value;
End;

Function TCExprVarNode.IsVariable: Boolean;
Begin
  Result := True;
End;

{ TCExprConstNode }

Constructor TCExprConstNode.Create(AValue: Integer);
Begin
  FValue := AValue;
End;

Function TCExprConstNode.GetAsString: String;
Begin
  Result := IntToStr(Value)
End;

Function TCExprConstNode.GetValue: Integer;
Begin
  Result := FValue;
End;

Function TCExprConstNode.IsVariable: Boolean;
Begin
  Result := False;
End;

{ TCExprUnaryNode }

Constructor TCExprUnaryNode.Create(AOp: TCUnaryOp; AOperand: TCExprNode);
Begin
  FOp := AOp;
  FOperand := AOperand;
End;

Destructor TCExprUnaryNode.Destroy;
Begin
  FOperand.Free;
  Inherited Destroy;
End;

Function TCExprUnaryNode.GetAsString: String;

  Function ToString(Operand: TCExprNode): String;
  Begin
    Result := Operand.AsString;
    If (Operand Is TCExprUnaryNode) Or (Operand Is TCExprBinaryNode) Or (Operand Is TCExprTernaryNode) Then
      Result := '(' + Result + ')';
  End;

Begin
  Result := UnaryOpSymbols[Op] + ToString(Operand);
End;

Function TCExprUnaryNode.GetValue: Integer;
Begin
  Case Op Of
    OpNegate:
      Result := -Operand.Value;
    OpBitNot:
      Result := Not Operand.Value;
    OpLogNot:
      Result := Ord(Operand.Value = 0);
  Else
    Result := Operand.Value
  End;
End;

Function TCExprUnaryNode.IsVariable: Boolean;
Begin
  Result := Operand.IsVariable;
End;

{ TCExprBinaryNode }

Constructor TCExprBinaryNode.Create(AOp: TCBinaryOp; AOperandLeft, AOperandRight: TCExprNode);
Begin
  FOp := AOp;
  FOperandLeft := AOperandLeft;
  FOperandRight := AOperandRight;
End;

Destructor TCExprBinaryNode.Destroy;
Begin
  FOperandLeft.Free;
  FOperandRight.Free;
  Inherited Destroy;
End;

Function TCExprBinaryNode.GetAsString: String;

  Function ToString(Operand: TCExprNode): String;
  Begin
    Result := Operand.AsString;
    If (Operand Is TCExprTernaryNode) Or
      ((Operand Is TCExprBinaryNode) And ((TCExprBinaryNode(Operand).Op <> Op) Or Not(Op In [OpMultiply, OpAdd,
      OpSubtract, OpBitAnd, OpBitOr, OpBitXor, OpLogAnd, OpLogOr])) And
      (Not(TCExprBinaryNode(Operand).Op In [OpAdd, OpSubtract]) Or Not(Op In [OpAdd, OpSubtract]))) Then
      Result := '(' + Result + ')';
  End;

Begin
  Result := ToString(OperandLeft) + ' ' + BinaryOpSymbols[Op] + ' ' + ToString(OperandRight);
End;

Function TCExprBinaryNode.GetValue: Integer;
Begin
  Case Op Of
    OpAdd:
      Result := OperandLeft.Value + OperandRight.Value;
    OpSubtract:
      Result := OperandLeft.Value - OperandRight.Value;
    OpMultiply:
      Result := OperandLeft.Value * OperandRight.Value;
    OpDivide:
      Result := OperandLeft.Value Div OperandRight.Value;
    OpModulus:
      Result := OperandLeft.Value Mod OperandRight.Value;
    OpBitAnd:
      Result := OperandLeft.Value And OperandRight.Value;
    OpBitOr:
      Result := OperandLeft.Value Or OperandRight.Value;
    OpBitXor:
      Result := OperandLeft.Value Xor OperandRight.Value;
    OpShiftLeft:
      Result := OperandLeft.Value Shl OperandRight.Value;
    OpShiftRight:
      Result := OperandLeft.Value Shr OperandRight.Value;
    OpEqual:
      Result := Ord(OperandLeft.Value = OperandRight.Value);
    OpNotEqual:
      Result := Ord(OperandLeft.Value <> OperandRight.Value);
    OpGreaterThan:
      Result := Ord(OperandLeft.Value > OperandRight.Value);
    OpGreaterThanOrEqual:
      Result := Ord(OperandLeft.Value >= OperandRight.Value);
    OpLessThan:
      Result := Ord(OperandLeft.Value < OperandRight.Value);
    OpLessThanOrEqual:
      Result := Ord(OperandLeft.Value <= OperandRight.Value);
    OpLogAnd:
      Begin
        Result := Ord(OperandLeft.Value <> 0);
        If Result <> 0 Then
          Result := Ord(OperandRight.Value <> 0);
      End;
    OpLogOr:
      Begin
        Result := Ord(OperandLeft.Value <> 0);
        If Result = 0 Then
          Result := Ord(OperandRight.Value <> 0);
      End
  Else
    Result := 0; // To prevent compiler warning; code never reaches here.
  End;
End;

Function TCExprBinaryNode.IsVariable: Boolean;
Begin
  Result := OperandLeft.IsVariable Or OperandRight.IsVariable;
End;

{ TCExprTernaryNode }

Constructor TCExprTernaryNode.Create(ACondition, AOperandTrue, AOperandFalse: TCExprNode);
Begin
  FCondition := ACondition;
  FOperandTrue := AOperandTrue;
  FOperandFalse := AOperandFalse;
End;

Destructor TCExprTernaryNode.Destroy;
Begin
  FCondition.Free;
  FOperandTrue.Free;
  FOperandFalse.Free;
  Inherited Destroy;
End;

Function TCExprTernaryNode.GetAsString: String;

  Function ToString(Operand: TCExprNode): String;
  Begin
    Result := Operand.AsString;
    If (Operand Is TCExprBinaryNode) Or (Operand Is TCExprTernaryNode) Then
      Result := '(' + Result + ')';
  End;

Begin
  Result := ToString(Condition) + ' ? ' + ToString(OperandTrue) + ' : ' + ToString(OperandFalse);
End;

Function TCExprTernaryNode.GetValue: Integer;
Begin
  If Condition.Value <> 0 Then
    Result := OperandTrue.Value
  Else
    Result := OperandFalse.Value;
End;

Function TCExprTernaryNode.IsVariable: Boolean;
Begin
  If Not Condition.IsVariable Then
    If Condition.Value <> 0 Then
      Result := OperandTrue.IsVariable
    Else
      Result := OperandFalse.IsVariable
  Else
    Result := True;
End;

{ TCExpression }

Constructor TCExpression.Create(Const Expression: String);
Begin
  FSource := Expression;
  FVars := TStringList.Create;
  FVars.CaseSensitive := False;
  FVars.OwnsObjects := True;
  FVars.Sorted := True;
  FVars.Duplicates := DupIgnore;
  Expr := Parse(Expression);
End;

Destructor TCExpression.Destroy;
Begin
  If Assigned(Expr) Then
    Expr.Free;
  FVars.Free;
  Inherited Destroy;
End;

Function TCExpression.GetFormattedSource: String;
Begin
  Result := Expr.AsString;
End;

Function TCExpression.GetVarCount: Integer;
Begin
  Result := FVars.Count;
End;

Function TCExpression.GetVarNames(Index: Integer): String;
Begin
  Result := FVars[Index];
End;

Function TCExpression.GetVars(Const Name: String): TCExprVar;
Var
  Index: Integer;
Begin
  Index := FVars.Add(Name);
  Result := TCExprVar(FVars.Objects[Index]);
  If Result = Nil Then
  Begin
    Result := TCExprVar.Create(Name);
    FVars.Objects[Index] := Result;
  End;
End;

Function TCExpression.Evaluate: Integer;
Begin
  Result := Expr.Value;
End;

Function TCExpression.IsVariable: Boolean;
Begin
  Result := Expr.IsVariable;
End;

Function TCExpression.Parse(Const Expression: String): TCExprNode;

Type
  TCExprToken = (CeNone, CeIdentifier, CeNumber, CeSymbol);
  TBinaryOperations = Set Of TCBinaryOp;

Const
  BinaryOpPrecedence: Array [1 .. 10] Of TBinaryOperations = ([OpMultiply, OpDivide, OpModulus],
    [OpAdd, OpSubtract], [OpShiftLeft, OpShiftRight], [OpGreaterThan, OpGreaterThanOrEqual, OpLessThan,
    OpLessThanOrEqual], [OpEqual, OpNotEqual], [OpBitAnd], [OpBitXor], [OpBitOr], [OpLogAnd], [OpLogOr]);

Var
  S: PChar;
  Token: String;
  TokenID: TCExprToken;

  Procedure NextToken;
  Var
    P: PChar;
  Begin
    While CharInSet(S^, [' ', ^I, #10, #13]) Do
      Inc(S);
    P := S;
    Case S^ Of
      #0:
        Begin
          Token := '';
          TokenID := CeNone;
        End;
      '0' .. '9':
        Begin
          Repeat
            Inc(S)
          Until Not CharInSet(S^, ['0' .. '9']);
          SetString(Token, P, S - P);
          TokenID := CeNumber;
        End;
      '_', 'A' .. 'Z', 'a' .. 'z':
        Begin
          Repeat
            Inc(S)
          Until Not CharInSet(S^, ['_', 'A' .. 'Z', 'a' .. 'z', '0' .. '9']);
          SetString(Token, P, S - P);
          TokenID := CeIdentifier;
        End;
    Else
      Inc(S);
      Case P^ Of
        '<':
          If CharInSet(S^, ['<', '=']) Then
            Inc(S);
        '>':
          If CharInSet(S^, ['>', '=']) Then
            Inc(S);
        '=':
          If S^ = '=' Then
            Inc(S);
        '!':
          If S^ = '=' Then
            Inc(S);
        '&':
          If S^ = '&' Then
            Inc(S);
        '|':
          If S^ = '|' Then
            Inc(S);
      End;
      SetString(Token, P, S - P);
      TokenID := CeSymbol;
    End;
  End;

  Function IsTokenUnaryOp(Out UnaryOp: TCUnaryOp): Boolean;
  Var
    Op: TCUnaryOp;
  Begin
    If TokenID = CeSymbol Then
      For Op := Low(UnaryOpSymbols) To High(UnaryOpSymbols) Do
        If UnaryOpSymbols[Op] = Token Then
        Begin
          UnaryOp := Op;
          Result := True;
          Exit;
        End;
    Result := False;
  End;

  Function IsTokenBinaryOp(Out BinaryOp: TCBinaryOp): Boolean;
  Var
    Op: TCBinaryOp;
  Begin
    If TokenID = CeSymbol Then
      For Op := Low(BinaryOpSymbols) To High(BinaryOpSymbols) Do
        If BinaryOpSymbols[Op] = Token Then
        Begin
          BinaryOp := Op;
          Result := True;
          Exit;
        End;
    Result := False;
  End;

  Function ParseExpression: TCExprNode; Forward;

    Function ParseFactor: TCExprNode;
    Var
      UnaryOp, Temp: TCUnaryOp;
      Value: Integer;
    Begin
      If IsTokenUnaryOp(UnaryOp) Then
        NextToken
      Else
        UnaryOp := OpNone;
      Case TokenID Of
        CeIdentifier:
          Begin
            Result := TCExprVarNode.Create(Vars[Token]);
            NextToken;
          End;
        CeNumber:
          Begin
            Result := TCExprConstNode.Create(StrToInt(Token));
            NextToken;
          End;
        CeSymbol:
          Begin
            If Token = '(' Then
            Begin
              NextToken;
              Result := ParseExpression;
              If Token <> ')' Then
              Begin
                Result.Free;
                Raise ECExpressionError.CreateResFmt(@SMissingExpectedSymbolError, [')']);
              End;
              NextToken;
            End
            Else If IsTokenUnaryOp(Temp) Then
              Result := ParseFactor
            Else
              Raise ECExpressionError.CreateResFmt(@SUnexpectedSymbolError, [Token]);
          End;
      Else
        Raise ECExpressionError.CreateRes(@SUnexpectedEndOfExpressionError);
      End;
      If UnaryOp <> OpNone Then
      Begin
        Result := TCExprUnaryNode.Create(UnaryOp, Result);
        If Not Result.IsVariable Then
        Begin
          Try
            Value := Result.Value;
          Finally
            Result.Free;
          End;
          Result := TCExprConstNode.Create(Value);
        End;
      End;
    End;

    Function ParseTerm(Level: Integer = High(BinaryOpPrecedence)): TCExprNode;
    Var
      NextOperand: TCExprNode;
      BinaryOp: TCBinaryOp;
      Value: Integer;
    Begin
      If Level = Low(BinaryOpPrecedence) Then
        Result := ParseFactor
      Else
        Result := ParseTerm(Level - 1);
      While IsTokenBinaryOp(BinaryOp) And (BinaryOp In BinaryOpPrecedence[Level]) Do
      Begin
        NextToken;
        Try
          If Level = Low(BinaryOpPrecedence) Then
            NextOperand := ParseFactor
          Else
            NextOperand := ParseTerm(Level - 1);
        Except
          Result.Free;
          Raise;
        End;
        Result := TCExprBinaryNode.Create(BinaryOp, Result, NextOperand);
        If Not Result.IsVariable Then
        Begin
          Try
            Value := Result.Value;
          Finally
            Result.Free;
          End;
          Result := TCExprConstNode.Create(Value);
        End;
      End;
    End;

    Function ParseExpression: TCExprNode;
    Var
      OperandTrue, OperandFalse: TCExprNode;
    Begin
      Result := ParseTerm;
      If Token = '?' Then
      Begin
        NextToken;
        Try
          OperandTrue := ParseExpression;
        Except
          ;
          Result.Free;
          Raise;
        End;
        If Token <> ':' Then
        Begin
          Result.Free;
          Raise ECExpressionError.CreateResFmt(@SMissingExpectedSymbolError, [':']);
        End;
        NextToken;
        Try
          OperandFalse := ParseExpression;
        Except
          Result.Free;
          OperandTrue.Free;
          Raise;
        End;
        If Result.IsVariable Then
          Result := TCExprTernaryNode.Create(Result, OperandTrue, OperandFalse)
        Else
        Begin
          If Result.Value <> 0 Then
          Begin
            Result.Free;
            Result := OperandTrue;
            OperandFalse.Free;
          End
          Else
          Begin
            Result.Free;
            Result := OperandFalse;
            OperandTrue.Free;
          End;
        End;
      End;
    End;

  Var
    Error: ECExpressionError;
  Begin
    S := PChar(Expression);
    NextToken;
    Try
      Result := ParseExpression;
      If TokenID <> CeNone Then
      Begin
        Result.Free;
        Raise ECExpressionError.CreateResFmt(@SUnexpectedSymbolError, [Token]);
      End;
    Except
      On E: ECExpressionError Do
      Begin
        E.SetErrorDetails(Expression, S - PChar(Expression), Token);
        Raise;
      End;
      On E: Exception Do
      Begin
        Error := ECExpressionError.Create(E.Message);
        Error.SetErrorDetails(Expression, S - PChar(Expression), Token);
        Raise Error;
      End;
    End;
  End;

End.
