{ ------------------------------------------------------------------------------ }
{ }
{ i18n Package }
{ by Kambiz R. Khojasteh }
{ }
{ kambiz@delphiarea.com }
{ http://www.delphiarea.com }
{ }
{ ------------------------------------------------------------------------------ }

/// This unit implements classes and functions to collect information and work
/// with locales, countries and currencies.
Unit i18nCore;

{$I DELPHIAREA.INC}
{ -$DEFINE EXCLUDE_TERRITORIES_WITH_NO_CULTURE_INFO }
{$DEFINE EXCLUDE_TERRITORIES_WITH_X_CODE}

Interface

Uses
  Windows, SysUtils, Classes, I18nHashList, I18nUnicode,{ I18nPlurals,}
  I18nCalendar, I18nCalendars, I18nWinNLS;

Const
{$REGION 'xmldoc'}
  /// <summary>
  /// Specifies the start range of custom primary language identifers.</summary>
{$ENDREGION}
  CUSTOM_LANG = $0200;
{$REGION 'xmldoc'}
  /// <summary>
  /// Specifies the end range of custom primary language identifers.</summary>
{$ENDREGION}
  CUSTOM_LANG_LAST = $03FF;
{$REGION 'xmldoc'}
  /// <summary>
  /// Specifies the start range of custom secondary language identifers.</summary>
{$ENDREGION}
  CUSTOM_SUBLANG = $20;
{$REGION 'xmldoc'}
  /// <summary>
  /// Specifies the end range of custom secondary language identifers.</summary>
{$ENDREGION}
  CUSTOM_SUBLANG_LAST = $3F;

Const
{$REGION 'xmldoc'}
  /// <summary>
  /// Lists nominal digits as a string of characters.</summary>
{$ENDREGION}
  NOMINAL_DIGITS = '0123456789';

Type

  TCultureInfo = Class;
  TReadonlyCultureList = Class;
  TCultureList = Class;
  TTerritoryInfo = Class;
  TReadonlyTerritoryList = Class;
  TTerritoryList = Class;
  TCurrencyInfo = Class;
  TReadonlyCurrencyList = Class;
  TCurrencyList = Class;

{$REGION 'xmldoc'}
  /// <summary>
  /// This data type represents a pointer to <see cref="TFormatSettings"/> record.</summary>
{$ENDREGION}
  PFormatSettings = ^TFormatSettings;

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the type of code pages.</summary>
{$ENDREGION}
  TCodePageType = (
{$REGION 'xmldoc'}
    /// OEM code page
{$ENDREGION}
    CpOEM,
{$REGION 'xmldoc'}
    /// ANSI code page
{$ENDREGION}
    CpANSI,
{$REGION 'xmldoc'}
    /// MAC code page
{$ENDREGION}
    CpMAC);

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the measurement systems.</summary>
{$ENDREGION}
  TMeasurementSystem = (
{$REGION 'xmldoc'}
    /// S.I. system
{$ENDREGION}
    MsMetric,
{$REGION 'xmldoc'}
    /// U.S. system
{$ENDREGION}
    MsUS);

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the reading direction of languages.</summary>
{$ENDREGION}
  TReadingLayout = (
{$REGION 'xmldoc'}
    /// Read from left to right, as for the English (United States) locale.
{$ENDREGION}
    RlLeftToRight,
{$REGION 'xmldoc'}
    /// Read from right to left, as for Arabic locales.
{$ENDREGION}
    RlRightToLeft,
{$REGION 'xmldoc'}
    /// Read vertically from top to bottom with columns to the right, and read
    /// from right to left, as for the Japanese (Japan) locale.
{$ENDREGION}
    RlTopToBottomColRightToLeft,
{$REGION 'xmldoc'}
    /// Read vertically from top to bottom with columns proceeding to the right,
    /// as for the Mongolian (Mongolian) locale.
{$ENDREGION}
    RlTopToBottomColLeftToRight);

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the possible display names for
  /// <see cref="TCultureInfo"/> objects.</summary>
{$ENDREGION}
  TCultureDisplayName = (
{$REGION 'xmldoc'}
    /// Nothing
{$ENDREGION}
    CnNone,
{$REGION 'xmldoc'}
    /// Locale name, a multi-part tag to uniquely identify the locale.
{$ENDREGION}
    CnLocale,
{$REGION 'xmldoc'}
    /// Two characters of the international language code specified in ISO 639-1.
{$ENDREGION}
    CnLanguage,
{$REGION 'xmldoc'}
    /// Three characters of the international language code specified in ISO 639-2/T.
{$ENDREGION}
    CnLanguage3,
{$REGION 'xmldoc'}
    /// Full localized (in Windows user's language) name of the locale, for example,
    /// Deutsch (Deutschland) for German (Germany).
{$ENDREGION}
    CnLocalizedDisplayName,
{$REGION 'xmldoc'}
    /// Display name of the locale in English. Usually the display name consists of
    /// the language and the country/region, for example, German (Germany) for
    /// Deutsch (Deutschland).
{$ENDREGION}
    CnEnglishDisplayName,
{$REGION 'xmldoc'}
    /// Display name of the locale in its native language, for example, Deutsch
    /// (Deutschland) for the locale German (Germany).
{$ENDREGION}
    CnNativeDisplayName,
{$REGION 'xmldoc'}
    /// Full localized (in Windows user's language) name of the language, for example,
    /// Deutsch representing German.
{$ENDREGION}
    CnLocalizedLanguageName,
{$REGION 'xmldoc'}
    /// English name of the language, for example, German for Deutsch, from
    /// International ISO Standard 639.
{$ENDREGION}
    CnEnglishLanguageName,
{$REGION 'xmldoc'}
    /// Native name of the language, for example, Deutsch for German (Germany).
{$ENDREGION}
    CnNativeLanguageName,
{$REGION 'xmldoc'}
    /// Native name of the country/region associated with the locale, in the native
    /// language of the locale.
{$ENDREGION}
    CnNativeCountryName,
{$REGION 'xmldoc'}
    /// Native name of the currency associated with the locale, in the native language
    /// of the locale.
{$ENDREGION}
    CnNativeCurrencyName);

{$REGION 'xmldoc'}
  /// <summary>
  /// This class represents a culture (locale).</summary>
  /// <remarks>
  /// Each instance of TCultureInfo class represents a specific locale. This class has
  /// properties and methods to query information about the locale, and perform the
  /// locale specific tasks.
  ///
  /// NOTE: Do not create or destroy an instance of TCultureInfo class. Use <see cref="CultureOf"/>
  /// global function to get TCultureInfo instance of a locale.</remarks>
  /// <seealso cref="CultureOf"/>
  /// <seealso cref="World.Cultures"/>
{$ENDREGION}

  TCultureInfo = Class Sealed(TObject)
  Private
    FLocaleID: LCID;
    FGeoID: GEOID;
    FLocale: String;
    FLanguage2: String;
    FLanguage3: String;
    FScriptCodes: String;
    FNativeDigits: String;
    FNativeCalendarType: TCalendarClass;
    FCalendarTypes: TCalendarClassList;
    FScripts: TUnicodeScripts;
    FReadingLayout: TReadingLayout;
    FMeasurementSystem: TMeasurementSystem;
    FDefaultCodePages: Array [TCodePageType] Of Integer;
    FNativeCalendar: TCalendar;
    FLocalizedDisplayName: String;
    FEnglishDisplayName: String;
    FNativeDisplayName: String;
    FLocalizedLanguageName: String;
    FEnglishLanguageName: String;
    FNativeLanguageName: String;
    FNativeCountryName: String;
    FNativeCurrencyName: String;
    FPercentNegative: String;
    FPercentPositive: String;
//    FPluralRule: String;
    FCountry: TTerritoryInfo;
    FCurrency: TCurrencyInfo;
    FFormatSettings: PFormatSettings;
    Function GetLangID: LANGID; Inline;
    Function GetPrimaryLangID: LANGID; Inline;
    Function GetSubLangID: LANGID; Inline;
    Function GetBiDiMode: TBiDiMode; Inline;
    Function GetDefaultCodePages(CodePageType: TCodePageType): Integer;
    Function GetDisplayNames(DisplayName: TCultureDisplayName): String;
    Function GetFormatSettings: PFormatSettings;
    Function GetNativeCalendar: TCalendar;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for the locale specified by its identifier.</summary>
    /// <param name="ALocaleID">
    /// The identifier of the locale.</param>
{$ENDREGION}
    Constructor Create(ALocaleID: LCID); Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for the locale specified by its name.</summary>
    /// <param name="ALocale">
    /// The name of the locale.</param>
{$ENDREGION}
    Constructor Create(Const ALocale: String); Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroyes the object and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a string that its embedded nominal digits (0-9) are substituted with
    /// their corresponding digits in this culture.</summary>
    /// <remarks>
    /// NominalDigitsToNative substitutes the nominal digits in the string specified
    /// by <paramref name="Str"/> to their corresponding native digits of this culture.
    ///
    /// If the argument passed as <paramref name="IgnoreUCC"/> parameter is false, the
    /// method will not substitute nominal digits that are marked with NODS (Nominal digit
    /// shapes) Unicode control character, otherwise all nominal digits will be substituted.</remarks>
    /// <param name="Str">
    /// The string to convert.</param>
    /// <param name="IgnoreUCCs">
    /// Indicates whether ignore Unicode control characters.</param>
    /// <returns>
    /// The string that its digits are represented in native digits.</returns>
    /// <seealso cref="NominalDigitsToNative"/>
    /// <seealso cref="FreezeDigits"/>
    /// <seealso cref="UnfreezeDigits"/>
    /// <see cref="UCC_NODS"/>
{$ENDREGION}
    Function NominalDigitsToNative(Const Str: String; IgnoreUCCs: Boolean = False): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a string that its embedded native digits are substitued with their
    /// corresponding nominal digits (0-9).</summary>
    /// <remarks>
    /// NativeDigitsToNominal substitutes the native digits of this culture in the
    /// string specified by <paramref name="Str"/> to their corresponding nominal digits.
    ///
    /// If the argument passed as <paramref name="IgnoreUCC"/> parameter is false, the
    /// method will not substitute native digits that are marked with NADS (Native digit
    /// shapes substitution) Unicode control character, otherwise all native digits will
    /// be substituted.</remarks>
    /// <param name="Str">
    /// The string to convert.</param>
    /// <param name="IgnoreUCCs">
    /// Indicates whether ignore Unicode control characters.</param>
    /// <returns>
    /// The string that its digits are represented in nominal digits.</returns>
    /// <seealso cref="NativeDigitsToNominal"/>
    /// <seealso cref="FreezeDigits"/>
    /// <seealso cref="UnfreezeDigits"/>
    /// <see cref="UCC_NADS"/>
{$ENDREGION}
    Function NativeDigitsToNominal(Const Str: String; IgnoreUCCs: Boolean = False): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a string that its embeded digits will always be displayed as either native
    /// or nominal, regardless of Windows settings.</summary>
    /// <remarks>
    /// Based on the user's selection (configurable via Windows control panel), Windows may
    /// display nominal digits (0-9) in different digit shapes.
    ///
    /// FreezeDigits inserts appropriate Unicode control character in front of the string
    /// to prevent automatic digit substitution by Windows.
    ///
    /// If argument passed as <paramref name="NativeDigits"/> is true, FreezeDigits will
    /// also substitude nominal digits with native digits.</remarks>
    /// <param name="Str">
    /// The string to convert.</param>
    /// <param name="UseNativeDigits">
    /// Indicates whether to use native digits or nominal digits.</param>
    /// <returns>
    /// The string that its digits are fixed as either nominal or native.</returns>
    /// <seealso cref="UnfreezeDigits"/>
    /// <seealso cref="NominalDigitsToNative"/>
    /// <seealso cref="NativeDigitsToNominal"/>
{$ENDREGION}
    Function FreezeDigits(Const Str: String; UseNativeDigits: Boolean): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a string that its embedded digits can be displayed automatically as either
    /// native or nominal by Windows.</summary>
    /// <remarks>
    /// UnfreezeDigits reverts changes made by <see cref="FreezeDigits"/> on the string
    /// specified by the <paramref name="Str"/> parameter.
    ///
    /// That means, if the string specfied by the <paramref name="Str"/> parameter has
    /// NODS (Nominal digit shapes) Unicode control character in front, the control
    /// character will be removed and its native digits will be substituted with their
    /// corresponding nominal digits.</remarks>
    /// <param name="Str">
    /// The string to convert.</param>
    /// <returns>
    /// The string that its digits can be displayed as either native or nominal.</returns>
    /// <seealso cref="FreezeDigits"/>
{$ENDREGION}
    Function UnfreezeDigits(Const Str: String): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a formatted string assembled from a format string and an array of
    /// arguments.</summary>
    /// <remarks>
    /// Format method acts like Format function of Delphi, except that it uses this
    /// culture's preferences for formatting the values.
    ///
    /// Optionally, this method can represent numbers in native digits.</remarks>
    /// <param name="Fmt">
    /// The format string as in used by Delphi's standard Format function.</param>
    /// <param name="Args">
    /// The array of arguments to apply to the format specifiers in the format string.</param>
    /// <param name="UseNativeDigits">
    /// Indicates whether to use native digits or nominal digits for numbers.</param>
    /// <returns>
    /// The formatted string.</returns>
{$ENDREGION}
    Function Format(Const Fmt: String; Const Args: Array Of Const; UseNativeDigits: Boolean = True): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Formats a floating point value.</summary>
    /// <remarks>
    /// FormatNumber method acts like FormatFloat function of Delphi, except that it
    /// uses this culture's preferences for formatting the number.
    ///
    /// Optionally, this method can represent the number in native digits.</remarks>
    /// <param name="Fmt">
    /// The format string.</param>
    /// <param name="Value">
    /// The floating point value to format.</param>
    /// <param name="UseNativeDigits">
    /// Indicates whether to use native digits or nominal digits for the number.</param>
    /// <returns>
    /// The formatted string.</returns>
{$ENDREGION}
    Function FormatNumber(Const Fmt: String; Const Value: Extended; UseNativeDigits: Boolean = True): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Formats a floating point value as a percentage.</summary>
    /// <remarks>
    /// FormatPercent represents a floating point value as a percentage using the
    /// percentage format of this culture.
    ///
    /// Optionally, this method can represent the percentage in native digits.</remarks>
    /// <param name="Value">
    /// The floating point value to format.</param>
    /// <param name="Decimals">
    /// The number of decimals.</param>
    /// <param name="UseNativeDigits">
    /// Indicates whether to use native digits or nominal digits for the percentage.</param>
    /// <returns>
    /// The formatted string.</returns>
{$ENDREGION}
    Function FormatPercent(Const Value: Extended; Decimals: Integer = 2;
      UseNativeDigits: Boolean = True): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Formats a <see cref="TDateTime"/> value.</summary>
    /// <remarks>
    /// FormatDateTime formats the <see cref="TDateTime"/> value given by <paramref name="Value"/>
    /// using the format given by <paramref name="Fmt"/>.
    ///
    /// FormatDateTime uses <see cref="NativeCalendar"/> object to extrcat components of
    /// the specified <see cref="TDateTime"/> value.
    ///
    /// Optionally, this method can represent the numbers in native digits.</remarks>
    /// <param name="Fmt">
    /// The format string. See the <see cref="TCalendar.Format"/> method for details.</param>
    /// <param name="Value">
    /// The <see cref="TDateTime"/> value to format.</param>
    /// <param name="UseNativeDigits">
    /// Indicates whether to use native digits or nominal digits for the numbers.</param>
    /// <returns>
    /// The formatted date-time string.</returns>
    /// <seealso cref="NativeCalendar"/>
    /// <seealso cref="TryParseDateTime"/>
    /// <seealso cref="ParseDateTime"/>
{$ENDREGION}
    Function FormatDateTime(Const Fmt: String; Const Value: TDateTime; UseNativeDigits: Boolean = True): String;
{$REGION 'xmldoc'}
    /// <summary>
    /// Converts a string to a <see cref="TDateTime"/> value using a specified format
    /// string.</summary>
    /// <remarks>
    /// The ParseDateTime method scans the string given by <paramref name="Str"/>
    /// using the format string given by <paramref name="Fmt"/> to extract its
    /// date and time values.
    ///
    /// ParseDateTime uses the calendar object specified by <see cref="NativeCalendar"/>
    /// property to extract the date and time components from the input string.</remarks>
    /// <param name="Fmt">
    /// The format string. See the <see cref="TCalendar.Format"/> method for details.</param>
    /// <param name="Str">
    /// The date and time value as a string.</param>
    /// <returns>
    /// The <see cref="TDateTime"/> value.</returns>
    /// <exception cref="EConvertError">
    /// Occurs when the conversion is failed.</exception>
    /// <seealso cref="TryParseDateTime"/>
    /// <seealso cref="FormatDateTime"/>
{$ENDREGION}
    Function ParseDateTime(Const Fmt: String; Const Str: String): TDateTime;
{$REGION 'xmldoc'}
    /// <summary>
    /// Converts a string to a <see cref="TDateTime"/> value using a specified format
    /// string.</summary>
    /// <remarks>
    /// The TryParseDateTime method scans the string given by <paramref name="Str"/>
    /// using the format string given by <paramref name="Fmt"/> to extract its
    /// date and time values.
    ///
    /// TryParseDateTime uses the calendar object specified by <see cref="NativeCalendar"/>
    /// property to extract the date and time components of the input string.</remarks>
    /// <param name="Fmt">
    /// The format string. See the <see cref="TCalendar.Format"/> method for details.</param>
    /// <param name="Str">
    /// The date and time value as a string.</param>
    /// <param name="DateTime">
    /// The <see cref="TDateTime"/> value.</param>
    /// <returns>
    /// Returns <see langword="true"/> if the conversion is succeeded, otherwise
    /// returns <see langword="false"/>.</returns>
    /// <seealso cref="ParseDateTime"/>
    /// <seealso cref="FormatDateTime"/>
{$ENDREGION}
    Function TryParseDateTime(Const Fmt: String; Const Str: String; Var DateTime: TDateTime): Boolean;
{$REGION 'xmldoc'}
    /// <summary>
    /// Verifies whether a speicfied string used the scripts of this culture.</summary>
    /// <param name="Str">
    /// The string to examine.</param>
    /// <param name="AllowLatin">
    /// Indicates whether the Latin script is allowed anyway.</param>
    /// <returns>
    /// Returns <see langword="true"/> if the string uses scripts of this culture,
    /// otherwise returns <see langword="false"/>.</returns>
    /// <seealso cref="ScriptCodes"/>
    /// <seealso cref="Scripts"/>
{$ENDREGION}
    Function VerifyString(Const Str: String; AllowLatin: Boolean = True): Boolean;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Converts a string in default OEM code page of this culture's language to Unicode.</summary>
//    /// <param name="Str">
//    /// The OEM string.</param>
//    /// <returns>
//    /// The Unicode string.</returns>
//    /// <seealso cref="UnicodeToOem"/>
//    /// <seealso cref="AnsiToUnicode"/>
//    /// <seealso cref="MacToUnicode"/>
//    /// <seealso cref="DefaultCodePages"/>
//{$ENDREGION}
//    Function OemToUnicode(Const Str: RawByteString): String;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Converts a string from Unicode to default OEM code page of this culture's language.</summary>
//    /// <param name="Str">
//    /// The Unicode string</param>
//    /// <returns>
//    /// The OEM string.</returns>
//    /// <seealso cref="OemToUnicode"/>
//    /// <seealso cref="UnicodeToAnsi"/>
//    /// <seealso cref="UnicodeToMac"/>
//    /// <seealso cref="DefaultCodePages"/>
//{$ENDREGION}
//    Function UnicodeToOem(Const Str: String): RawByteString;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Converts a string in default ANSI code page of this culture's language to Unicode.</summary>
//    /// <param name="Str">
//    /// The ANSI string.</param>
//    /// <returns>
//    /// The Unicode string.</returns>
//    /// <seealso cref="UnicodeToAnsi"/>
//    /// <seealso cref="OemToUnicode"/>
//    /// <seealso cref="MacToUnicode"/>
//    /// <seealso cref="DefaultCodePages"/>
//{$ENDREGION}
//    Function AnsiToUnicode(Const Str: RawByteString): String;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Converts a string from Unicode to default ANSI code page of this culture's language.</summary>
//    /// <param name="Str">
//    /// The Unicode string</param>
//    /// <returns>
//    /// The ANSI string.</returns>
//    /// <seealso cref="AnsiToUnicode"/>
//    /// <seealso cref="UnicodeToOem"/>
//    /// <seealso cref="UnicodeToMac"/>
//    /// <seealso cref="DefaultCodePages"/>
//{$ENDREGION}
//    Function UnicodeToAnsi(Const Str: String): RawByteString;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Converts a string in default MAC code page of this culture's language to Unicode.</summary>
//    /// <param name="Str">
//    /// The MAC string.</param>
//    /// <returns>
//    /// The Unicode string.</returns>
//    /// <seealso cref="UnicodeToMac"/>
//    /// <seealso cref="AnsiToUnicode"/>
//    /// <seealso cref="OemToUnicode"/>
//    /// <seealso cref="DefaultCodePages"/>
//{$ENDREGION}
//    Function MacToUnicode(Const Str: RawByteString): String;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Converts a string from Unicode to default MAC code page of this culture's language.</summary>
//    /// <param name="Str">
//    /// The Unicode string</param>
//    /// <returns>
//    /// The MAC string.</returns>
//    /// <seealso cref="MacToUnicode"/>
//    /// <seealso cref="UnicodeToAnsi"/>
//    /// <seealso cref="UnicodeToOem"/>
//    /// <seealso cref="DefaultCodePages"/>
//{$ENDREGION}
//    Function UnicodeToMac(Const Str: String): RawByteString;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the adjusted DrawText flags based on the language direction of this
    /// culture.</summary>
    /// <param name="Flags">
    /// The DrawText flags to adjust.</param>
    /// <returns>
    /// The adjusted DrawText flags.</returns>
    /// <seealso cref="DrawTextBiDiModeFlagsReadingOnly"/>
{$ENDREGION}
    Function DrawTextBiDiModeFlags(Flags: Cardinal): Integer;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the DrawText reading order flag based on the language direction of
    /// this culture.</summary>
    /// <returns>
    /// The DrawText reading order flag.</returns>
    /// <seealso cref="DrawTextBiDiModeFlags"/>
{$ENDREGION}
    Function DrawTextBiDiModeFlagsReadingOnly: Integer; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a specified culture object has the same primary language
    /// of this culture.</summary>
    /// <param name="Culture">
    /// The culture object to examined.</param>
    /// <returns>
    /// Returns <see langword="true"/> if primary languages are identical, otherwise
    /// returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsDialectOf(Culture: TCultureInfo): Boolean; Overload; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a specified locale name represents the primary language
    /// of this culture.</summary>
    /// <param name="Locale">
    /// The locale name to examine.</param>
    /// <returns>
    /// Returns <see langword="true"/> if primary languages are identical, otherwise
    /// returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsDialectOf(Const Locale: String): Boolean; Overload; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a specified language identifier represents the primary
    /// language of this culture.</summary>
    /// <param name="LangID">
    /// The language identifier to examine.</param>
    /// <returns>
    /// Returns <see langword="true"/> if primary languages are identical, otherwise
    /// returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsDialectOf(LangID: LANGID): Boolean; Overload; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the nominal digits (0-9) are the native digits of this culture.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the nominal digits are the native digits of this
    /// culture, otherwise returns <see langword="false"/>.</returns>
    /// <see cref="NativeDigits"/>
{$ENDREGION}
    Function IsUsingNominalDigits: Boolean; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the Gregorian calendar in Common Era (C.E.) is the native digits
    /// of this culture.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the Gregorian calendar in Common Era is the native
    /// calendar of this culture, otherwise returns <see langword="false"/>.</returns>
    /// <see cref="NativeCalendar"/>
{$ENDREGION}
    Function IsUsingCommonEraCalendar: Boolean; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the language direction of this culture is from right to left.</summary>
    /// <returns>
    /// Returns <see langword="true"/> if the language direction of this culture
    /// is from right to left, otherwise returns <see langword="false"/>.</returns>
    /// <seealso cref="ReadingLayout"/>
{$ENDREGION}
    Function IsRightToLeft: Boolean; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether the language represented by this culture is the default
    /// dialect (sub-language) of its main language (primary language).</summary>
    /// <returns>
    /// Returns <see langword="true"/> if this culture's language is the default
    /// dialect of its primary language, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function IsDefault: Boolean; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the locale identifier of the culture.</summary>
{$ENDREGION}
    Property LocaleID: LCID Read FLocaleID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the language identifier of the culture.</summary>
{$ENDREGION}
    Property LangID: LANGID Read GetLangID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the primary language identifier of the culture.</summary>
{$ENDREGION}
    Property PrimaryLangID: LANGID Read GetPrimaryLangID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the sub-language identifier of the culture.</summary>
{$ENDREGION}
    Property SubLangID: LANGID Read GetSubLangID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the geographical identifier of the culture.</summary>
{$ENDREGION}
    Property GeoID: GEOID Read FGeoID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the locale name of the culture.</summary>
{$ENDREGION}
    Property Locale: String Read FLocale;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the two-character ISO 639-1 international language code of the culture.</summary>
{$ENDREGION}
    Property Language2: String Read FLanguage2;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the three-character ISO 639-2/T international language code of the culture.</summary>
{$ENDREGION}
    Property Language3: String Read FLanguage3;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets country information of the culture.</summary>
{$ENDREGION}
    Property Country: TTerritoryInfo Read FCountry;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets currency information of the culture.</summary>
{$ENDREGION}
    Property Currency: TCurrencyInfo Read FCurrency;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the four-character ISO 15924 script codes of the culture separated by
    /// semicolon.</summary>
{$ENDREGION}
    Property ScriptCodes: String Read FScriptCodes;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the set of culture's scripts.</summary>
{$ENDREGION}
    Property Scripts: TUnicodeScripts Read FScripts;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the default measurement system used by the culture.</summary>
{$ENDREGION}
    Property MeasurementSystem: TMeasurementSystem Read FMeasurementSystem;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native digits of the culture starting from 0 to 9.</summary>
{$ENDREGION}
    Property NativeDigits: String Read FNativeDigits;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the <see cref="TCalendar"/> class that manages the default calendar system
    /// used by the culture.</summary>
{$ENDREGION}
    Property NativeCalendarType: TCalendarClass Read FNativeCalendarType;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native calendar used by the culture.</summary>
{$ENDREGION}
    Property NativeCalendar: TCalendar Read GetNativeCalendar;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the reading order of the scripts in the culture.</summary>
{$ENDREGION}
    Property ReadingLayout: TReadingLayout Read FReadingLayout;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the TBiDiMode value of the culture's reading order.</summary>
{$ENDREGION}
    Property BiDiMode: TBiDiMode Read GetBiDiMode;
//{$REGION 'xmldoc'}
//    /// <summary>
//    /// Gets the language's plural rule that determines how the language handles
//    /// plurals of nouns or unit expressions.
//    ///
//    /// NOTE: During development of the i18n package, I could not find the plural
//    /// rule of some languages. For those languages, the PluralRule property will
//    /// return a default rule that is 'nplurals=1; plural=0; ASSUMED'. So, you can
//    /// check for existence of word 'ASSUMED' in a plural rule returned by this
//    /// property to find out whether the rule is correct or is just a default
//    /// assumption.</summary>
//    /// <seealso cref="TPluralForms"/>
//{$ENDREGION}
//    Property PluralRule: String Read FPluralRule;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the localized (in Windows user's language) name of the culture.</summary>
{$ENDREGION}
    Property LocalizedDisplayName: String Read FLocalizedDisplayName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the English name of the culture.</summary>
{$ENDREGION}
    Property EnglishDisplayName: String Read FEnglishDisplayName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native name of the culture.</summary>
{$ENDREGION}
    Property NativeDisplayName: String Read FNativeDisplayName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the localized (in Windows user's language) name of the culture's language.</summary>
{$ENDREGION}
    Property LocalizedLanguageName: String Read FLocalizedLanguageName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the English name of the culture's language.</summary>
{$ENDREGION}
    Property EnglishLanguageName: String Read FEnglishLanguageName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native name of the culture's language.</summary>
{$ENDREGION}
    Property NativeLanguageName: String Read FNativeLanguageName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native name of the culture's country.</summary>
{$ENDREGION}
    Property NativeCountryName: String Read FNativeCountryName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native name of the culture's currency.</summary>
{$ENDREGION}
    Property NativeCurrencyName: String Read FNativeCurrencyName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets pointer to the culture's format settings.</summary>
{$ENDREGION}
    Property FormatSettings: PFormatSettings Read GetFormatSettings;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists the default ANSI, OEM, and MAC codepages of the culture.</summary>
{$ENDREGION}
    Property DefaultCodePages[CodePageType: TCodePageType]: Integer Read GetDefaultCodePages;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists display names of the culture.</summary>
{$ENDREGION}
    Property DisplayNames[DisplayName: TCultureDisplayName]: String Read GetDisplayNames;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCultureEnumerator enumerates through the <see cref="TCultureInfo"/> objects
  /// of an instance of <see cref="TReadonlyCultureList"/> or its descendent class.</summary>
{$ENDREGION}

  TCultureEnumerator = Class(TListEnumerator)
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for a specified list.</summary>
    /// <param name="list">
    /// The list to enumerate its items.</param>
{$ENDREGION}
    Constructor Create(List: TReadonlyCultureList);
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the current <see cref="TCultureInfo"/> object of the list being
    /// enumerated.</summary>
    /// <returns>
    /// Returns the current <see cref="TCultureInfo"/> object.</returns>
{$ENDREGION}
    Function GetCurrent: TCultureInfo; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the current <see cref="TCultureInfo"/> object.</summary>
{$ENDREGION}
    Property Current: TCultureInfo Read GetCurrent;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This class maintains a read-only list of <see cref="TCultureInfo"/> objects.</summary>
  /// <remarks>
  /// TReadonlyCultureList maintains a read-only list of <see cref="TCultureInfo"/>
  /// objects, and provides properties and methods to locate and access the objects.
  /// The objects in the list are hashed and can be located quickly.</remarks>
  /// <see cref="TCultureList"/>
{$ENDREGION}

  TReadonlyCultureList = Class(TPersistent)
  Private
    Lookup_LocaleID: TKeyLookup<LCID, Integer>;
    Lookup_Locale: TKeyLookup<String, Integer>;
    IsLookupTableValid_LocaleID: Boolean;
    IsLookupTableValid_Locale: Boolean;
    Function GetCount: Integer; Inline;
    Function GetItems(Index: Integer): TCultureInfo; Inline;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Provides direct accress to the individual <see cref="TCultureInfo"/> objects.</summary>
{$ENDREGION}
    List: TList;
{$REGION 'xmldoc'}
    /// <summary>
    /// Attaches a specified <see cref="TCultureInfo"/> object with a specified index
    /// in the list.
    ///
    /// NOTE: This method does not add the object to the list, it only creats its lookup
    /// entries for all the lookup tables.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to add to the lookup tables.</param>
    /// <param name="Index">
    /// The index of the <see cref="TCultureInfo"/> object in the list.</param>
    /// <seealso cref="RemoveLookupEntries"/>
{$ENDREGION}
    Procedure AddLookupEntries(Culture: TCultureInfo; Index: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Detaches a specified <see cref="TCultureInfo"/> object from the previosuly attached
    /// index.
    ///
    /// NOTE: This method does not remove the object from the list, it only deletes its
    /// lookup entries from all lookup tables.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to remove from the lookup tables.</param>
    /// <seealso cref="AddLookupEntries"/>
{$ENDREGION}
    Procedure RemoveLookupEntries(Culture: TCultureInfo);
{$REGION 'xmldoc'}
    /// <summary>
    /// Clears all entries from all lookup tables.</summary>
{$ENDREGION}
    Procedure ClearLookupTables;
{$REGION 'xmldoc'}
    /// <summary>
    /// Schedules rebuilding of all lookup tables.</summary>
{$ENDREGION}
    Procedure InvalidateLookupTables; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Validates lookup table for <see cref="TCultureInfo.LocaleID"/> values.</summary>
{$ENDREGION}
    Procedure ValidateLookupTable_LocaleID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Validates lookup table for <see cref="TCultureInfo.Locale"/> values.</summary>
{$ENDREGION}
    Procedure ValidateLookupTable_Locale;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creats an instance of the class.</summary>
{$ENDREGION}
    Constructor Create; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a <see cref="TCultureEnumerator"/> reference, which enumerates
    /// all <see cref="TCultureInfo"/> objects in the list.</summary>
    /// <returns>
    /// Returns a reference to a <see cref="TCultureEnumerator"/> instance.</returns>
{$ENDREGION}
    Function GetEnumerator: TCultureEnumerator;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the first <see cref="TCultureInfo"/> object in the list.</summary>
    /// <returns>
    /// The first <see cref="TCultureInfo"/> object in the list.</returns>
    /// <seealso cref="Last"/>
{$ENDREGION}
    Function First: TCultureInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the last <see cref="TCultureInfo"/> object in the list.</summary>
    /// <returns>
    /// The last <see cref="TCultureInfo"/> object in the list.</returns>
    /// <seealso cref="First"/>
{$ENDREGION}
    Function Last: TCultureInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies <see cref="TCultureInfo"/> objects in the list to another object.</summary>
    /// <param name="Dest">
    /// The destination object.</param>
    /// <seealso cref="AssignLabelsTo"/>
{$ENDREGION}
    Procedure AssignTo(Dest: TPersistent); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies display name of <see cref="TCultureInfo"/> objects in the list to
    /// a <see cref="TStrings"/> object.</summary>
    /// <param name="Dest">
    /// The destination <see cref="TStrings"/> object.</param>
    /// <param name="DisplayName">
    /// Determines which <see cref="TCultureInfo.DisplayNames"/> value should be used.</param>
    /// <seealso cref="AssignTo"/>
    /// <seealso cref="TCultureInfo.DisplayNames"/>
{$ENDREGION}
    Procedure AssignLabelsTo(Dest: TStrings; DisplayName: TCultureDisplayName);
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TCultureInfo"/> object in the list by its locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// The <see cref="TCultureInfo"/> object with the specified locale identifier
    /// or <see langword="nil"/> if the object is not found.</returns>
    /// <seealso cref="FindNearest"/>
{$ENDREGION}
    Function Find(LocaleID: LCID): TCultureInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TCultureInfo"/> object in the list by its locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// The <see cref="TCultureInfo"/> object with the specified locale name
    /// or <see langword="nil"/> if the object is not found.</returns>
    /// <seealso cref="FindNearest"/>
{$ENDREGION}
    Function Find(Const Locale: String): TCultureInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TCultureInfo"/> object in the list, which is either
    /// identical to or a dialect of a specified <see cref="TCultureInfo"/> object.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to be located.</param>
    /// <returns>
    /// The matched <see cref="TCultureInfo"/> object or <see langword="nil"/> if
    /// the object is not found.</returns>
    /// <seealso cref="Find"/>
{$ENDREGION}
    Function FindNearest(Culture: TCultureInfo): TCultureInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TCultureInfo"/> object in the list, which is either
    /// identical to or a dialect of a specified locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// The matched <see cref="TCultureInfo"/> object or <see langword="nil"/> if
    /// the object is not found.</returns>
    /// <seealso cref="Find"/>
{$ENDREGION}
    Function FindNearest(LocaleID: LCID): TCultureInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TCultureInfo"/> object in the list, which is either
    /// identical to or a dialect of a specified locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// The matched <see cref="TCultureInfo"/> object or <see langword="nil"/> if
    /// the object is not found.</returns>
    /// <seealso cref="Find"/>
{$ENDREGION}
    Function FindNearest(Const Locale: String): TCultureInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a specified <see cref="TCultureInfo"/> object in the list.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to locate.</param>
    /// <returns>
    /// The index of <see cref="TCultureInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
    /// <seealso cref="NearestIndexOf"/>
{$ENDREGION}
    Function IndexOf(Culture: TCultureInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCultureInfo"/> object in the list by its locale
    /// identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// The index of <see cref="TCultureInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
    /// <seealso cref="NearestIndexOf"/>
{$ENDREGION}
    Function IndexOf(LocaleID: LCID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCultureInfo"/> object in the list by its locale
    /// name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// The index of <see cref="TCultureInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
    /// <seealso cref="NearestIndexOf"/>
{$ENDREGION}
    Function IndexOf(Const Locale: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCultureInfo"/> object in the list by its display name.</summary>
    /// <param name="AName">
    /// The display name of the target object.</param>
    /// <param name="DisplayName">
    /// Indicates which <see cref="TCultureInfo.DisplayNames"/> value should be used.</param>
    /// <returns>
    /// The index of <see cref="TCultureInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOf"/>
{$ENDREGION}
    Function IndexOfName(Const AName: String; DisplayName: TCultureDisplayName): Integer;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCultureInfo"/> object in the list, which is either
    /// identical to or a dialect of a specified <see cref="TCultureInfo"/> object.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to locate.</param>
    /// <returns>
    /// The index of the matched <see cref="TCultureInfo"/> object or -1 if the
    /// object is not found.</returns>
    /// <seealso cref="IndexOf"/>
{$ENDREGION}
    Function NearestIndexOf(Culture: TCultureInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCultureInfo"/> object in the list, which is either
    /// identical to or a dialect of a specified locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// The index of the matched <see cref="TCultureInfo"/> object or -1 if the
    /// object is not found.</returns>
    /// <seealso cref="IndexOf"/>
{$ENDREGION}
    Function NearestIndexOf(LocaleID: LCID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCultureInfo"/> object in the list, which is either
    /// identical to or a dialect of a specified locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// The index of the matched <see cref="TCultureInfo"/> object or -1 if the
    /// object is not found.</returns>
    /// <seealso cref="IndexOf"/>
{$ENDREGION}
    Function NearestIndexOf(Const Locale: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a specified <see cref="TCultureInfo"/> object is in the list.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to examine.</param>
    /// <returns>
    /// Returns <see langword="true"/> if the <see cref="TCultureInfo"/> object is in the
    /// list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(Culture: TCultureInfo): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a <see cref="TCultureInfo"/> object with a specified locale
    /// identifier is in the list.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// Returns <see langword="true"/> if a <see cref="TCultureInfo"/> object with the
    /// specified locale identifier is in the list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(LocaleID: LCID): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a <see cref="TCultureInfo"/> object with a specified locale
    /// name is in the list.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// Returns <see langword="true"/> if a <see cref="TCultureInfo"/> object with the
    /// specified locale identifier is in the list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(Const Locale: String): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the number of <see cref="TCultureInfo"/> objects in the list.</summary>
{$ENDREGION}
    Property Count: Integer Read GetCount;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists the <see cref="TCultureInfo"/> objects in the list.</summary>
{$ENDREGION}
    Property Items[Index: Integer]: TCultureInfo Read GetItems; Default;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This class maintains a list of <see cref="TCultureInfo"/> objects.</summary>
  /// <remarks>
  /// Use TCultureList to store and maintain a list of <see cref="TCultureInfo"/>
  /// objects. TCultureList provides properties and methods to add, delete, rearrange,
  /// locate, access, and sort <see cref="TCultureInfo"/> objects.</remarks>
{$ENDREGION}

  TCultureList = Class(TReadonlyCultureList)
  Private
    FDuplicates: TDuplicates;
    FUpdateCount: Integer;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    Function GetCapacity: Integer; Inline;
    Procedure SetCapacity(Value: Integer);
    Procedure ReadLocales(Reader: TReader);
    Procedure WriteLocales(Writer: TWriter);
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Reads and writes the <see cref="Items"/> property as if it were published.</summary>
    /// <param name="Filer">
    /// The current <see cref="TReader"/> or <see cref="TWriter"/> object that is
    /// loading or saving the list.</param>
{$ENDREGION}
    Procedure DefineProperties(Filer: TFiler); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Generates an <see cref="OnChanging"/> event.</summary>
    /// <seealso cref="Change"/>
{$ENDREGION}
    Procedure Changing; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Generates an <see cref="OnChange"/> event.</summary>
    /// <seealso cref="Changing"/>
{$ENDREGION}
    Procedure Change; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the update level.</summary>
{$ENDREGION}
    Property UpdateCount: Integer Read FUpdateCount;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Prevents generating of the <see cref="OnChanging"/> and <see cref="OnChange"/>
    /// events until the <see cref="EndUpdate"/> method is called.</summary>
    /// <seealso cref="EndUpdate"/>
{$ENDREGION}
    Procedure BeginUpdate;
{$REGION 'xmldoc'}
    /// <summary>
    /// Reenables <see cref="OnChanging"/> and <see cref="OnChange"/> events generation
    /// that was turned off with the <see cref="BeginUpdate"/> method.</summary>
    /// <seealso cref="BeginUpdate"/>
{$ENDREGION}
    Procedure EndUpdate;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes all items from the list.</summary>
{$ENDREGION}
    Procedure Clear;
{$REGION 'xmldoc'}
    /// <summary>
    /// Sorts the items in the list by a specified display name.</summary>
    /// <param name="OrderBy">
    /// Determines which <see cref="TCultureInfo.DisplayNames"/> value should the
    /// list be sorted by.</param>
{$ENDREGION}
    Procedure Sort(OrderBy: TCultureDisplayName);
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies the list of <see cref="TCultureInfo"/> objects from another object.</summary>
    /// <param name="Source">
    /// The source list.</param>
    /// <seealso cref="Apply"/>
{$ENDREGION}
    Procedure Assign(Source: TPersistent); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies <see cref="TCultureInfo"/> objects of one list to another.</summary>
    /// <remarks>
    /// Call Apply to assign the elements of another list to this one. Assign combines
    /// the source list with this one using the logical operator specified by the
    /// <paramref name="Op"/> parameter.
    ///
    /// If the <paramref name="ListB"/> parameter is specified or not <see langword="nil"/>,
    /// then Apply first replaces all the elements of this list with those in <paramref name="ListA"/>,
    /// and then merges <paramref name="ListB"/> into this list using the operator specified
    /// by <paramref name="Op"/>.
    ///
    /// These operators are acceptable in Apply method:
    ///
    /// <list type="table">
    /// <listheader>
    /// <term>Operator</term>
    /// <description>Description</description>
    /// </listheader>
    /// <item>
    /// <term>laCopy</term>
    /// <description>Destination becomes a copy of the source</description>
    /// </item>
    /// <item>
    /// <term>laAnd</term>
    /// <description>Intersection of the two lists</description>
    /// </item>
    /// <item>
    /// <term>laOr</term>
    /// <description>Union of the two lists</description>
    /// </item>
    /// <item>
    /// <term>laXor</term>
    /// <description>Only those not in both lists</description>
    /// </item>
    /// <item>
    /// <term>laSrcUnique</term>
    /// <description>Only those unique to source (same as laAnd followed by laXor)</description>
    /// </item>
    /// <item>
    /// <term>laDestUnique</term>
    /// <description>Only those unique to destination (same as laOr followed by laXor)</description>
    /// </item>
    /// </list>
    ///
    /// </remarks>
    /// <param name="ListA">
    /// The source list.</param>
    /// <param name="Op">
    /// The operator that indicates how the two lists should be merged.</param>
    /// <param name="ListB">
    /// The optional and aditional source list.</param>
    /// <seealso cref="Assign"/>
{$ENDREGION}
    Procedure Apply(ListA: TReadonlyCultureList; Op: TListAssignOp = LaCopy; ListB: TReadonlyCultureList = Nil);
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a specified <see cref="TCultureInfo"/> object to the list.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to add.</param>
    /// <returns>
    /// The index of new <see cref="TCultureInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(Culture: TCultureInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a <see cref="TCultureInfo"/> object to the list by its locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// The index of new <see cref="TCultureInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(LocaleID: LCID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a <see cref="TCultureInfo"/> object to the list by its locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// The index of new <see cref="TCultureInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(Const Locale: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a specified <see cref="TCultureInfo"/> object from the list.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object to remove.</param>
    /// <returns>
    /// The index of the removed <see cref="TCultureInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(Culture: TCultureInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TCultureInfo"/> object from the list by its locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <returns>
    /// The index of the removed <see cref="TCultureInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(LocaleID: LCID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TCultureInfo"/> object from the list by its locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <returns>
    /// The index of the removed <see cref="TCultureInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(Const Locale: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TCultureInfo"/> object at a specified index of the list.</summary>
    /// <param name="Index">
    /// The index of <see cref="TCultureInfo"/> object to removed.</param>
    /// <seealso cref="Remove"/>
{$ENDREGION}
    Procedure Delete(Index: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets the number of <see cref="TCultureInfo"/> objects the TCultureList object can hold.</summary>
{$ENDREGION}
    Property Capacity: Integer Read GetCapacity Write SetCapacity;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets whether duplicate <see cref="TCultureInfo"/> objects can be added to the list.</summary>
{$ENDREGION}
    Property Duplicates: TDuplicates Read FDuplicates Write FDuplicates;
{$REGION 'xmldoc'}
    /// <summary>
    /// Occurs immediately after the list of <see cref="TCultureInfo"/> objects changes.</summary>
{$ENDREGION}
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
{$REGION 'xmldoc'}
    /// <summary>
    /// Occurs immediately before the list of <see cref="TCultureInfo"/> objects changes.</summary>
{$ENDREGION}
    Property OnChanging: TNotifyEvent Read FOnChanging Write FOnChanging;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the possible display names for
  /// <see cref="TTerritoryInfo"/> objects.</summary>
{$ENDREGION}

  TTerritoryDisplayName = (
{$REGION 'xmldoc'}
    /// Nothing
{$ENDREGION}
    TnNone,
{$REGION 'xmldoc'}
    /// Two characters of the international country code specified in ISO 3166-2.
{$ENDREGION}
    TnCode2,
{$REGION 'xmldoc'}
    /// Three characters of the international country code specified in ISO 3166-3.
{$ENDREGION}
    TnCode3,
{$REGION 'xmldoc'}
    /// The friendly name of the nation, for example, Germany.
{$ENDREGION}
    TnFriendlyName,
{$REGION 'xmldoc'}
    /// The official name of the nation, for example, Federal Republic of Germany.
{$ENDREGION}
    TnOfficialName,
{$REGION 'xmldoc'}
    /// Full localized (in Windows user's language) name of the country/region, for
    /// example, Deutschland for Germany.
{$ENDREGION}
    TnLocalizedName,
{$REGION 'xmldoc'}
    /// English name of the country/region, for example, Germany for Deutschland.
{$ENDREGION}
    TnEnglishName,
{$REGION 'xmldoc'}
    /// Native name of the country/region, for example, Deutschland for Germany.
{$ENDREGION}
    TnNativeName);

{$REGION 'xmldoc'}
  /// <summary>
  /// This class represents a country/region.</summary>
  /// <remarks>
  /// Each instance of TTerritoryInfo class represents a specific geographical location,
  /// and has properties and methods to query information about it.
  ///
  /// NOTE: Do not create or destroy an instance of TTerritoryInfo class. Use <see cref="TerritoryOf"/>
  /// global function to get TTerritoryInfo instance of a geographical location.</remarks>
  /// <seealso cref="TerritoryOf"/>
  /// <seealso cref="World.Territories"/>
{$ENDREGION}

  TTerritoryInfo = Class Sealed(TObject)
  Private
    FGeoID: GEOID;
    FCode2: String;
    FCode3: String;
    FLatitude: Double;
    FLongitude: Double;
    FFriendlyName: String;
    FOfficialName: String;
    FLocalizedName: String;
    FEnglishName: String;
    FCurrencies: TCurrencyList;
    FCultures: TCultureList;
    Function GetNativeName: String;
    Function GetCultures: TReadonlyCultureList; Inline;
    Function GetCurrencies: TReadonlyCurrencyList; Inline;
    Function GetDisplayNames(DisplayName: TTerritoryDisplayName): String;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Associates the specified <see cref="TCultureInfo"/> and <see cref="TCurrencyInfo"/>
    /// objects to this TTerritoryInfo object.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object.</param>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object.</param>
{$ENDREGION}
    Procedure Associate(Culture: TCultureInfo; Currency: TCurrencyInfo);
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the object for a specified geographical identifier.</summary>
    /// <param name="AGeoID">
    /// The geographical identifier.</param>
    /// <seealso cref="CreateByLocaleID"/>
    /// <seealso cref="CreateByLocale"/>
{$ENDREGION}
    Constructor Create(AGeoID: GEOID);
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the object for a specified locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
    /// <param name="Dummy">
    /// This parameter is for compatibility with c++.</param>
    /// <seealso cref="Create"/>
    /// <seealso cref="CreateByLocale"/>
{$ENDREGION}
    Constructor CreateByLocaleID(LocaleID: LCID; Dummy: Integer = 0);
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the object for a specified locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
    /// <seealso cref="Create"/>
    /// <seealso cref="CreateByLocaleID"/>
{$ENDREGION}
    Constructor CreateByLocale(Const Locale: String);
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroyes the object and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the geographical identifier of the country/region.</summary>
{$ENDREGION}
    Property GeoID: GEOID Read FGeoID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the two-character ISO 3166-2 international country code.</summary>
{$ENDREGION}
    Property Code2: String Read FCode2;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the three-character ISO 3166-3 international country code.</summary>
{$ENDREGION}
    Property Code3: String Read FCode3;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the geographical latitude of the territory's captital, in degrees.</summary>
{$ENDREGION}
    Property Latitude: Double Read FLatitude;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the geographical longitude of the territory's captital, in degrees.</summary>
{$ENDREGION}
    Property Longitude: Double Read FLongitude;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the friendly name of the nation.</summary>
{$ENDREGION}
    Property FriendlyName: String Read FFriendlyName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the official name of the nation.</summary>
{$ENDREGION}
    Property OfficialName: String Read FOfficialName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the localized (in Windows user's language) name of the country/region.</summary>
{$ENDREGION}
    Property LocalizedName: String Read FLocalizedName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the English name of the country/region.</summary>
{$ENDREGION}
    Property EnglishName: String Read FEnglishName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the Native name of the country/region.</summary>
{$ENDREGION}
    Property NativeName: String Read GetNativeName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists different cultures in this country/region.</summary>
{$ENDREGION}
    Property Cultures: TReadonlyCultureList Read GetCultures;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists authoritative currencies in this country/region.</summary>
{$ENDREGION}
    Property Currencies: TReadonlyCurrencyList Read GetCurrencies;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists display names of the country/region.</summary>
{$ENDREGION}
    Property DisplayNames[DisplayName: TTerritoryDisplayName]: String Read GetDisplayNames;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TTerritoryEnumerator enumerates through the <see cref="TTerritoryInfo"/> objects
  /// of an instance of <see cref="TReadonlyTerritoryList"/> or its descendent class.</summary>
{$ENDREGION}

  TTerritoryEnumerator = Class(TListEnumerator)
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for a specified list.</summary>
    /// <param name="list">
    /// The list to enumerate its items.</param>
{$ENDREGION}
    Constructor Create(List: TReadonlyTerritoryList);
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the current <see cref="TTerritoryInfo"/> object of the list being
    /// enumerated.</summary>
    /// <returns>
    /// Returns the current <see cref="TTerritoryInfo"/> object.</returns>
{$ENDREGION}
    Function GetCurrent: TTerritoryInfo; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the current <see cref="TTerritoryInfo"/> object.</summary>
{$ENDREGION}
    Property Current: TTerritoryInfo Read GetCurrent;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This class maintains a read-only list of <see cref="TTerritoryInfo"/> objects.</summary>
  /// <remarks>
  /// TReadonlyTerritoryList maintains a read-only list of <see cref="TTerritoryInfo"/>
  /// objects, and provides properties and methods to locate and access the objects.
  /// The objects in the list are hashed and can be located quickly.</remarks>
  /// <see cref="TTerritoryList"/>
{$ENDREGION}

  TReadonlyTerritoryList = Class(TPersistent)
  Private
    LookupTable_GeoID: TKeyLookup<GEOID, Integer>;
    LookupTable_Code2: TKeyLookup<String, Integer>;
    LookupTable_Code3: TKeyLookup<String, Integer>;
    IsLookupTableValid_GeoID: Boolean;
    IsLookupTableValid_Code2: Boolean;
    IsLookupTableValid_Code3: Boolean;
    Function GetCount: Integer; Inline;
    Function GetItems(Index: Integer): TTerritoryInfo; Inline;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Provides direct accress to the individual <see cref="TTerritoryInfo"/> objects.</summary>
{$ENDREGION}
    List: TList;
{$REGION 'xmldoc'}
    /// <summary>
    /// Attaches a specified <see cref="TTerritoryInfo"/> object with a specified index
    /// in the list.
    ///
    /// NOTE: This method does not add the object to the list, it only creats its lookup
    /// entries for all the lookup tables.</summary>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object to add to the lookup tables.</param>
    /// <param name="Index">
    /// The index of the <see cref="TTerritoryInfo"/> object in the list.</param>
    /// <seealso cref="RemoveLookupEntries"/>
{$ENDREGION}
    Procedure AddLookupEntries(Territory: TTerritoryInfo; Index: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Detaches a specified <see cref="TTerritoryInfo"/> object from the previosuly attached
    /// index.
    ///
    /// NOTE: This method does not remove the object from the list, it only deletes its
    /// lookup entries from all lookup tables.</summary>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object to remove from the lookup tables.</param>
    /// <seealso cref="AddLookupEntries"/>
{$ENDREGION}
    Procedure RemoveLookupEntries(Territory: TTerritoryInfo);
{$REGION 'xmldoc'}
    /// <summary>
    /// Clears all entries from all lookup tables.</summary>
{$ENDREGION}
    Procedure ClearLookupTables;
{$REGION 'xmldoc'}
    /// <summary>
    /// Schedules rebuilding of all lookup tables.</summary>
{$ENDREGION}
    Procedure InvalidateLookupTables; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Validates lookup table for <see cref="TTerritoryInfo.GeoID"/> values.</summary>
{$ENDREGION}
    Procedure ValidateLookupTable_GeoID;
{$REGION 'xmldoc'}
    /// <summary>
    /// Validates lookup table for <see cref="TTerritoryInfo.Code2"/> values.</summary>
{$ENDREGION}
    Procedure ValidateLookupTable_Code2;
{$REGION 'xmldoc'}
    /// <summary>
    /// Validates lookup table for <see cref="TTerritoryInfo.Code3"/> values.</summary>
{$ENDREGION}
    Procedure ValidateTookupTable_Code3;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creats an instance of the class.</summary>
{$ENDREGION}
    Constructor Create; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a <see cref="TTerritoryEnumerator"/> reference, which enumerates
    /// all <see cref="TTerritoryInfo"/> objects in the list.</summary>
    /// <returns>
    /// Returns an instance of <see cref="TTerritoryEnumerator"/> class.</returns>
{$ENDREGION}
    Function GetEnumerator: TTerritoryEnumerator;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the first <see cref="TTerritoryInfo"/> object in the list.</summary>
    /// <returns>
    /// The first <see cref="TTerritoryInfo"/> object in the list.</returns>
    /// <seealso cref="Last"/>
{$ENDREGION}
    Function First: TTerritoryInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the last <see cref="TTerritoryInfo"/> object in the list.</summary>
    /// <returns>
    /// The last <see cref="TTerritoryInfo"/> object in the list.</returns>
    /// <seealso cref="First"/>
{$ENDREGION}
    Function Last: TTerritoryInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies <see cref="TTerritoryInfo"/> objects in the list to another object.</summary>
    /// <param name="Dest">
    /// The destination object.</param>
    /// <seealso cref="AssignLabelsTo"/>
{$ENDREGION}
    Procedure AssignTo(Dest: TPersistent); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies display name of <see cref="TTerritoryInfo"/> objects in the list to
    /// a <see cref="TStrings"/> object.</summary>
    /// <param name="Dest">
    /// The destination <see cref="TStrings"/> object.</param>
    /// <param name="DisplayName">
    /// Determines which <see cref="TTerritoryInfo.DisplayNames"/> value should be used.</param>
    /// <seealso cref="AssignTo"/>
    /// <seealso cref="TTerritoryInfo.DisplayNames"/>
{$ENDREGION}
    Procedure AssignLabelsTo(Dest: TStrings; DisplayName: TTerritoryDisplayName);
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TTerritoryInfo"/> object in the list by its geographical identifier.</summary>
    /// <param name="GeoID">
    /// The geographical identifier.</param>
    /// <returns>
    /// The <see cref="TTerritoryInfo"/> object with the specified geographical identifier
    /// or <see langword="nil"/> if the object is not found.</returns>
{$ENDREGION}
    Function Find(GeoID: GEOID): TTerritoryInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TTerritoryInfo"/> object in the list by its two or three
    /// characters international country/region code.</summary>
    /// <param name="Code">
    /// The two or three characters international code.</param>
    /// <returns>
    /// The <see cref="TTerritoryInfo"/> object with the specified international code
    /// or <see langword="nil"/> if the object is not found.</returns>
{$ENDREGION}
    Function Find(Const Code: String): TTerritoryInfo; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a specified <see cref="TTerritoryInfo"/> object in the list.</summary>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object to locate.</param>
    /// <returns>
    /// The index of <see cref="TTerritoryInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
{$ENDREGION}
    Function IndexOf(Territory: TTerritoryInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TTerritoryInfo"/> object in the list by its geographical
    /// identifier.</summary>
    /// <param name="GeoID">
    /// The geographical identifier.</param>
    /// <returns>
    /// The index of <see cref="TTerritoryInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
{$ENDREGION}
    Function IndexOf(GeoID: GEOID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TTerritoryInfo"/> object in the list by its two or three
    /// characters international code.</summary>
    /// <param name="Code">
    /// The two or three characters international code.</param>
    /// <returns>
    /// The index of <see cref="TTerritoryInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
{$ENDREGION}
    Function IndexOf(Const Code: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TTerritoryInfo"/> object in the list by its display name.</summary>
    /// <param name="AName">
    /// The display name of the target object.</param>
    /// <param name="DisplayName">
    /// Indicates which <see cref="TTerritoryInfo.DisplayNames"/> value should be used.</param>
    /// <returns>
    /// The index of <see cref="TTerritoryInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOf"/>
{$ENDREGION}
    Function IndexOfName(Const AName: String; DisplayName: TTerritoryDisplayName): Integer;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a specified <see cref="TTerritoryInfo"/> object is in the list.</summary>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object to examine.</param>
    /// <returns>
    /// Returns <see langword="true"/> if the <see cref="TTerritoryInfo"/> object is in the
    /// list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(Territory: TTerritoryInfo): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a <see cref="TTerritoryInfo"/> object with a specified geograpgical
    /// identifier is in the list.</summary>
    /// <param name="GeoID">
    /// The geographical identifier.</param>
    /// <returns>
    /// Returns <see langword="true"/> if a <see cref="TTerritoryInfo"/> object with the
    /// specified geographical identifier is in the list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(GeoID: GEOID): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a <see cref="TTerritoryInfo"/> object with a specified two or three
    /// charachters international code is in the list.</summary>
    /// <param name="Code">
    /// The two or three characters international code.</param>
    /// <returns>
    /// Returns <see langword="true"/> if a <see cref="TTerritoryInfo"/> object with the
    /// specified international code is in the list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(Const Code: String): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the number of <see cref="TTerritoryInfo"/> objects in the list.</summary>
{$ENDREGION}
    Property Count: Integer Read GetCount;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists the <see cref="TTerritoryInfo"/> objects in the list.</summary>
{$ENDREGION}
    Property Items[Index: Integer]: TTerritoryInfo Read GetItems; Default;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This class maintains a list of <see cref="TTerritoryInfo"/> objects.</summary>
  /// <remarks>
  /// Use TTerritoryList to store and maintain a list of <see cref="TTerritoryInfo"/>
  /// objects. TTerritoryList provides properties and methods to add, delete, rearrange,
  /// locate, access, and sort <see cref="TTerritoryInfo"/> objects.</remarks>
{$ENDREGION}

  TTerritoryList = Class(TReadonlyTerritoryList)
  Private
    FDuplicates: TDuplicates;
    FUpdateCount: Integer;
    FOnChanging: TNotifyEvent;
    FOnChange: TNotifyEvent;
    Function GetCapacity: Integer; Inline;
    Procedure SetCapacity(Value: Integer);
    Procedure ReadTerritory(Reader: TReader);
    Procedure WriteTerritory(Writer: TWriter);
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Reads and writes the <see cref="Items"/> property as if it were published.</summary>
    /// <param name="Filer">
    /// The current <see cref="TReader"/> or <see cref="TWriter"/> object that is
    /// loading or saving the list.</param>
{$ENDREGION}
    Procedure DefineProperties(Filer: TFiler); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Generates an <see cref="OnChanging"/> event.</summary>
    /// <seealso cref="Change"/>
{$ENDREGION}
    Procedure Changing; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Generates an <see cref="OnChange"/> event.</summary>
    /// <seealso cref="Changing"/>
{$ENDREGION}
    Procedure Change; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the update level.</summary>
{$ENDREGION}
    Property UpdateCount: Integer Read FUpdateCount;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Prevents generating of the <see cref="OnChanging"/> and <see cref="OnChange"/>
    /// events until the <see cref="EndUpdate"/> method is called.</summary>
    /// <seealso cref="EndUpdate"/>
{$ENDREGION}
    Procedure BeginUpdate;
{$REGION 'xmldoc'}
    /// <summary>
    /// Reenables <see cref="OnChanging"/> and <see cref="OnChange"/> events generation
    /// that was turned off with the <see cref="BeginUpdate"/> method.</summary>
    /// <seealso cref="BeginUpdate"/>
{$ENDREGION}
    Procedure EndUpdate;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes all items from the list.</summary>
{$ENDREGION}
    Procedure Clear;
{$REGION 'xmldoc'}
    /// <summary>
    /// Sorts the items in the list by a specified display name.</summary>
    /// <param name="OrderBy">
    /// Determines which <see cref="TTerritoryInfo.DisplayNames"/> value should the
    /// list be sorted by.</param>
{$ENDREGION}
    Procedure Sort(OrderBy: TTerritoryDisplayName);
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies the list of <see cref="TTerritoryInfo"/> objects from another object.</summary>
    /// <param name="Source">
    /// The source list.</param>
    /// <seealso cref="Apply"/>
{$ENDREGION}
    Procedure Assign(Source: TPersistent); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies <see cref="TTerritoryInfo"/> objects of one list to another.</summary>
    /// <remarks>
    /// Call Apply to assign the elements of another list to this one. Assign combines
    /// the source list with this one using the logical operator specified by the
    /// <paramref name="Op"/> parameter.
    ///
    /// If the <paramref name="ListB"/> parameter is specified or not <see langword="nil"/>,
    /// then Apply first replaces all the elements of this list with those in <paramref name="ListA"/>,
    /// and then merges <paramref name="ListB"/> into this list using the operator specified
    /// by <paramref name="Op"/>.
    ///
    /// These operators are acceptable in Apply method:
    ///
    /// <list type="table">
    /// <listheader>
    /// <term>Operator</term>
    /// <description>Description</description>
    /// </listheader>
    /// <item>
    /// <term>laCopy</term>
    /// <description>Destination becomes a copy of the source</description>
    /// </item>
    /// <item>
    /// <term>laAnd</term>
    /// <description>Intersection of the two lists</description>
    /// </item>
    /// <item>
    /// <term>laOr</term>
    /// <description>Union of the two lists</description>
    /// </item>
    /// <item>
    /// <term>laXor</term>
    /// <description>Only those not in both lists</description>
    /// </item>
    /// <item>
    /// <term>laSrcUnique</term>
    /// <description>Only those unique to source (same as laAnd followed by laXor)</description>
    /// </item>
    /// <item>
    /// <term>laDestUnique</term>
    /// <description>Only those unique to destination (same as laOr followed by laXor)</description>
    /// </item>
    /// </list>
    ///
    /// </remarks>
    /// <param name="ListA">
    /// The source list.</param>
    /// <param name="Op">
    /// The operator that indicates how the two lists should be merged.</param>
    /// <param name="ListB">
    /// The optional and aditional source list.</param>
    /// <seealso cref="Assign"/>
{$ENDREGION}
    Procedure Apply(ListA: TReadonlyTerritoryList; Op: TListAssignOp = LaCopy;
      ListB: TReadonlyTerritoryList = Nil);
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a specified <see cref="TTerritoryInfo"/> object to the list.</summary>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object to add.</param>
    /// <returns>
    /// The index of new <see cref="TTerritoryInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(Territory: TTerritoryInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a <see cref="TTerritoryInfo"/> object to the list by its geographical identifier.</summary>
    /// <param name="GeoID">
    /// The geographical identifier.</param>
    /// <returns>
    /// The index of new <see cref="TTerritoryInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(GeoID: GEOID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a <see cref="TTerritoryInfo"/> object to the list by its two or three
    /// character international code.</summary>
    /// <param name="Code">
    /// The two or three character international code.</param>
    /// <returns>
    /// The index of new <see cref="TTerritoryInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(Const Code: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a specified <see cref="TTerritoryInfo"/> object from the list.</summary>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object to remove.</param>
    /// <returns>
    /// The index of the removed <see cref="TTerritoryInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(Territory: TTerritoryInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TTerritoryInfo"/> object from the list by its geographical
    /// identifier.</summary>
    /// <param name="GeoID">
    /// The geographical identifier.</param>
    /// <returns>
    /// The index of the removed <see cref="TTerritoryInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(GeoID: GEOID): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TTerritoryInfo"/> object from the list by its two or three
    /// characters international code.</summary>
    /// <param name="Code">
    /// The two or three character sinternational code.</param>
    /// <returns>
    /// The index of the removed <see cref="TTerritoryInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(Const Code: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TTerritoryInfo"/> object at a specified index of the list.</summary>
    /// <param name="Index">
    /// The index of <see cref="TTerritoryInfo"/> object to remove.</param>
    /// <seealso cref="Remove"/>
{$ENDREGION}
    Procedure Delete(Index: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets the number of <see cref="TTerritoryInfo"/> objects the TTerritoryList object can hold.</summary>
{$ENDREGION}
    Property Capacity: Integer Read GetCapacity Write SetCapacity;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets whether duplicate <see cref="TTerritoryInfo"/> objects can be added to the list.</summary>
{$ENDREGION}
    Property Duplicates: TDuplicates Read FDuplicates Write FDuplicates;
{$REGION 'xmldoc'}
    /// <summary>
    /// Occurs immediately after the list of <see cref="TTerritoryInfo"/> objects changes.</summary>
{$ENDREGION}
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
{$REGION 'xmldoc'}
    /// <summary>
    /// Occurs immediately before the list of <see cref="TTerritoryInfo"/> objects changes.</summary>
{$ENDREGION}
    Property OnChanging: TNotifyEvent Read FOnChanging Write FOnChanging;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This enumeration type identifies the possible display names for
  /// <see cref="TCurrencyInfo"/> objects.</summary>
{$ENDREGION}

  TCurrencyDisplayName = (
{$REGION 'xmldoc'}
    /// Nothing
{$ENDREGION}
    CrnNone,
{$REGION 'xmldoc'}
    /// Three characters of the international monetary symbol specified in ISO 4217.
{$ENDREGION}
    CrnIntlSymbol,
{$REGION 'xmldoc'}
    /// The full English name of the currency.
{$ENDREGION}
    CrnEnglishName,
{$REGION 'xmldoc'}
    /// The native name of the currency.
{$ENDREGION}
    CrnNativeName);

{$REGION 'xmldoc'}
  /// <summary>
  /// This class represents a currency.</summary>
  /// <remarks>
  /// Each instance of TCurrencyInfo class represents a specific currency and has properties
  /// and methods to query information about it.
  ///
  /// NOTE: Do not create or destroy an instance of TCurrencyInfo class. Use <see cref="CurrencyOf"/>
  /// global function to get TCurrencyInfo instance of a currency.</remarks>
  /// <seealso cref="CurrencyOf"/>
  /// <seealso cref="World.Currencies"/>
{$ENDREGION}

  TCurrencyInfo = Class Sealed(TObject)
  Private
    FIntlSymbol: String;
    FLocalSymbol: String;
    FEnglishName: String;
    FCultures: TCultureList;
    FCountries: TTerritoryList;
    FOriginCountry: TTerritoryInfo;
    Function GetNativeName: String; Inline;
    Function GetCultures: TReadonlyCultureList; Inline;
    Function GetCountries: TReadonlyTerritoryList; Inline;
    Function GetDisplayNames(DisplayName: TCurrencyDisplayName): String;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Associates the specified <see cref="TCultureInfo"/> and <see cref="TTerritoryInfo"/>
    /// objects to this TCurrencyInfo object.</summary>
    /// <param name="Culture">
    /// The <see cref="TCultureInfo"/> object.</param>
    /// <param name="Territory">
    /// The <see cref="TTerritoryInfo"/> object.</param>
{$ENDREGION}
    Procedure Associate(Culture: TCultureInfo; Territory: TTerritoryInfo);
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the object for a specified locale identifier.</summary>
    /// <param name="LocaleID">
    /// The locale identifier.</param>
{$ENDREGION}
    Constructor Create(LocaleID: LCID); Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the object for a specified locale name.</summary>
    /// <param name="Locale">
    /// The locale name.</param>
{$ENDREGION}
    Constructor Create(Const Locale: String); Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroyes the object and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Specifies the three-character ISO 4217 international monetary symbol of
    /// the currency.</summary>
{$ENDREGION}
    Property IntlSymbol: String Read FIntlSymbol;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the local monetary symbol of the currency.</summary>
{$ENDREGION}
    Property LocalSymbol: String Read FLocalSymbol;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the English name of the currency.</summary>
{$ENDREGION}
    Property EnglishName: String Read FEnglishName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the native name of the currency.</summary>
{$ENDREGION}
    Property NativeName: String Read GetNativeName;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists cultures that this currency is authoritative in.</summary>
{$ENDREGION}
    Property Cultures: TReadonlyCultureList Read GetCultures;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists countries/regions that this currency is authoritative in.</summary>
{$ENDREGION}
    Property Countries: TReadonlyTerritoryList Read GetCountries;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the main country/region of the currency. The value of this property
    /// can be <see langword="nil"/>. For example, Euro has no main country.</summary>
{$ENDREGION}
    Property OriginCountry: TTerritoryInfo Read FOriginCountry;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists display names of the currency.</summary>
{$ENDREGION}
    Property DisplayNames[DisplayName: TCurrencyDisplayName]: String Read GetDisplayNames;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// TCurrencyEnumerator enumerates through the <see cref="TCurrencyInfo"/> objects
  /// of an instance of <see cref="TReadonlyCurrencyList"/> or its descendent class.</summary>
{$ENDREGION}

  TCurrencyEnumerator = Class(TListEnumerator)
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates an instance of the class for a specified list.</summary>
    /// <param name="list">
    /// The list to enumerate its items.</param>
{$ENDREGION}
    Constructor Create(List: TReadonlyCurrencyList);
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the current <see cref="TCurrencyInfo"/> object of the list being
    /// enumerated.</summary>
    /// <returns>
    /// Returns the current <see cref="TCurrencyInfo"/> object.</returns>
{$ENDREGION}
    Function GetCurrent: TCurrencyInfo; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the current <see cref="TCurrencyInfo"/> object.</summary>
{$ENDREGION}
    Property Current: TCurrencyInfo Read GetCurrent;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This class maintains a read-only list of <see cref="TCurrencyInfo"/> objects.</summary>
  /// <remarks>
  /// TReadonlyCurrencyList maintains a read-only list of <see cref="TCurrencyInfo"/>
  /// objects, and provides properties and methods to locate and access the objects.
  /// The objects in the list are hashed and can be located quickly.</remarks>
  /// <see cref="TCurrencyList"/>
{$ENDREGION}

  TReadonlyCurrencyList = Class(TPersistent)
  Private
    LookupTable_IntlSymbol: TKeyLookup<String, Integer>;
    IsLookupTableValid_IntlSymbol: Boolean;
    Function GetCount: Integer; Inline;
    Function GetItems(Index: Integer): TCurrencyInfo; Inline;
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Provides direct accress to the individual <see cref="TCurrencyInfo"/> objects.</summary>
{$ENDREGION}
    List: TList;
{$REGION 'xmldoc'}
    /// <summary>
    /// Attaches a specified <see cref="TCurrencyInfo"/> object with a specified index
    /// in the list.
    ///
    /// NOTE: This method does not add the object to the list, it only creats its lookup
    /// entries for all the lookup tables.</summary>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object to add to the lookup tables.</param>
    /// <param name="Index">
    /// The index of the <see cref="TCurrencyInfo"/> object in the list.</param>
    /// <seealso cref="RemoveLookupEntries"/>
{$ENDREGION}
    Procedure AddLookupEntries(Currency: TCurrencyInfo; Index: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Detaches a specified <see cref="TCurrencyInfo"/> object from the previosuly
    /// attached index.
    ///
    /// NOTE: This method does not remove the object from the list, it only deletes its
    /// lookup entries from all lookup tables.</summary>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object to remove from the lookup tables.</param>
    /// <seealso cref="AddLookupEntries"/>
{$ENDREGION}
    Procedure RemoveLookupEntries(Currency: TCurrencyInfo);
{$REGION 'xmldoc'}
    /// <summary>
    /// Clears all entries from all lookup tables.</summary>
{$ENDREGION}
    Procedure ClearLookupTables;
{$REGION 'xmldoc'}
    /// <summary>
    /// Schedules rebuilding of all lookup tables.</summary>
{$ENDREGION}
    Procedure InvalidateLookupTables; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Validates lookup table for <see cref="TCurrencyInfo.IntlSymbol"/> values.</summary>
{$ENDREGION}
    Procedure ValidateLookupTable_IntlSymbol;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creats an instance of the class.</summary>
{$ENDREGION}
    Constructor Create; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Destroys the instance and releases its allocated memory.</summary>
{$ENDREGION}
    Destructor Destroy; Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns a <see cref="TCurrencyEnumerator"/> reference, which enumerates
    /// all <see cref="TCurrencyInfo"/> objects in the list.</summary>
    /// <returns>
    /// Returns an instance of <see cref="TCurrencyEnumerator"/> class.</returns>
{$ENDREGION}
    Function GetEnumerator: TCurrencyEnumerator;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the first <see cref="TCurrencyInfo"/> object in the list.</summary>
    /// <returns>
    /// The first <see cref="TCurrencyInfo"/> object in the list.</returns>
    /// <seealso cref="Last"/>
{$ENDREGION}
    Function First: TCurrencyInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the last <see cref="TCurrencyInfo"/> object in the list.</summary>
    /// <returns>
    /// The last <see cref="TCurrencyInfo"/> object in the list.</returns>
    /// <seealso cref="First"/>
{$ENDREGION}
    Function Last: TCurrencyInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies <see cref="TCurrencyInfo"/> objects in the list to another object.</summary>
    /// <param name="Dest">
    /// The destination object.</param>
    /// <seealso cref="AssignLabelsTo"/>
{$ENDREGION}
    Procedure AssignTo(Dest: TPersistent); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies display name of <see cref="TCurrencyInfo"/> objects in the list to
    /// a <see cref="TStrings"/> object.</summary>
    /// <param name="Dest">
    /// The destination <see cref="TStrings"/> object.</param>
    /// <param name="DisplayName">
    /// Determines which <see cref="TCurrencyInfo.DisplayNames"/> value should be used.</param>
    /// <seealso cref="AssignTo"/>
    /// <seealso cref="TCurrencyInfo.DisplayNames"/>
{$ENDREGION}
    Procedure AssignLabelsTo(Dest: TStrings; DisplayName: TCurrencyDisplayName);
{$REGION 'xmldoc'}
    /// <summary>
    /// Locates a <see cref="TCurrencyInfo"/> object in the list by its three characters
    /// international monetary symbol.</summary>
    /// <param name="IntlSymbol">
    /// The international monetary sumbol.</param>
    /// <returns>
    /// The <see cref="TCurrencyInfo"/> object with the specified international monetary
    /// sumbol or <see langword="nil"/> if the object is not found.</returns>
{$ENDREGION}
    Function Find(Const IntlSymbol: String): TCurrencyInfo;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a specified <see cref="TCurrencyInfo"/> object in the list.</summary>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object to locate.</param>
    /// <returns>
    /// The index of <see cref="TCurrencyInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
{$ENDREGION}
    Function IndexOf(Currency: TCurrencyInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCurrencyInfo"/> object in the list by its three
    /// characters international monetary symbol.</summary>
    /// <param name="IntlSymbol">
    /// The international monetary symbol.</param>
    /// <returns>
    /// The index of <see cref="TCurrencyInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOfName"/>
{$ENDREGION}
    Function IndexOf(Const IntlSymbol: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Finds index of a <see cref="TCurrencyInfo"/> object in the list by its display name.</summary>
    /// <param name="AName">
    /// The display name of the target object.</param>
    /// <param name="DisplayName">
    /// Indicates which <see cref="TCurrencyInfo.DisplayNames"/> value should be used.</param>
    /// <returns>
    /// The index of <see cref="TCurrencyInfo"/> object or -1 if the object is not found.</returns>
    /// <seealso cref="IndexOf"/>
{$ENDREGION}
    Function IndexOfName(Const AName: String; DisplayName: TCurrencyDisplayName): Integer;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a specified <see cref="TCurrencyInfo"/> object is in the list.</summary>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object to examine.</param>
    /// <returns>
    /// Returns <see langword="true"/> if the <see cref="TCurrencyInfo"/> object is in the
    /// list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(Currency: TCurrencyInfo): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Indicates whether a <see cref="TCurrencyInfo"/> object with a specified three characters
    /// international monetary symbol is in the list.</summary>
    /// <param name="IntlSymbol">
    /// The international monetary symbol.</param>
    /// <returns>
    /// Returns <see langword="true"/> if a <see cref="TCurrencyInfo"/> object with the
    /// specified international monetary symbol is in the list, otherwise returns <see langword="false"/>.</returns>
{$ENDREGION}
    Function Exists(Const IntlSymbol: String): Boolean; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the number of <see cref="TCurrencyInfo"/> objects in the list.</summary>
{$ENDREGION}
    Property Count: Integer Read GetCount;
{$REGION 'xmldoc'}
    /// <summary>
    /// Lists the <see cref="TCurrencyInfo"/> objects in the list.</summary>
{$ENDREGION}
    Property Items[Index: Integer]: TCurrencyInfo Read GetItems; Default;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// This class maintains a list of <see cref="TCurrencyInfo"/> objects.</summary>
  /// <remarks>
  /// Use TCurrencyList to store and maintain a list of <see cref="TCurrencyInfo"/>
  /// objects. TCurrencyList provides properties and methods to add, delete, rearrange,
  /// locate, access, and sort <see cref="TCurrencyInfo"/> objects.</remarks>
{$ENDREGION}

  TCurrencyList = Class(TReadonlyCurrencyList)
  Private
    FDuplicates: TDuplicates;
    FUpdateCount: Integer;
    FOnChanging: TNotifyEvent;
    FOnChange: TNotifyEvent;
    Function GetCapacity: Integer; Inline;
    Procedure SetCapacity(Value: Integer);
    Procedure ReadCurrencies(Reader: TReader);
    Procedure WriteCurrencies(Writer: TWriter);
  Protected
{$REGION 'xmldoc'}
    /// <summary>
    /// Reads and writes the <see cref="Items"/> property as if it were published.</summary>
    /// <param name="Filer">
    /// The current <see cref="TReader"/> or <see cref="TWriter"/> object that is
    /// loading or saving the list.</param>
{$ENDREGION}
    Procedure DefineProperties(Filer: TFiler); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Generates an <see cref="OnChanging"/> event.</summary>
    /// <seealso cref="Change"/>
{$ENDREGION}
    Procedure Changing; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Generates an <see cref="OnChange"/> event.</summary>
    /// <seealso cref="Changing"/>
{$ENDREGION}
    Procedure Change; Virtual;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the update level.</summary>
{$ENDREGION}
    Property UpdateCount: Integer Read FUpdateCount;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Prevents generating of the <see cref="OnChanging"/> and <see cref="OnChange"/>
    /// events until the <see cref="EndUpdate"/> method is called.</summary>
    /// <seealso cref="EndUpdate"/>
{$ENDREGION}
    Procedure BeginUpdate;
{$REGION 'xmldoc'}
    /// <summary>
    /// Reenables <see cref="OnChanging"/> and <see cref="OnChange"/> events generation
    /// that was turned off with the <see cref="BeginUpdate"/> method.</summary>
    /// <seealso cref="BeginUpdate"/>
{$ENDREGION}
    Procedure EndUpdate;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes all items from the list.</summary>
{$ENDREGION}
    Procedure Clear;
{$REGION 'xmldoc'}
    /// <summary>
    /// Sorts the items in the list by a specified display name.</summary>
    /// <param name="OrderBy">
    /// Determines which <see cref="TCurrencyInfo.DisplayNames"/> value should the
    /// list be sorted by.</param>
{$ENDREGION}
    Procedure Sort(OrderBy: TCurrencyDisplayName);
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies the list of <see cref="TCurrencyInfo"/> objects from another object.</summary>
    /// <param name="Source">
    /// The source list.</param>
    /// <seealso cref="Apply"/>
{$ENDREGION}
    Procedure Assign(Source: TPersistent); Override;
{$REGION 'xmldoc'}
    /// <summary>
    /// Copies <see cref="TCurrencyInfo"/> objects of one list to another.</summary>
    /// <remarks>
    /// Call Apply to assign the elements of another list to this one. Assign combines
    /// the source list with this one using the logical operator specified by the
    /// <paramref name="Op"/> parameter.
    ///
    /// If the <paramref name="ListB"/> parameter is specified or not <see langword="nil"/>,
    /// then Apply first replaces all the elements of this list with those in <paramref name="ListA"/>,
    /// and then merges <paramref name="ListB"/> into this list using the operator specified
    /// by <paramref name="Op"/>.
    ///
    /// These operators are acceptable in Apply method:
    ///
    /// <list type="table">
    /// <listheader>
    /// <term>Operator</term>
    /// <description>Description</description>
    /// </listheader>
    /// <item>
    /// <term>laCopy</term>
    /// <description>Destination becomes a copy of the source</description>
    /// </item>
    /// <item>
    /// <term>laAnd</term>
    /// <description>Intersection of the two lists</description>
    /// </item>
    /// <item>
    /// <term>laOr</term>
    /// <description>Union of the two lists</description>
    /// </item>
    /// <item>
    /// <term>laXor</term>
    /// <description>Only those not in both lists</description>
    /// </item>
    /// <item>
    /// <term>laSrcUnique</term>
    /// <description>Only those unique to source (same as laAnd followed by laXor)</description>
    /// </item>
    /// <item>
    /// <term>laDestUnique</term>
    /// <description>Only those unique to destination (same as laOr followed by laXor)</description>
    /// </item>
    /// </list>
    ///
    /// </remarks>
    /// <param name="ListA">
    /// The source list.</param>
    /// <param name="Op">
    /// The operator that indicates how the two lists should be merged.</param>
    /// <param name="ListB">
    /// The optional and aditional source list.</param>
    /// <seealso cref="Assign"/>
{$ENDREGION}
    Procedure Apply(ListA: TReadonlyCurrencyList; Op: TListAssignOp = LaCopy;
      ListB: TReadonlyCurrencyList = Nil);
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a specified <see cref="TCurrencyInfo"/> object to the list.</summary>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object to add.</param>
    /// <returns>
    /// The index of new <see cref="TCurrencyInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(Currency: TCurrencyInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a <see cref="TCurrencyInfo"/> object to the list by its three characters
    /// international monetary symbol.</summary>
    /// <param name="IntlSymbol">
    /// The international monetary symbol.</param>
    /// <returns>
    /// The index of new <see cref="TCurrencyInfo"/> object in the list.</returns>
{$ENDREGION}
    Function Add(Const IntlSymbol: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a specified <see cref="TCurrencyInfo"/> object from the list.</summary>
    /// <param name="Currency">
    /// The <see cref="TCurrencyInfo"/> object to remove.</param>
    /// <returns>
    /// The index of the removed <see cref="TCurrencyInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(Currency: TCurrencyInfo): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TCurrencyInfo"/> object from the list by its three characters
    /// international monetry symbol.</summary>
    /// <param name="IntlSymbol">
    /// The international monetry symbol.</param>
    /// <returns>
    /// The index of the removed <see cref="TCurrencyInfo"/> object, or -1 if the object
    /// is not found.</returns>
    /// <seealso cref="Delete"/>
{$ENDREGION}
    Function Remove(Const IntlSymbol: String): Integer; Overload;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a <see cref="TCurrencyInfo"/> object at a specified index of the list.</summary>
    /// <param name="Index">
    /// The index of <see cref="TCurrencyInfo"/> object to remove.</param>
    /// <seealso cref="Remove"/>
{$ENDREGION}
    Procedure Delete(Index: Integer);
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets the number of <see cref="TCurrencyInfo"/> objects the TCurrencyList object can hold.</summary>
{$ENDREGION}
    Property Capacity: Integer Read GetCapacity Write SetCapacity;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets or sets whether duplicate <see cref="TCurrencyInfo"/> objects can be added to the list.</summary>
{$ENDREGION}
    Property Duplicates: TDuplicates Read FDuplicates Write FDuplicates;
{$REGION 'xmldoc'}
    /// <summary>
    /// Occurs immediately after the list of <see cref="TCurrencyInfo"/> objects changes.</summary>
{$ENDREGION}
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
{$REGION 'xmldoc'}
    /// <summary>
    /// Occurs immediately before the list of <see cref="TCurrencyInfo"/> objects changes.</summary>
{$ENDREGION}
    Property OnChanging: TNotifyEvent Read FOnChanging Write FOnChanging;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// World class provides information about all cultures, territories, and
  /// currencies.</summary>
  /// <remarks>
  /// World class provides properties to locate and access all available
  /// <see cref="TCultureInfo"/>, <see cref="TTerritoryInfo"/>, and
  /// <see cref="TCurrencyInfo"/> objects.</remarks>
{$ENDREGION}

  World = Class
  Private
    Class Var Changing: Boolean;
    Class Var CultureList: TCultureList;
    Class Var TerritoryList: TTerritoryList;
    Class Var CurrencyList: TCurrencyList;
    Class Procedure Startup; Static;
    Class Procedure Cleanup; Static;
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the read-only list of all available <see cref="TCultureInfo"/> objects.</summary>
    /// <returns>
    /// The read-only list of all available <see cref="TCultureInfo"/> objects.</returns>
{$ENDREGION}
    Class Function Cultures: TReadonlyCultureList; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the read-only list of all available <see cref="TTerritoryInfo"/> objects.</summary>
    /// <returns>
    /// The read-only list of all available <see cref="TTerritoryInfo"/> objects.</returns>
{$ENDREGION}
    Class Function Territories: TReadonlyTerritoryList; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the read-only list of all available <see cref="TCurrencyInfo"/> objects.</summary>
    /// <returns>
    /// The read-only list of all available <see cref="TCurrencyInfo"/> objects.</returns>
{$ENDREGION}
    Class Function Currencies: TReadonlyCurrencyList; Static;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// Indicates whether the i18n package functionality is available for the current
  /// version of Windows.</summary>
  /// <remarks>
  /// The i18n package is only supported by Windows XP and newer versions of Windows.
  /// Use i18nAvailable to determine whether the current version of Windows can support
  /// the i18n package or not.</remarks>
  /// <returns>
  /// Returns <see langword="true"/> if the i18n package can be used, otherwise returns
  /// <see langword="false"/>.</returns>
{$ENDREGION}

Function I18nAvailable: Boolean; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCultureInfo"/> object thet represents the application's
/// default locale.</summary>
/// <returns>
/// The <see cref="TCultureInfo"/> object.</returns>
/// <seealso cref="GetSystemDefaultCulture"/>
/// <seealso cref="GetUserDefaultCulture"/>
/// <seealso cref="GetUserDefaultUICulture"/>
{$ENDREGION}
Function GetApplicationDefaultCulture: TCultureInfo;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCultureInfo"/> object thet represents the system
/// default locale.</summary>
/// <returns>
/// The <see cref="TCultureInfo"/> object.</returns>
/// <seealso cref="GetUserDefaultCulture"/>
/// <seealso cref="GetUserDefaultUICulture"/>
{$ENDREGION}
Function GetSystemDefaultCulture: TCultureInfo; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCultureInfo"/> object thet represents the user's
/// default locale. </summary>
/// <returns>
/// The <see cref="TCultureInfo"/> object.</returns>
/// <seealso cref="GetUserDefaultUICulture"/>
/// <seealso cref="GetSystemDefaultCulture"/>
{$ENDREGION}
Function GetUserDefaultCulture: TCultureInfo; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCultureInfo"/> object thet represents the user's
/// default user interface locale. </summary>
/// <returns>
/// The <see cref="TCultureInfo"/> object.</returns>
/// <seealso cref="GetUserDefaultCulture"/>
/// <seealso cref="GetSystemDefaultCulture"/>
{$ENDREGION}
Function GetUserDefaultUICulture: TCultureInfo;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TTerritoryInfo"/> object that represents the geographical
/// location of the user.</summary>
/// <returns>
/// The <see cref="TTerritoryInfo"/> object.</returns>
{$ENDREGION}
Function GetUserTerritory: TTerritoryInfo; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCultureInfo"/> object that represents a locale specified
/// by its identifier.</summary>
/// <remarks>
/// Use CultureOf to get information about a locale specified by its identifier.
///
/// When the argument passed as <paramref name="Exact"/> parameter is <see langword="true"/>,
/// the returned <see cref="TCultureInfo"/> object has the same locale of the locale
/// specified by the <paramref name="LocaleID"/> parameter.
///
/// When the argument passed as <paramref name="Exact"/> parameter is <see langword="false"/>,
/// the returned <see cref="TCultureInfo"/> object is the nearest locale to the locale
/// specified by the <paramref name="LocaleID"/> parameter if the exact match is not found.</remarks>
/// <param name="LocaleID">
/// The locale identifier.</param>
/// <param name="Exact">
/// Indicates whether a dialect of the locale can be a match.</param>
/// <returns>
/// The <see cref="TCultureInfo"/> object or <see langword="nil"/> if the locale
/// identifier is undefined.</returns>
{$ENDREGION}
Function CultureOf(LocaleID: LCID; Exact: Boolean = True): TCultureInfo; Overload; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCultureInfo"/> object that represents a locale specified
/// by its name.</summary>
/// <remarks>
/// Use CultureOf to get information about a locale specified by its name.
///
/// When the argument passed as <paramref name="Exact"/> parameter is <see langword="true"/>,
/// the returned <see cref="TCultureInfo"/> object has the same locale of the locale
/// specified by the <paramref name="Locale"/> parameter.
///
/// When the argument passed as <paramref name="Exact"/> parameter is <see langword="false"/>,
/// the returned <see cref="TCultureInfo"/> object is the nearest locale to the locale
/// specified by the <paramref name="Locale"/> parameter if the exact match is not found.</remarks>
/// <param name="Locale">
/// The locale name.</param>
/// <param name="Exact">
/// Indicates whether a dialect of the locale can be a match.</param>
/// <returns>
/// The <see cref="TCultureInfo"/> object or <see langword="nil"/> if the locale
/// identifier is undefined.</returns>
{$ENDREGION}
Function CultureOf(Const Locale: String; Exact: Boolean = True): TCultureInfo; Overload; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TTerritoryInfo"/> object that represents a country/region
/// specified by its geographical identifier.</summary>
/// <param name="GeoID">
/// The geographical identifier.</param>
/// <returns>
/// The <see cref="TTerritoryInfo"/> object or <see langword="nil"/> if the geographical
/// identifier is undefined.</returns>
{$ENDREGION}
Function TerritoryOf(GeoID: GEOID): TTerritoryInfo; Overload; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TTerritoryInfo"/> object that represents a country/region
/// specified by its two or three characters international code.</summary>
/// <param name="Code">
/// The two or three characters of the international code.</param>
/// <returns>
/// The <see cref="TTerritoryInfo"/> object or <see langword="nil"/> if the international
/// code is undefined.</returns>
{$ENDREGION}
Function TerritoryOf(Const Code: String): TTerritoryInfo; Overload; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Returns the <see cref="TCurrencyInfo"/> object that represents a currency specified
/// by its international monetary symbol.</summary>
/// <param name="IntlSymbol">
/// The three characters of the international monetary symbol.</param>
/// <returns>
/// The <see cref="TCurrencyInfo"/> object or <see langword="nil"/> if the international
/// monetary symbol is undefined.</returns>
{$ENDREGION}
Function CurrencyOf(Const IntlSymbol: String): TCurrencyInfo; Inline;

{$REGION 'xmldoc'}
/// <summary>
/// Indicates whether two locale names represent the same primary language.</summary>
/// <param name="Locale1">
/// The first locale name.</param>
/// <param name="Locale2">
/// The second locale name.</param>
/// <returns>
/// Returns <see langword="true"/> if both locales represent the same primary
/// language, otherwise returns <see langword="false"/>. </returns>
/// <seealso cref="SameCountry"/>
{$ENDREGION}
Function SameLanguage(Const Locale1, Locale2: String): Boolean;

{$REGION 'xmldoc'}
/// <summary>
/// Indicates whether two locale names represent the same country.</summary>
/// <param name="Locale1">
/// The first locale name.</param>
/// <param name="Locale2">
/// The second locale name.</param>
/// <returns>
/// Returns <see langword="true"/> if both locales represent the same country,
/// otherwise returns <see langword="false"/>. </returns>
/// <seealso cref="SameLanguage"/>
{$ENDREGION}
Function SameCountry(Const Locale1, Locale2: String): Boolean;

{$REGION 'xmldoc'}
/// <summary>
/// Call DefaultCalendar to get the <see cref="TCalendar"/> object that indicates
/// the default calendar of the application.</summary>
/// <returns>
/// Returns a <see cref="TCalendar"/> object.</returns>
/// <seealso cref="ChangeDefaultCalendarType"/>
{$ENDREGION}
Function DefaultCalendar: TCalendar;

{$REGION 'xmldoc'}
/// <summary>
/// Call ChangeDefaultCalendarType to change the default calendar system of the
/// application.</summary>
/// <param name="CalendarClass">
/// The <see cref="TCalendar"/> class that represents the new calendar system.</param>
/// <seealso cref="DefaultCalendar"/>
{$ENDREGION}
Procedure ChangeDefaultCalendarType(CalendarClass: TCalendarClass);

Implementation

Uses
  Types, {I18nUtils,} I18nCalGregorian;

Resourcestring
  SInstanceCreateError = 'You cannot create an instance of %s class.';
  SInstanceDestroyError = 'You cannot destroy an instance of %s class.';
  SDuplicateItemError = 'List does not allow duplicates (%u)';

  { Helper Functions }

Var
  TheDefaultCalendar: TCalendar = Nil;
  TheDefaultCalendarClass: TCalendarClass = TGregorianCalendar;

Function I18nAvailable: Boolean;
Begin
  Result := (WinVer >= $0501); // Windows XP or Later
End;

Function DefaultCalendar: TCalendar;
Begin
  If Not Assigned(TheDefaultCalendar) Then
    TheDefaultCalendar := TheDefaultCalendarClass.Create(LOCALE_USER_DEFAULT);
  Result := TheDefaultCalendar;
End;

Procedure ChangeDefaultCalendarType(CalendarClass: TCalendarClass);
Begin
  If TheDefaultCalendarClass <> CalendarClass Then
  Begin
    TheDefaultCalendarClass := CalendarClass;
    If Assigned(TheDefaultCalendar) Then
      FreeAndNil(TheDefaultCalendar);
  End;
End;

Function GetApplicationDefaultCulture: TCultureInfo;
Var
  AppFileName: String;
  VerInfoSize, Dummy: Cardinal;
  VerInfo, Trans: Pointer;
Begin
  Result := Nil;
  AppFileName := ParamStr(0);
  VerInfoSize := GetFileVersioninfoSize(PChar(AppFileName), Dummy);
  If VerInfoSize <> 0 Then
  Begin
    GetMem(VerInfo, VerInfoSize);
    Try
      If GetFileVersionInfo(PChar(PChar(AppFileName)), Dummy, VerInfoSize, VerInfo) And
        VerQueryValue(VerInfo, '\VarFileInfo\Translation', Trans, Dummy) Then
        Result := CultureOf(LoWord(PDWORD(Trans)^));
    Finally
      FreeMem(VerInfo);
    End;
  End;
  If Not Assigned(Result) Then
    Result := CultureOf('en-US');
End;

Function GetSystemDefaultCulture: TCultureInfo;
Begin
  Result := World.Cultures.Find(GetSystemDefaultLCID);
End;

Function GetUserDefaultCulture: TCultureInfo;
Begin
  Result := World.Cultures.Find(GetUserDefaultLCID);
End;

{$INLINE OFF}

Function GetUserDefaultUICulture: TCultureInfo;
Begin
  Result := World.Cultures.Find(GetUserDefaultUILanguage);
End;
{$INLINE AUTO}

Function GetUserTerritory: TTerritoryInfo;
Begin
  Result := World.Territories.Find(GetUserGeoID(GEOCLASS_NATION));
End;

Function CultureOf(LocaleID: LCID; Exact: Boolean): TCultureInfo;
Begin
  If Exact Then
    Result := World.Cultures.Find(LocaleID)
  Else
    Result := World.Cultures.FindNearest(LocaleID);
End;

Function CultureOf(Const Locale: String; Exact: Boolean): TCultureInfo;
Begin
  If Exact Then
    Result := World.Cultures.Find(Locale)
  Else
    Result := World.Cultures.FindNearest(Locale);
End;

Function TerritoryOf(GeoID: GEOID): TTerritoryInfo;
Begin
  Result := World.Territories.Find(GeoID);
End;

Function TerritoryOf(Const Code: String): TTerritoryInfo;
Begin
  Result := World.Territories.Find(Code);
End;

Function CurrencyOf(Const IntlSymbol: String): TCurrencyInfo;
Begin
  Result := World.Currencies.Find(IntlSymbol);
End;

Function SameLanguage(Const Locale1, Locale2: String): Boolean;
Var
  S1, S2: PChar;
Begin
  Result := False;
  S1 := PChar(Locale1);
  S2 := PChar(Locale2);
  While (S1^ <> #0) And (S1^ <> '-') Do
  Begin
    If UpCase(S1^) <> UpCase(S2^) Then
      Exit;
    Inc(S1);
    Inc(S2);
  End;
  Result := (S2^ = #0) Or (S2^ = '-');
End;

Function SameCountry(Const Locale1, Locale2: String): Boolean;
Var
  P1, P2: PChar;
Begin
  P1 := StrRScan(PChar(Locale1), '-');
  P2 := StrRScan(PChar(Locale2), '-');
  Result := (P1 <> Nil) And (P2 <> Nil) And (StrIComp(P1, P2) = 0);
End;

Function IsRightToLeftLanguage(LocaleID: LCID): Boolean;
Var
  I: Integer;
  BufferA: Array [128 .. 255] Of AnsiChar;
  BufferW: Array [128 .. 256] Of WideChar;
Begin
  Result := False;
  For I := Low(BufferA) To High(BufferA) Do
    BufferA[I] := AnsiChar(I);
  GetStringTypeExA(LocaleID, CT_CTYPE2, @BufferA, High(BufferA) - Low(BufferA) + 1, BufferW);
  For I := Low(BufferA) To High(BufferA) Do
    If Ord(BufferW[I]) = C2_RIGHTTOLEFT Then
    Begin
      Result := True;
      Exit;
    End;
End;

{ TCulture }

Constructor TCultureInfo.Create(ALocaleID: LCID);
Var
  CurrencySymbol: String;
  PercentSign: String;
Begin
  If Not World.Changing Then
    Raise Exception.CreateResFmt(@SInstanceCreateError, [ClassName]);
  FLocaleID := ALocaleID;
  FLocale := LCIDToLocaleName(LocaleID);
  FLanguage2 := GetLocaleStr(LocaleID, LOCALE_SISO639LANGNAME, '');
  FLanguage3 := GetLocaleStr(LocaleID, LOCALE_SISO639LANGNAME2, '');
  FGeoID := GetLocaleInt(LocaleID, LOCALE_IGEOID, 0);
  FLocalizedDisplayName := GetLocaleStr(LocaleID, LOCALE_SLOCALIZEDDISPLAYNAME, '');
  FEnglishDisplayName := GetLocaleStr(LocaleID, LOCALE_SENGLISHDISPLAYNAME, '');
  FNativeDisplayName := GetLocaleStr(LocaleID, LOCALE_SNATIVEDISPLAYNAME, '');
  FLocalizedLanguageName := GetLocaleStr(LocaleID, LOCALE_SLOCALIZEDLANGUAGENAME, '');
  FEnglishLanguageName := GetLocaleStr(LocaleID, LOCALE_SENGLISHLANGUAGENAME, '');
  FNativeLanguageName := GetLocaleStr(LocaleID, LOCALE_SNATIVELANGUAGENAME, '');
  FNativeCountryName := GetLocaleStr(LocaleID, LOCALE_SNATIVECOUNTRYNAME, '');
  FNativeCurrencyName := GetLocaleStr(LocaleID, LOCALE_SNATIVECURRNAME, '');
  FScriptCodes := GetLocaleScripts(Locale);
  FScripts := ScriptCodesToScripts(ScriptCodes);
  Case GetLocaleInt(LocaleID, LOCALE_IREADINGLAYOUT, -1) Of
    0:
      FReadingLayout := RlLeftToRight;
    1:
      FReadingLayout := RlRightToLeft;
    2:
      FReadingLayout := RlTopToBottomColRightToLeft;
    3:
      FReadingLayout := RlTopToBottomColLeftToRight;
  Else
    If IsRightToLeftLanguage(LocaleID) Then
      FReadingLayout := RlRightToLeft;
  End;
  If GetLocaleInt(LocaleID, LOCALE_IMEASURE, 0) = 0 Then
    FMeasurementSystem := MsMetric
  Else
    FMeasurementSystem := MsUS;
  FDefaultCodePages[CpOEM] := GetLocaleInt(LocaleID, LOCALE_IDEFAULTCODEPAGE, 0);
  FDefaultCodePages[CpANSI] := GetLocaleInt(LocaleID, LOCALE_IDEFAULTANSICODEPAGE, 0);
  FDefaultCodePages[CpMAC] := GetLocaleInt(LocaleID, LOCALE_IDEFAULTMACCODEPAGE, 0);
  FNativeDigits := GetLocaleStr(LocaleID, LOCALE_SNATIVEDIGITS, NOMINAL_DIGITS);
  FNativeCalendarType := CalendarTypes.ByLocale(Locale);
  If Not Assigned(FNativeCalendarType) Then
    FNativeCalendarType := TGregorianCalendar;
  If GeoID <> 0 Then
  Begin
    FCountry := World.TerritoryList.Find(GeoID);
    If Not Assigned(FCountry) Then
    Begin
      FCountry := TTerritoryInfo.CreateByLocaleID(LocaleID);
      World.TerritoryList.Add(FCountry);
    End;
  End;
  CurrencySymbol := GetLocaleStr(LocaleID, LOCALE_SINTLSYMBOL, '');
  If CurrencySymbol <> '' Then
  Begin
    FCurrency := World.CurrencyList.Find(CurrencySymbol);
    If Not Assigned(FCurrency) Then
    Begin
      FCurrency := TCurrencyInfo.Create(LocaleID);
      World.CurrencyList.Add(FCurrency);
    End;
  End;
  If Assigned(FCountry) Then
    FCountry.Associate(Self, FCurrency);
  If Assigned(FCurrency) Then
    FCurrency.Associate(Self, FCountry);
  Case GetLocaleInt(LocaleID, LOCALE_IPOSITIVEPERCENT, 0) Of
    0:
      FPercentPositive := '# %';
    1:
      FPercentPositive := '#%';
    2:
      FPercentPositive := '%#';
    3:
      FPercentPositive := '% #';
  End;
  Case GetLocaleInt(LocaleID, LOCALE_INEGATIVEPERCENT, 0) Of
    0:
      FPercentNegative := '-# %';
    1:
      FPercentNegative := '-#%';
    2:
      FPercentNegative := '-%#';
    3:
      FPercentNegative := '%-#';
    4:
      FPercentNegative := '%#-';
    5:
      FPercentNegative := '#-%';
    6:
      FPercentNegative := '#%-';
    7:
      FPercentNegative := '-% #';
    8:
      FPercentNegative := '# %-';
    9:
      FPercentNegative := '% #-';
    10:
      FPercentNegative := '% -#';
    11:
      FPercentNegative := '#- %';
  End;
  PercentSign := GetLocaleStr(LocaleID, LOCALE_SPERCENT, '%');
  If PercentSign <> '%' Then
  Begin
    FPercentPositive := StringReplace(FPercentPositive, '%', PercentSign, []);
    FPercentNegative := StringReplace(FPercentNegative, '%', PercentSign, []);
  End;
//  FPluralRule := TPluralForms.RuleOf(Locale, 'nplurals=1; plural=0; ASSUMED');
End;

Constructor TCultureInfo.Create(Const ALocale: String);
Begin
  Create(LocaleNameToLCID(ALocale));
End;

Destructor TCultureInfo.Destroy;
Begin
  If Not World.Changing Then
    Raise Exception.CreateResFmt(@SInstanceDestroyError, [ClassName]);
  If Assigned(FFormatSettings) Then
    Dispose(FFormatSettings);
  If Assigned(FNativeCalendar) Then
    FNativeCalendar.Free;
  If Assigned(FCalendarTypes) Then
    FCalendarTypes.Free;
  Inherited Destroy;
End;

Function TCultureInfo.GetDisplayNames(DisplayName: TCultureDisplayName): String;
Begin
  Case DisplayName Of
    CnLocale:
      Result := Locale;
    CnLanguage:
      Result := Language2;
    CnLanguage3:
      Result := Language3;
    CnLocalizedDisplayName:
      Result := LocalizedDisplayName;
    CnEnglishDisplayName:
      Result := EnglishDisplayName;
    CnNativeDisplayName:
      Result := NativeDisplayName;
    CnLocalizedLanguageName:
      Result := LocalizedLanguageName;
    CnEnglishLanguageName:
      Result := EnglishLanguageName;
    CnNativeLanguageName:
      Result := NativeLanguageName;
    CnNativeCountryName:
      Result := NativeCountryName;
    CnNativeCurrencyName:
      Result := NativeCurrencyName;
  Else
    Result := '';
  End;
End;

Function TCultureInfo.GetDefaultCodePages(CodePageType: TCodePageType): Integer;
Begin
  Result := FDefaultCodePages[CodePageType];
End;

Function TCultureInfo.GetLangID: LANGID;
Begin
  Result := Windows.LANGID(LocaleID);
End;

Function TCultureInfo.GetPrimaryLangID: LANGID;
Begin
  Result := Windows.PRIMARYLANGID(LocaleID);
End;

Function TCultureInfo.GetSubLangID: LANGID;
Begin
  Result := Windows.SUBLANGID(LocaleID);
End;

Function TCultureInfo.GetBiDiMode: TBiDiMode;
Begin
  If IsRightToLeft Then
    Result := BdRightToLeft
  Else
    Result := BdLeftToRight;
End;

Function TCultureInfo.GetFormatSettings: PFormatSettings;
Begin
  If Not Assigned(FFormatSettings) Then
  Begin
    New(FFormatSettings);
{$IFDEF COMPILER_XE2_UP}
    FFormatSettings^ := TFormatSettings.Create(LocaleID);
{$ELSE}
    GetLocaleFormatSettings(LocaleID, FFormatSettings^);
{$ENDIF}
  End;
  Result := FFormatSettings;
End;

Function TCultureInfo.GetNativeCalendar: TCalendar;
Begin
  If Not Assigned(FNativeCalendar) Then
    FNativeCalendar := NativeCalendarType.Create(Locale);
  Result := FNativeCalendar;
End;

Function TCultureInfo.IsRightToLeft: Boolean;
Begin
  Result := (ReadingLayout = RlRightToLeft);
End;

Function TCultureInfo.IsDefault: Boolean;
Begin
  Result := (SubLangID = SUBLANG_DEFAULT);
End;

Function TCultureInfo.IsDialectOf(Culture: TCultureInfo): Boolean;
Begin
  Result := (Culture.PrimaryLangID = PrimaryLangID);
End;

Function TCultureInfo.IsDialectOf(Const Locale: String): Boolean;
Begin
  Result := (Pos(Language2, LowerCase(Locale)) = 1) And
    ((Length(Language2) = Length(Locale)) Or (Locale[Length(Language2) + 1] = '-'));
End;

Function TCultureInfo.IsDialectOf(LangID: LANGID): Boolean;
Begin
  Result := (Windows.PRIMARYLANGID(LangID) = PrimaryLangID);
End;

Function TCultureInfo.IsUsingNominalDigits: Boolean;
Begin
  Result := (Pos('0', NativeDigits) = 1);
End;

Function TCultureInfo.IsUsingCommonEraCalendar: Boolean;
Begin
  Result := (NativeCalendarType = TGregorianCalendar);
End;

Function TCultureInfo.DrawTextBiDiModeFlags(Flags: Cardinal): Integer;
Begin
  Result := Flags;
  If IsRightToLeft Then
  Begin
    If (Result And DT_RIGHT) = DT_RIGHT Then
      Result := Result And Not DT_RIGHT
    Else If (Result And DT_CENTER) <> DT_CENTER Then
      Result := Result Or DT_RIGHT;
  End;
  Result := Result Or DrawTextBiDiModeFlagsReadingOnly;
End;

Function TCultureInfo.DrawTextBiDiModeFlagsReadingOnly: Integer;
Begin
  If IsRightToLeft Then
    Result := DT_RTLREADING
  Else
    Result := 0;
End;

Function TCultureInfo.VerifyString(Const Str: String; AllowLatin: Boolean): Boolean;
Begin
  Result := VerifyScripts(ScriptCodes, GetStringScripts(Str));
End;
{
Function TCultureInfo.OemToUnicode(Const Str: RawByteString): String;
Begin
  Result := CodePageToUnicode(DefaultCodePages[CpOEM], Str);
End;

Function TCultureInfo.UnicodeToOem(Const Str: String): RawByteString;
Begin
  Result := UnicodeToCodePage(DefaultCodePages[CpOEM], Str);
End;

Function TCultureInfo.AnsiToUnicode(Const Str: RawByteString): String;
Begin
  Result := CodePageToUnicode(DefaultCodePages[CpANSI], Str);
End;

Function TCultureInfo.UnicodeToAnsi(Const Str: String): RawByteString;
Begin
  Result := UnicodeToCodePage(DefaultCodePages[CpANSI], Str);
End;

Function TCultureInfo.MacToUnicode(Const Str: RawByteString): String;
Begin
  Result := CodePageToUnicode(DefaultCodePages[CpMAC], Str);
End;

Function TCultureInfo.UnicodeToMac(Const Str: String): RawByteString;
Begin
  Result := UnicodeToCodePage(DefaultCodePages[CpMAC], Str);
End;
     }
Function TCultureInfo.NominalDigitsToNative(Const Str: String; IgnoreUCCs: Boolean): String;
Var
  S: PChar;
  Zero: Char;
  CanChange: Boolean;
Begin
  Result := Str;
  Zero := PChar(NativeDigits)^;
  If Zero <> '0' Then
  Begin
    CanChange := True;
    S := PChar(Result);
    While S^ <> #0 Do
    Begin
      If S^ = UCC_NODS Then
        CanChange := IgnoreUCCs
      Else If S^ = UCC_NADS Then
        CanChange := True
      Else If CanChange And (S^ >= '0') And (S^ <= '9') Then
        S^ := Chr(Ord(Zero) + Ord(S^) - Ord('0'));
      Inc(S);
    End;
  End;
End;

Function TCultureInfo.NativeDigitsToNominal(Const Str: String; IgnoreUCCs: Boolean): String;
Var
  S: PChar;
  Zero: Char;
  CanChange: Boolean;
Begin
  Result := Str;
  Zero := PChar(NativeDigits)^;
  If Zero <> '0' Then
  Begin
    CanChange := True;
    S := PChar(Result);
    While S^ <> #0 Do
    Begin
      If S^ = UCC_NADS Then
        CanChange := IgnoreUCCs
      Else If S^ = UCC_NODS Then
        CanChange := True
      Else If CanChange And (Cardinal(Ord(S^) - Ord(Zero)) <= 9) Then
        S^ := Chr(Ord('0') + Ord(S^) - Ord(Zero));
      Inc(S);
    End;
  End;
End;

Function TCultureInfo.FreezeDigits(Const Str: String; UseNativeDigits: Boolean): String;
Begin
  If UseNativeDigits Then
    Result := UCC_NODS + NominalDigitsToNative(Str, False)
  Else
    Result := UCC_NODS + Str;
End;

Function TCultureInfo.UnfreezeDigits(Const Str: String): String;
Begin
  Result := Str;
  If PChar(Result)^ = UCC_NODS Then
  Begin
    Delete(Result, 1, 1);
    Result := NativeDigitsToNominal(Result, False);
  End;
End;

Function TCultureInfo.Format(Const Fmt: String; Const Args: Array Of Const; UseNativeDigits: Boolean): String;
Begin
  Result := FreezeDigits(SysUtils.Format(Fmt, Args, FormatSettings^), UseNativeDigits);
End;

Function TCultureInfo.FormatNumber(Const Fmt: String; Const Value: Extended; UseNativeDigits: Boolean): String;
Begin
  Result := FreezeDigits(SysUtils.FormatFloat(Fmt, Value, FormatSettings^), UseNativeDigits);
End;

Function TCultureInfo.FormatPercent(Const Value: Extended; Decimals: Integer; UseNativeDigits: Boolean): String;
Var
  Str, Fmt: String;
Begin
  If Decimals < 0 Then
    Fmt := '#,##0.' + StringOfChar('0', Abs(Decimals))
  Else
    Fmt := '#,##0.' + StringOfChar('#', Decimals);
  Str := FormatNumber(Fmt, Abs(Value * 100.0), UseNativeDigits);
  If Value < 0 Then
    Result := StringReplace(FPercentNegative, '#', Str, [])
  Else
    Result := StringReplace(FPercentPositive, '#', Str, []);
End;

Function TCultureInfo.FormatDateTime(Const Fmt: String; Const Value: TDateTime;
  UseNativeDigits: Boolean): String;
Begin
  Result := FreezeDigits(NativeCalendar.Format(Fmt, Value), UseNativeDigits);
End;

Function TCultureInfo.ParseDateTime(Const Fmt: String; Const Str: String): TDateTime;
Begin
  Result := NativeCalendar.Parse(Fmt, UnfreezeDigits(Str));
End;

Function TCultureInfo.TryParseDateTime(Const Fmt: String; Const Str: String; Var DateTime: TDateTime): Boolean;
Begin
  Result := NativeCalendar.TryParse(Fmt, UnfreezeDigits(Str), DateTime);
End;

{ TCultureEnumerator }

Constructor TCultureEnumerator.Create(List: TReadonlyCultureList);
Begin
  Inherited Create(List.List);
End;

Function TCultureEnumerator.GetCurrent: TCultureInfo;
Begin
  Result := TCultureInfo( Inherited GetCurrent);
End;

{ TReadonlyCultureList }

Constructor TReadonlyCultureList.Create;
Begin
  List := TList.Create;
End;

Destructor TReadonlyCultureList.Destroy;
Begin
  List.Free;
  If Assigned(Lookup_LocaleID) Then
    Lookup_LocaleID.Free;
  If Assigned(Lookup_Locale) Then
    Lookup_Locale.Free;
  Inherited;
End;

Function TReadonlyCultureList.GetCount: Integer;
Begin
  Result := List.Count;
End;

Function TReadonlyCultureList.GetItems(Index: Integer): TCultureInfo;
Begin
  Result := TCultureInfo(List[Index]);
End;

Function TReadonlyCultureList.GetEnumerator: TCultureEnumerator;
Begin
  Result := TCultureEnumerator.Create(Self);
End;

Procedure TReadonlyCultureList.AddLookupEntries(Culture: TCultureInfo; Index: Integer);
Begin
  If IsLookupTableValid_LocaleID Then
    Lookup_LocaleID.Add(Culture.LocaleID, Index);
  If IsLookupTableValid_Locale Then
    Lookup_Locale.Add(AnsiLowerCase(Culture.Locale), Index);
End;

Procedure TReadonlyCultureList.RemoveLookupEntries(Culture: TCultureInfo);
Begin
  If IsLookupTableValid_LocaleID Then
    Lookup_LocaleID.Remove(Culture.LocaleID);
  If IsLookupTableValid_Locale Then
    Lookup_Locale.Remove(AnsiLowerCase(Culture.Locale));
End;

Procedure TReadonlyCultureList.ClearLookupTables;
Begin
  If Assigned(Lookup_LocaleID) Then
  Begin
    Lookup_LocaleID.Clear;
    IsLookupTableValid_LocaleID := False;
  End;
  If Assigned(Lookup_Locale) Then
  Begin
    Lookup_Locale.Clear;
    IsLookupTableValid_Locale := False;
  End;
End;

Procedure TReadonlyCultureList.InvalidateLookupTables;
Begin
  IsLookupTableValid_LocaleID := False;
  IsLookupTableValid_Locale := False;
End;

Procedure TReadonlyCultureList.ValidateLookupTable_LocaleID;
Var
  I: Integer;
Begin
  If Not IsLookupTableValid_LocaleID Then
  Begin
    If Assigned(Lookup_LocaleID) Then
      Lookup_LocaleID.Clear
    Else
      Lookup_LocaleID := TKeyLookup<LCID, Integer>.Create;
    For I := 0 To Count - 1 Do
      Lookup_LocaleID.Add(Items[I].LocaleID, I);
    IsLookupTableValid_LocaleID := True;
  End;
End;

Procedure TReadonlyCultureList.ValidateLookupTable_Locale;
Var
  I: Integer;
Begin
  If Not IsLookupTableValid_Locale Then
  Begin
    If Assigned(Lookup_Locale) Then
      Lookup_Locale.Clear
    Else
      Lookup_Locale := TKeyLookup<String, Integer>.Create;
    For I := 0 To Count - 1 Do
      Lookup_Locale.Add(LowerCase(Items[I].Locale), I);
    IsLookupTableValid_Locale := True;
  End;
End;

Function TReadonlyCultureList.First: TCultureInfo;
Begin
  If Count <> 0 Then
    Result := Items[0]
  Else
    Result := Nil;
End;

Function TReadonlyCultureList.Last: TCultureInfo;
Begin
  If Count <> 0 Then
    Result := Items[Count - 1]
  Else
    Result := Nil;
End;

Procedure TReadonlyCultureList.AssignTo(Dest: TPersistent);
Var
  Strings: TStrings Absolute Dest;
  I: Integer;
Begin
  If Dest Is TStrings Then
  Begin
    Strings.BeginUpdate;
    Try
      Strings.Clear;
      For I := 0 To Count - 1 Do
        Strings.AddObject(Items[I].Locale, Items[I]);
    Finally
      Strings.EndUpdate;
    End;
  End
  Else
    Inherited AssignTo(Dest);
End;

Procedure TReadonlyCultureList.AssignLabelsTo(Dest: TStrings; DisplayName: TCultureDisplayName);
Var
  I: Integer;
Begin
  Dest.BeginUpdate;
  Try
    Dest.Clear;
    For I := 0 To Count - 1 Do
      Dest.AddObject(Items[I].DisplayNames[DisplayName], Items[I]);
  Finally
    Dest.EndUpdate;
  End;
End;

Function TReadonlyCultureList.Find(LocaleID: LCID): TCultureInfo;
Var
  Index: Integer;
Begin
  If Not IsLookupTableValid_LocaleID Then
    ValidateLookupTable_LocaleID;
  If Lookup_LocaleID.Retrieve(LocaleID, Index) Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyCultureList.Find(Const Locale: String): TCultureInfo;
Var
  Index: Integer;
Begin
  If Not IsLookupTableValid_Locale Then
    ValidateLookupTable_Locale;
  If Lookup_Locale.Retrieve(LowerCase(Locale), Index) Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyCultureList.FindNearest(Culture: TCultureInfo): TCultureInfo;
Var
  Index: Integer;
Begin
  Index := NearestIndexOf(Culture);
  If Index >= 0 Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyCultureList.FindNearest(LocaleID: LCID): TCultureInfo;
Var
  Index: Integer;
Begin
  Index := NearestIndexOf(LocaleID);
  If Index >= 0 Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyCultureList.FindNearest(Const Locale: String): TCultureInfo;
Var
  Index: Integer;
Begin
  Index := NearestIndexOf(Locale);
  If Index >= 0 Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyCultureList.IndexOf(Culture: TCultureInfo): Integer;
Begin
  If Assigned(Culture) Then
    Result := IndexOf(Culture.LocaleID)
  Else
    Result := -1;
End;

Function TReadonlyCultureList.IndexOf(LocaleID: LCID): Integer;
Begin
  If Not IsLookupTableValid_LocaleID Then
    ValidateLookupTable_LocaleID;
  If Not Lookup_LocaleID.Retrieve(LocaleID, Result) Then
    Result := -1;
End;

Function TReadonlyCultureList.IndexOf(Const Locale: String): Integer;
Begin
  If Not IsLookupTableValid_Locale Then
    ValidateLookupTable_Locale;
  If Not Lookup_Locale.Retrieve(LowerCase(Locale), Result) Then
    Result := -1;
End;

Function TReadonlyCultureList.IndexOfName(Const AName: String; DisplayName: TCultureDisplayName): Integer;
Var
  I: Integer;
Begin
  For I := 0 To Count - 1 Do
    If AnsiSameText(Items[I].DisplayNames[DisplayName], AName) Then
    Begin
      Result := I;
      Exit;
    End;
  Result := -1;
End;

Function TReadonlyCultureList.NearestIndexOf(Culture: TCultureInfo): Integer;
Begin
  If Assigned(Culture) Then
    Result := NearestIndexOf(Culture.LocaleID)
  Else
    Result := -1;
End;

Function TReadonlyCultureList.NearestIndexOf(LocaleID: LCID): Integer;
Var
  PriID, SubID: LANGID;
  I: Integer;
Begin
  Result := IndexOf(LocaleID);
  If Result >= 0 Then
    Exit;
  PriID := Windows.PRIMARYLANGID(LocaleID);
  If PriID = 0 Then
    Exit;
  SubID := High(LANGID);
  For I := 0 To Count - 1 Do
    With Items[I] Do
    Begin
      If PrimaryLangID <> PriID Then
        Continue;
      If IsDefault Then
      Begin
        Result := I;
        Exit;
      End;
      If (Result < 0) Or (SubLangID < SubID) Then
      Begin
        Result := I;
        SubID := SubLangID;
      End;
    End;
End;

Function TReadonlyCultureList.NearestIndexOf(Const Locale: String): Integer;
Var
  Lang: String;
  SubID: LANGID;
  I: Integer;
Begin
  Result := IndexOf(Locale);
  If Result >= 0 Then
    Exit;
  I := Pos('-', Locale);
  If I <> 0 Then
    Lang := LowerCase(Copy(Locale, 1, I - 1))
  Else
    Lang := LowerCase(Locale);
  If Lang = '' Then
    Exit;
  SubID := High(LANGID);
  For I := 0 To Count - 1 Do
    With Items[I] Do
    Begin
      If Language2 <> Lang Then
        Continue;
      If IsDefault Then
      Begin
        Result := I;
        Exit;
      End;
      If (Result < 0) Or (SubLangID < SubID) Then
      Begin
        Result := I;
        SubID := SubLangID;
      End;
    End;
End;

Function TReadonlyCultureList.Exists(Culture: TCultureInfo): Boolean;
Begin
  Result := Assigned(Culture) And Exists(Culture.LocaleID);
End;

Function TReadonlyCultureList.Exists(LocaleID: LCID): Boolean;
Begin
  If Not IsLookupTableValid_LocaleID Then
    ValidateLookupTable_LocaleID;
  Result := Lookup_LocaleID.Exists(LocaleID);
End;

Function TReadonlyCultureList.Exists(Const Locale: String): Boolean;
Begin
  If Not IsLookupTableValid_Locale Then
    ValidateLookupTable_Locale;
  Result := Lookup_Locale.Exists(LowerCase(Locale));
End;

{ TCultureList }

Function TCultureList.GetCapacity: Integer;
Begin
  Result := List.Capacity;
End;

Procedure TCultureList.SetCapacity(Value: Integer);
Begin
  List.Capacity := Value;
  If Value < Count Then
    InvalidateLookupTables;
End;

Procedure TCultureList.ReadLocales(Reader: TReader);
Var
  Locale: String;
Begin
  BeginUpdate;
  Try
    Clear;
    Reader.ReadListBegin;
    While Not Reader.EndOfList Do
    Begin
      Locale := Reader.ReadString;
      If Locale = '*' Then
      Begin
        Assign(World.Cultures);
        Break;
      End;
      Add(Locale);
    End;
    Reader.ReadListEnd;
  Finally
    EndUpdate;
  End;
End;

Procedure TCultureList.WriteLocales(Writer: TWriter);
Var
  I: Integer;
Begin
  Writer.WriteListBegin;
  If (Duplicates <> DupAccept) And (Count = World.Cultures.Count) Then
    Writer.WriteString('*') // all cultures
  Else
    For I := 0 To Count - 1 Do
      Writer.WriteString(Items[I].Locale);
  Writer.WriteListEnd;
End;

Procedure TCultureList.DefineProperties(Filer: TFiler);
Begin
  Inherited DefineProperties(Filer);
  Filer.DefineProperty('Cultures', ReadLocales, WriteLocales, Count <> 0);
End;

Procedure TCultureList.Change;
Begin
  If (UpdateCount = 0) And Assigned(OnChange) Then
    OnChange(Self);
End;

Procedure TCultureList.Changing;
Begin
  If (UpdateCount = 0) And Assigned(OnChanging) Then
    OnChanging(Self);
End;

Procedure TCultureList.BeginUpdate;
Begin
  If UpdateCount = 0 Then
    Changing;
  Inc(FUpdateCount);
End;

Procedure TCultureList.EndUpdate;
Begin
  Dec(FUpdateCount);
  If UpdateCount = 0 Then
    Change;
End;

Procedure TCultureList.Clear;
Begin
  If Count <> 0 Then
  Begin
    Changing;
    Try
      List.Clear;
      ClearLookupTables;
    Finally
      Change;
    End;
  End;
End;

Procedure TCultureList.Sort(OrderBy: TCultureDisplayName);

  Procedure QuickSort(L, R: Integer);
  Var
    I, J, P: Integer;
    Pivot: String;
  Begin
    Repeat
      I := L;
      J := R;
      P := (L + R) Shr 1;
      Repeat
        Pivot := Items[P].DisplayNames[OrderBy];
        While AnsiCompareText(Items[I].DisplayNames[OrderBy], Pivot) < 0 Do
          Inc(I);
        While AnsiCompareText(Items[J].DisplayNames[OrderBy], Pivot) > 0 Do
          Dec(J);
        If I <= J Then
        Begin
          If I <> J Then
          Begin
            List.Exchange(I, J);
            InvalidateLookupTables;
          End;
          If P = I Then
            P := J
          Else If P = J Then
            P := I;
          Inc(I);
          Dec(J);
        End;
      Until I > J;
      If L < J Then
        QuickSort(L, J);
      L := I;
    Until I >= R;
  End;

Begin
  If (OrderBy <> CnNone) And (Count > 1) Then
  Begin
    Changing;
    Try
      QuickSort(0, Count - 1);
    Finally
      Change;
    End;
  End;
End;

Procedure TCultureList.Assign(Source: TPersistent);
Begin
  If Source Is TReadonlyCultureList Then
    Apply(TReadonlyCultureList(Source), LaCopy)
  Else
    Inherited Assign(Source);
End;

Procedure TCultureList.Apply(ListA: TReadonlyCultureList; Op: TListAssignOp; ListB: TReadonlyCultureList);
Begin
  Changing;
  Try
    If Assigned(ListB) Then
      List.Assign(ListA.List, Op, ListB.List)
    Else
      List.Assign(ListA.List, Op, Nil);
    InvalidateLookupTables;
  Finally
    Change;
  End;
End;

Function TCultureList.Add(Culture: TCultureInfo): Integer;
Begin
  Result := -1;
  If Assigned(Culture) Then
  Begin
    If Duplicates <> DupAccept Then
    Begin
      Result := IndexOf(Culture);
      If Result >= 0 Then
      Begin
        If Duplicates = DupError Then
          Raise EListError.CreateResFmt(@SDuplicateItemError, [Result]);
        Exit;
      End;
    End;
    Changing;
    Try
      Result := List.Add(Culture);
      AddLookupEntries(Culture, Result);
    Finally
      Change;
    End;
  End;
End;

Function TCultureList.Add(LocaleID: LCID): Integer;
Begin
  Result := Add(World.Cultures.Find(LocaleID));
End;

Function TCultureList.Add(Const Locale: String): Integer;
Begin
  Result := Add(World.Cultures.Find(Locale));
End;

Function TCultureList.Remove(Culture: TCultureInfo): Integer;
Begin
  Result := IndexOf(Culture);
  If Result >= 0 Then
    Delete(Result);
End;

Function TCultureList.Remove(LocaleID: LCID): Integer;
Begin
  Result := IndexOf(LocaleID);
  If Result >= 0 Then
    Delete(Result);
End;

Function TCultureList.Remove(Const Locale: String): Integer;
Begin
  Result := IndexOf(Locale);
  If Result >= 0 Then
    Delete(Result);
End;

Procedure TCultureList.Delete(Index: Integer);
Begin
  Changing;
  Try
    RemoveLookupEntries(Items[Index]);
    List.Delete(Index);
  Finally
    Change;
  End;
End;

{ TTerritory }

Constructor TTerritoryInfo.Create(AGeoID: GEOID);
Begin
  If Not World.Changing Then
    Raise Exception.CreateResFmt(@SInstanceCreateError, [ClassName]);
  FCultures := TCultureList.Create;
  FCurrencies := TCurrencyList.Create;
  FGeoID := AGeoID;
  FCode2 := GetGeoStr(GeoID, GEO_ISO2, '');
  FCode3 := GetGeoStr(GeoID, GEO_ISO3, '');
  FLatitude := StrToFloatDef(GetGeoStr(GeoID, GEO_LATITUDE, '0'), 0);
  FLongitude := StrToFloatDef(GetGeoStr(GeoID, GEO_LONGITUDE, '0'), 0);
  FFriendlyName := GetGeoStr(GeoID, GEO_FRIENDLYNAME, '');
  FOfficialName := GetGeoStr(GeoID, GEO_OFFICIALNAME, '');
  FLocalizedName := FriendlyName;
  FEnglishName := FriendlyName;
End;

Constructor TTerritoryInfo.CreateByLocaleID(LocaleID: LCID; Dummy: Integer);
Begin
  Create(GetLocaleInt(LocaleID, LOCALE_IGEOID, 0));
  FLocalizedName := GetLocaleStr(LocaleID, LOCALE_SLOCALIZEDCOUNTRYNAME, '');
  FEnglishName := GetLocaleStr(LocaleID, LOCALE_SENGLISHCOUNTRYNAME, '');
End;

Constructor TTerritoryInfo.CreateByLocale(Const Locale: String);
Begin
  CreateByLocaleID(LocaleNameToLCID(Locale));
End;

Destructor TTerritoryInfo.Destroy;
Begin
  If Not World.Changing Then
    Raise Exception.CreateResFmt(@SInstanceDestroyError, [ClassName]);
  FCultures.Free;
  FCurrencies.Free;
  Inherited Destroy;
End;

Function TTerritoryInfo.GetDisplayNames(DisplayName: TTerritoryDisplayName): String;
Begin
  Case DisplayName Of
    TnCode2:
      Result := Code2;
    TnCode3:
      Result := Code3;
    TnFriendlyName:
      Result := FriendlyName;
    TnOfficialName:
      Result := OfficialName;
    TnLocalizedName:
      Result := LocalizedName;
    TnEnglishName:
      Result := EnglishName;
    TnNativeName:
      Result := NativeName;
  Else
    Result := '';
  End;
End;

Function TTerritoryInfo.GetNativeName: String;
Begin
  If Cultures.Count <> 0 Then
    Result := Cultures[0].NativeCountryName
  Else
    Result := '';
End;

Function TTerritoryInfo.GetCurrencies: TReadonlyCurrencyList;
Begin
  Result := FCurrencies;
End;

Function TTerritoryInfo.GetCultures: TReadonlyCultureList;
Begin
  Result := FCultures;
End;

Procedure TTerritoryInfo.Associate(Culture: TCultureInfo; Currency: TCurrencyInfo);
Begin
  If Assigned(Culture) And Not FCultures.Exists(Culture) Then
    FCultures.Add(Culture);
  If Assigned(Currency) And Not FCurrencies.Exists(Currency) Then
    FCurrencies.Add(Currency);
End;

{ TTerritoryEnumerator }

Constructor TTerritoryEnumerator.Create(List: TReadonlyTerritoryList);
Begin
  Inherited Create(List.List);
End;

Function TTerritoryEnumerator.GetCurrent: TTerritoryInfo;
Begin
  Result := TTerritoryInfo( Inherited GetCurrent);
End;

{ TReadonlyTerritoryList }

Constructor TReadonlyTerritoryList.Create;
Begin
  List := TList.Create;
End;

Destructor TReadonlyTerritoryList.Destroy;
Begin
  List.Free;
  If Assigned(LookupTable_GeoID) Then
    LookupTable_GeoID.Free;
  If Assigned(LookupTable_Code2) Then
    LookupTable_Code2.Free;
  If Assigned(LookupTable_Code3) Then
    LookupTable_Code3.Free;
  Inherited Destroy;
End;

Function TReadonlyTerritoryList.GetCount;
Begin
  Result := List.Count;
End;

Function TReadonlyTerritoryList.GetItems(Index: Integer): TTerritoryInfo;
Begin
  Result := TTerritoryInfo(List[Index]);
End;

Function TReadonlyTerritoryList.GetEnumerator: TTerritoryEnumerator;
Begin
  Result := TTerritoryEnumerator.Create(Self);
End;

Procedure TReadonlyTerritoryList.AddLookupEntries(Territory: TTerritoryInfo; Index: Integer);
Begin
  If IsLookupTableValid_GeoID Then
    LookupTable_GeoID.Add(Territory.GeoID, Index);
  If IsLookupTableValid_Code2 Then
    LookupTable_Code2.Add(Territory.Code2, Index);
  If IsLookupTableValid_Code3 Then
    LookupTable_Code3.Add(Territory.Code3, Index);
End;

Procedure TReadonlyTerritoryList.RemoveLookupEntries(Territory: TTerritoryInfo);
Begin
  If IsLookupTableValid_GeoID Then
    LookupTable_GeoID.Remove(Territory.GeoID);
  If IsLookupTableValid_Code2 Then
    LookupTable_Code2.Remove(Territory.Code2);
  If IsLookupTableValid_Code3 Then
    LookupTable_Code3.Remove(Territory.Code3);
End;

Procedure TReadonlyTerritoryList.ClearLookupTables;
Begin
  If Assigned(LookupTable_GeoID) Then
  Begin
    LookupTable_GeoID.Clear;
    IsLookupTableValid_GeoID := False;
  End;
  If Assigned(LookupTable_Code2) Then
  Begin
    LookupTable_Code2.Clear;
    IsLookupTableValid_Code2 := False;
  End;
  If Assigned(LookupTable_Code3) Then
  Begin
    LookupTable_Code3.Clear;
    IsLookupTableValid_Code3 := False;
  End;
End;

Procedure TReadonlyTerritoryList.InvalidateLookupTables;
Begin
  IsLookupTableValid_GeoID := False;
  IsLookupTableValid_Code2 := False;
  IsLookupTableValid_Code3 := False;
End;

Procedure TReadonlyTerritoryList.ValidateLookupTable_GeoID;
Var
  I: Integer;
Begin
  If Not IsLookupTableValid_GeoID Then
  Begin
    If Assigned(LookupTable_GeoID) Then
      LookupTable_GeoID.Clear
    Else
      LookupTable_GeoID := TKeyLookup<GEOID, Integer>.Create;
    For I := 0 To Count - 1 Do
      LookupTable_GeoID.Add(Items[I].GeoID, I);
    IsLookupTableValid_GeoID := True;
  End;
End;

Procedure TReadonlyTerritoryList.ValidateLookupTable_Code2;
Var
  I: Integer;
Begin
  If Not IsLookupTableValid_Code2 Then
  Begin
    If Assigned(LookupTable_Code2) Then
      LookupTable_Code2.Clear
    Else
      LookupTable_Code2 := TKeyLookup<String, Integer>.Create;
    For I := 0 To Count - 1 Do
      LookupTable_Code2.Add(Items[I].Code2, I);
    IsLookupTableValid_Code2 := True;
  End;
End;

Procedure TReadonlyTerritoryList.ValidateTookupTable_Code3;
Var
  I: Integer;
Begin
  If Not IsLookupTableValid_Code3 Then
  Begin
    If Assigned(LookupTable_Code3) Then
      LookupTable_Code3.Clear
    Else
      LookupTable_Code3 := TKeyLookup<String, Integer>.Create;
    For I := 0 To Count - 1 Do
      LookupTable_Code3.Add(Items[I].Code3, I);
    IsLookupTableValid_Code3 := True;
  End;
End;

Function TReadonlyTerritoryList.First: TTerritoryInfo;
Begin
  If Count <> 0 Then
    Result := Items[0]
  Else
    Result := Nil;
End;

Function TReadonlyTerritoryList.Last: TTerritoryInfo;
Begin
  If Count <> 0 Then
    Result := Items[Count - 1]
  Else
    Result := Nil;
End;

Procedure TReadonlyTerritoryList.AssignTo(Dest: TPersistent);
Var
  Strings: TStrings Absolute Dest;
  I: Integer;
Begin
  If Dest Is TStrings Then
  Begin
    Strings.BeginUpdate;
    Try
      Strings.Clear;
      For I := 0 To Count - 1 Do
        Strings.AddObject(Items[I].Code2, Items[I]);
    Finally
      Strings.EndUpdate;
    End;
  End
  Else
    Inherited AssignTo(Dest);
End;

Procedure TReadonlyTerritoryList.AssignLabelsTo(Dest: TStrings; DisplayName: TTerritoryDisplayName);
Var
  I: Integer;
Begin
  Dest.BeginUpdate;
  Try
    Dest.Clear;
    For I := 0 To Count - 1 Do
      Dest.AddObject(Items[I].DisplayNames[DisplayName], Items[I]);
  Finally
    Dest.EndUpdate;
  End;
End;

Function TReadonlyTerritoryList.Find(GeoID: GEOID): TTerritoryInfo;
Var
  Index: Integer;
Begin
  If Not IsLookupTableValid_GeoID Then
    ValidateLookupTable_GeoID;
  If LookupTable_GeoID.Retrieve(GeoID, Index) Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyTerritoryList.Find(Const Code: String): TTerritoryInfo;
Var
  Index: Integer;
Begin
  Result := Nil;
  Case Length(Code) Of
    2:
      Begin
        If Not IsLookupTableValid_Code2 Then
          ValidateLookupTable_Code2;
        If LookupTable_Code2.Retrieve(UpperCase(Code), Index) Then
          Result := Items[Index];
      End;
    3:
      Begin
        If Not IsLookupTableValid_Code3 Then
          ValidateTookupTable_Code3;
        If LookupTable_Code3.Retrieve(UpperCase(Code), Index) Then
          Result := Items[Index];
      End;
  End;
End;

Function TReadonlyTerritoryList.IndexOf(Territory: TTerritoryInfo): Integer;
Begin
  If Assigned(Territory) Then
    Result := IndexOf(Territory.GeoID)
  Else
    Result := -1;
End;

Function TReadonlyTerritoryList.IndexOf(GeoID: GEOID): Integer;
Begin
  If Not IsLookupTableValid_GeoID Then
    ValidateLookupTable_GeoID;
  If Not LookupTable_GeoID.Retrieve(GeoID, Result) Then
    Result := -1;
End;

Function TReadonlyTerritoryList.IndexOf(Const Code: String): Integer;
Begin
  Case Length(Code) Of
    2:
      Begin
        If Not IsLookupTableValid_Code2 Then
          ValidateLookupTable_Code2;
        If Not LookupTable_Code2.Retrieve(UpperCase(Code), Result) Then
          Result := -1;
      End;
    3:
      Begin
        If Not IsLookupTableValid_Code3 Then
          ValidateTookupTable_Code3;
        If Not LookupTable_Code3.Retrieve(UpperCase(Code), Result) Then
          Result := -1;
      End;
  Else
    Result := -1;
  End;
End;

Function TReadonlyTerritoryList.IndexOfName(Const AName: String; DisplayName: TTerritoryDisplayName): Integer;
Var
  I: Integer;
Begin
  For I := 0 To Count - 1 Do
    If AnsiSameText(Items[I].DisplayNames[DisplayName], AName) Then
    Begin
      Result := I;
      Exit;
    End;
  Result := -1;
End;

Function TReadonlyTerritoryList.Exists(Territory: TTerritoryInfo): Boolean;
Begin
  Result := Assigned(Territory) And Exists(Territory.GeoID);
End;

Function TReadonlyTerritoryList.Exists(GeoID: GEOID): Boolean;
Begin
  If Not IsLookupTableValid_GeoID Then
    ValidateLookupTable_GeoID;
  Result := LookupTable_GeoID.Exists(GeoID);
End;

Function TReadonlyTerritoryList.Exists(Const Code: String): Boolean;
Begin
  Case Length(Code) Of
    2:
      Begin
        If Not IsLookupTableValid_Code2 Then
          ValidateLookupTable_Code2;
        Result := LookupTable_Code2.Exists(UpperCase(Code));
      End;
    3:
      Begin
        If Not IsLookupTableValid_Code3 Then
          ValidateTookupTable_Code3;
        Result := LookupTable_Code3.Exists(UpperCase(Code));
      End;
  Else
    Result := False;
  End;
End;

{ TTerritoryList }

Function TTerritoryList.GetCapacity: Integer;
Begin
  Result := List.Capacity;
End;

Procedure TTerritoryList.SetCapacity(Value: Integer);
Begin
  List.Capacity := Value;
  If Value < Count Then
    InvalidateLookupTables;
End;

Procedure TTerritoryList.ReadTerritory(Reader: TReader);
Var
  Code2: String;
Begin
  BeginUpdate;
  Try
    Clear;
    Reader.ReadListBegin;
    While Not Reader.EndOfList Do
    Begin
      Code2 := Reader.ReadString;
      If Code2 = '*' Then
      Begin
        Assign(World.Territories);
        Break;
      End;
      Add(Code2);
    End;
    Reader.ReadListEnd;
  Finally
    EndUpdate;
  End;
End;

Procedure TTerritoryList.WriteTerritory(Writer: TWriter);
Var
  I: Integer;
Begin
  Writer.WriteListBegin;
  If (Duplicates <> DupAccept) And (Count = World.Territories.Count) Then
    Writer.WriteString('*') // all territories
  Else
    For I := 0 To Count - 1 Do
      Writer.WriteString(Items[I].Code2);
  Writer.WriteListEnd;
End;

Procedure TTerritoryList.DefineProperties(Filer: TFiler);
Begin
  Inherited DefineProperties(Filer);
  Filer.DefineProperty('Territories', ReadTerritory, WriteTerritory, Count <> 0);
End;

Procedure TTerritoryList.Changing;
Begin
  If (UpdateCount = 0) And Assigned(OnChanging) Then
    OnChanging(Self);
End;

Procedure TTerritoryList.Change;
Begin
  If (UpdateCount = 0) And Assigned(OnChange) Then
    OnChange(Self);
End;

Procedure TTerritoryList.BeginUpdate;
Begin
  If UpdateCount = 0 Then
    Changing;
  Inc(FUpdateCount);
End;

Procedure TTerritoryList.EndUpdate;
Begin
  Dec(FUpdateCount);
  If UpdateCount = 0 Then
    Change;
End;

Procedure TTerritoryList.Clear;
Begin
  If Count <> 0 Then
  Begin
    Changing;
    Try
      List.Clear;
      ClearLookupTables;
    Finally
      Change;
    End;
  End;
End;

Procedure TTerritoryList.Sort(OrderBy: TTerritoryDisplayName);

  Procedure QuickSort(L, R: Integer);
  Var
    I, J, P: Integer;
    Pivot: String;
  Begin
    Repeat
      I := L;
      J := R;
      P := (L + R) Shr 1;
      Repeat
        Pivot := Items[P].DisplayNames[OrderBy];
        While AnsiCompareText(Items[I].DisplayNames[OrderBy], Pivot) < 0 Do
          Inc(I);
        While AnsiCompareText(Items[J].DisplayNames[OrderBy], Pivot) > 0 Do
          Dec(J);
        If I <= J Then
        Begin
          If I <> J Then
          Begin
            List.Exchange(I, J);
            InvalidateLookupTables;
          End;
          If P = I Then
            P := J
          Else If P = J Then
            P := I;
          Inc(I);
          Dec(J);
        End;
      Until I > J;
      If L < J Then
        QuickSort(L, J);
      L := I;
    Until I >= R;
  End;

Begin
  If (OrderBy <> TnNone) And (Count > 1) Then
  Begin
    Changing;
    Try
      QuickSort(0, Count - 1);
    Finally
      Change;
    End;
  End;
End;

Procedure TTerritoryList.Assign(Source: TPersistent);
Begin
  If Source Is TReadonlyTerritoryList Then
    Apply(TReadonlyTerritoryList(Source), LaCopy)
  Else
    Inherited Assign(Source);
End;

Procedure TTerritoryList.Apply(ListA: TReadonlyTerritoryList; Op: TListAssignOp; ListB: TReadonlyTerritoryList);
Begin
  Changing;
  Try
    If Assigned(ListB) Then
      List.Assign(ListA.List, Op, ListB.List)
    Else
      List.Assign(ListA.List, Op, Nil);
    InvalidateLookupTables;
  Finally
    Change;
  End;
End;

Function TTerritoryList.Add(Territory: TTerritoryInfo): Integer;
Begin
  If Assigned(Territory) Then
  Begin
    If Duplicates <> DupAccept Then
    Begin
      Result := IndexOf(Territory);
      If Result >= 0 Then
      Begin
        If Duplicates = DupError Then
          Raise EListError.CreateResFmt(@SDuplicateItemError, [Result]);
        Exit;
      End;
    End;
    Changing;
    Try
      Result := List.Add(Territory);
      AddLookupEntries(Territory, Result);
    Finally
      Change;
    End;
  End
  Else
    Result := -1;
End;

Function TTerritoryList.Add(GeoID: GEOID): Integer;
Begin
  Result := Add(World.Territories.Find(GeoID));
End;

Function TTerritoryList.Add(Const Code: String): Integer;
Begin
  Result := Add(World.Territories.Find(Code));
End;

Function TTerritoryList.Remove(Territory: TTerritoryInfo): Integer;
Begin
  Result := IndexOf(Territory);
  If Result >= 0 Then
    Delete(Result);
End;

Function TTerritoryList.Remove(GeoID: GEOID): Integer;
Begin
  Result := IndexOf(GeoID);
  If Result >= 0 Then
    Delete(Result);
End;

Function TTerritoryList.Remove(Const Code: String): Integer;
Begin
  Result := IndexOf(Code);
  If Result >= 0 Then
    Delete(Result);
End;

Procedure TTerritoryList.Delete(Index: Integer);
Begin
  Changing;
  Try
    RemoveLookupEntries(Items[Index]);
    List.Delete(Index);
  Finally
    Change;
  End;
End;

{ TCurrency }

Constructor TCurrencyInfo.Create(LocaleID: LCID);
Begin
  If Not World.Changing Then
    Raise Exception.CreateResFmt(@SInstanceCreateError, [ClassName]);
  FCultures := TCultureList.Create;
  FCountries := TTerritoryList.Create;
  FIntlSymbol := GetLocaleStr(LocaleID, LOCALE_SINTLSYMBOL, '');
  FLocalSymbol := GetLocaleStr(LocaleID, LOCALE_SCURRENCY, '');
  FEnglishName := GetLocaleStr(LocaleID, LOCALE_SENGCURRNAME, '');
End;

Constructor TCurrencyInfo.Create(Const Locale: String);
Begin
  Create(LocaleNameToLCID(Locale));
End;

Destructor TCurrencyInfo.Destroy;
Begin
  If Not World.Changing Then
    Raise Exception.CreateResFmt(@SInstanceDestroyError, [ClassName]);
  FCultures.Free;
  FCountries.Free;
  Inherited Destroy;
End;

Function TCurrencyInfo.GetDisplayNames(DisplayName: TCurrencyDisplayName): String;
Begin
  Case DisplayName Of
    CrnIntlSymbol:
      Result := IntlSymbol;
    CrnEnglishName:
      Result := EnglishName;
    CrnNativeName:
      Result := NativeName;
  Else
    Result := '';
  End;
End;

Function TCurrencyInfo.GetNativeName: String;
Begin
  If Assigned(OriginCountry) Then
    Result := OriginCountry.Cultures[0].NativeCurrencyName
  Else
    Result := '';
End;

Function TCurrencyInfo.GetCultures: TReadonlyCultureList;
Begin
  Result := FCultures;
End;

Function TCurrencyInfo.GetCountries: TReadonlyTerritoryList;
Begin
  Result := FCountries;
End;

Procedure TCurrencyInfo.Associate(Culture: TCultureInfo; Territory: TTerritoryInfo);
Begin
  If Assigned(Culture) And Not FCultures.Exists(Culture) Then
    FCultures.Add(Culture);
  If Assigned(Territory) And Not FCountries.Exists(Territory) Then
  Begin
    FCountries.Add(Territory);
    If Assigned(FOriginCountry) And (Pos(FOriginCountry.Code2, IntlSymbol) = 0) Then
      FOriginCountry := Nil;
    If (FCountries.Count = 1) Or (Pos(Territory.Code2, IntlSymbol) <> 0) Then
      FOriginCountry := Territory;
  End;
End;

{ TCurrencyEnumerator }

Constructor TCurrencyEnumerator.Create(List: TReadonlyCurrencyList);
Begin
  Inherited Create(List.List);
End;

Function TCurrencyEnumerator.GetCurrent: TCurrencyInfo;
Begin
  Result := TCurrencyInfo( Inherited GetCurrent);
End;

{ TReadonlyCurrencyList }

Constructor TReadonlyCurrencyList.Create;
Begin
  List := TList.Create;
End;

Destructor TReadonlyCurrencyList.Destroy;
Begin
  List.Free;
  If Assigned(LookupTable_IntlSymbol) Then
    LookupTable_IntlSymbol.Free;
  Inherited Destroy;
End;

Function TReadonlyCurrencyList.GetCount: Integer;
Begin
  Result := List.Count;
End;

Function TReadonlyCurrencyList.GetItems(Index: Integer): TCurrencyInfo;
Begin
  Result := TCurrencyInfo(List[Index]);
End;

Function TReadonlyCurrencyList.GetEnumerator: TCurrencyEnumerator;
Begin
  Result := TCurrencyEnumerator.Create(Self);
End;

Procedure TReadonlyCurrencyList.AddLookupEntries(Currency: TCurrencyInfo; Index: Integer);
Begin
  If IsLookupTableValid_IntlSymbol Then
    LookupTable_IntlSymbol.Add(Currency.IntlSymbol, Index);
End;

Procedure TReadonlyCurrencyList.RemoveLookupEntries(Currency: TCurrencyInfo);
Begin
  If IsLookupTableValid_IntlSymbol Then
    LookupTable_IntlSymbol.Remove(Currency.IntlSymbol);
End;

Procedure TReadonlyCurrencyList.ClearLookupTables;
Begin
  If Assigned(LookupTable_IntlSymbol) Then
  Begin
    LookupTable_IntlSymbol.Clear;
    IsLookupTableValid_IntlSymbol := False;
  End;
End;

Procedure TReadonlyCurrencyList.InvalidateLookupTables;
Begin
  IsLookupTableValid_IntlSymbol := False;
End;

Procedure TReadonlyCurrencyList.ValidateLookupTable_IntlSymbol;
Var
  I: Integer;
Begin
  If Not IsLookupTableValid_IntlSymbol Then
  Begin
    If Assigned(LookupTable_IntlSymbol) Then
      LookupTable_IntlSymbol.Clear
    Else
      LookupTable_IntlSymbol := TKeyLookup<String, Integer>.Create;
    For I := 0 To Count - 1 Do
      LookupTable_IntlSymbol.Add(Items[I].IntlSymbol, I);
    IsLookupTableValid_IntlSymbol := True;
  End;
End;

Function TReadonlyCurrencyList.First: TCurrencyInfo;
Begin
  If Count <> 0 Then
    Result := Items[0]
  Else
    Result := Nil;
End;

Function TReadonlyCurrencyList.Last: TCurrencyInfo;
Begin
  If Count <> 0 Then
    Result := Items[Count - 1]
  Else
    Result := Nil;
End;

Procedure TReadonlyCurrencyList.AssignTo(Dest: TPersistent);
Var
  Strings: TStrings Absolute Dest;
  I: Integer;
Begin
  If Dest Is TStrings Then
  Begin
    Strings.BeginUpdate;
    Try
      Strings.Clear;
      For I := 0 To Count - 1 Do
        Strings.AddObject(Items[I].IntlSymbol, Items[I]);
    Finally
      Strings.EndUpdate;
    End;
  End
  Else
    Inherited AssignTo(Dest);
End;

Procedure TReadonlyCurrencyList.AssignLabelsTo(Dest: TStrings; DisplayName: TCurrencyDisplayName);
Var
  I: Integer;
Begin
  Dest.BeginUpdate;
  Try
    Dest.Clear;
    For I := 0 To Count - 1 Do
      Dest.AddObject(Items[I].DisplayNames[DisplayName], Items[I]);
  Finally
    Dest.EndUpdate;
  End;
End;

Function TReadonlyCurrencyList.Find(Const IntlSymbol: String): TCurrencyInfo;
Var
  Index: Integer;
Begin
  If Not IsLookupTableValid_IntlSymbol Then
    ValidateLookupTable_IntlSymbol;
  If LookupTable_IntlSymbol.Retrieve(UpperCase(IntlSymbol), Index) Then
    Result := Items[Index]
  Else
    Result := Nil;
End;

Function TReadonlyCurrencyList.IndexOf(Currency: TCurrencyInfo): Integer;
Begin
  If Assigned(Currency) Then
    Result := IndexOf(Currency.IntlSymbol)
  Else
    Result := -1;
End;

Function TReadonlyCurrencyList.IndexOf(Const IntlSymbol: String): Integer;
Begin
  If Not IsLookupTableValid_IntlSymbol Then
    ValidateLookupTable_IntlSymbol;
  If Not LookupTable_IntlSymbol.Retrieve(UpperCase(IntlSymbol), Result) Then
    Result := -1;
End;

Function TReadonlyCurrencyList.IndexOfName(Const AName: String; DisplayName: TCurrencyDisplayName): Integer;
Var
  I: Integer;
Begin
  For I := 0 To Count - 1 Do
    If AnsiSameText(Items[I].DisplayNames[DisplayName], AName) Then
    Begin
      Result := I;
      Exit;
    End;
  Result := -1;
End;

Function TReadonlyCurrencyList.Exists(Currency: TCurrencyInfo): Boolean;
Begin
  Result := Assigned(Currency) And Exists(Currency.IntlSymbol);
End;

Function TReadonlyCurrencyList.Exists(Const IntlSymbol: String): Boolean;
Begin
  If Not IsLookupTableValid_IntlSymbol Then
    ValidateLookupTable_IntlSymbol;
  Result := LookupTable_IntlSymbol.Exists(UpperCase(IntlSymbol));
End;

{ TCurrencyList }

Function TCurrencyList.GetCapacity: Integer;
Begin
  Result := List.Capacity;
End;

Procedure TCurrencyList.SetCapacity(Value: Integer);
Begin
  List.Capacity := Value;
  If Value < Count Then
    InvalidateLookupTables;
End;

Procedure TCurrencyList.ReadCurrencies(Reader: TReader);
Var
  IntlSymbol: String;
Begin
  BeginUpdate;
  Try
    Clear;
    Reader.ReadListBegin;
    While Not Reader.EndOfList Do
    Begin
      IntlSymbol := Reader.ReadString;
      If IntlSymbol = '*' Then
      Begin
        Assign(World.Currencies);
        Break;
      End;
      Add(IntlSymbol);
    End;
    Reader.ReadListEnd;
  Finally
    EndUpdate;
  End;
End;

Procedure TCurrencyList.WriteCurrencies(Writer: TWriter);
Var
  I: Integer;
Begin
  Writer.WriteListBegin;
  If (Duplicates <> DupAccept) And (Count = World.Currencies.Count) Then
    Writer.WriteString('*') // all currencies
  Else
    For I := 0 To Count - 1 Do
      Writer.WriteString(Items[I].IntlSymbol);
  Writer.WriteListEnd;
End;

Procedure TCurrencyList.DefineProperties(Filer: TFiler);
Begin
  Inherited DefineProperties(Filer);
  Filer.DefineProperty('Currencies', ReadCurrencies, WriteCurrencies, Count <> 0);
End;

Procedure TCurrencyList.Changing;
Begin
  If (UpdateCount = 0) And Assigned(OnChanging) Then
    OnChanging(Self);
End;

Procedure TCurrencyList.Change;
Begin
  If (UpdateCount = 0) And Assigned(OnChange) Then
    OnChange(Self);
End;

Procedure TCurrencyList.BeginUpdate;
Begin
  If UpdateCount = 0 Then
    Changing;
  Inc(FUpdateCount);
End;

Procedure TCurrencyList.EndUpdate;
Begin
  Dec(FUpdateCount);
  If UpdateCount = 0 Then
    Change;
End;

Procedure TCurrencyList.Clear;
Begin
  If Count <> 0 Then
  Begin
    Changing;
    Try
      List.Clear;
      ClearLookupTables;
    Finally
      Change;
    End;
  End;
End;

Procedure TCurrencyList.Sort(OrderBy: TCurrencyDisplayName);

  Procedure QuickSort(L, R: Integer);
  Var
    I, J, P: Integer;
    Pivot: String;
  Begin
    Repeat
      I := L;
      J := R;
      P := (L + R) Shr 1;
      Repeat
        Pivot := Items[P].DisplayNames[OrderBy];
        While AnsiCompareText(Items[I].DisplayNames[OrderBy], Pivot) < 0 Do
          Inc(I);
        While AnsiCompareText(Items[J].DisplayNames[OrderBy], Pivot) > 0 Do
          Dec(J);
        If I <= J Then
        Begin
          If I <> J Then
          Begin
            List.Exchange(I, J);
            InvalidateLookupTables;
          End;
          If P = I Then
            P := J
          Else If P = J Then
            P := I;
          Inc(I);
          Dec(J);
        End;
      Until I > J;
      If L < J Then
        QuickSort(L, J);
      L := I;
    Until I >= R;
  End;

Begin
  If (OrderBy <> CrnNone) And (Count > 1) Then
  Begin
    Changing;
    Try
      QuickSort(0, Count - 1);
    Finally
      Change;
    End;
  End;
End;

Procedure TCurrencyList.Assign(Source: TPersistent);
Begin
  If Source Is TReadonlyCurrencyList Then
    Apply(TReadonlyCurrencyList(Source), LaCopy)
  Else
    Inherited Assign(Source);
End;

Procedure TCurrencyList.Apply(ListA: TReadonlyCurrencyList; Op: TListAssignOp; ListB: TReadonlyCurrencyList);
Begin
  Changing;
  Try
    If Assigned(ListB) Then
      List.Assign(ListA.List, Op, ListB.List)
    Else
      List.Assign(ListA.List, Op, Nil);
    InvalidateLookupTables;
  Finally
    Change;
  End;
End;

Function TCurrencyList.Add(Currency: TCurrencyInfo): Integer;
Begin
  If Assigned(Currency) Then
  Begin
    If Duplicates <> DupAccept Then
    Begin
      Result := IndexOf(Currency);
      If Result >= 0 Then
      Begin
        If Duplicates = DupError Then
          Raise EListError.CreateResFmt(@SDuplicateItemError, [Result]);
        Exit;
      End;
    End;
    Changing;
    Try
      Result := List.Add(Currency);
      AddLookupEntries(Currency, Result);
    Finally
      Change;
    End;
  End
  Else
    Result := -1;
End;

Function TCurrencyList.Add(Const IntlSymbol: String): Integer;
Begin
  Result := Add(World.Currencies.Find(IntlSymbol));
End;

Function TCurrencyList.Remove(Currency: TCurrencyInfo): Integer;
Begin
  Result := IndexOf(Currency);
  If Result >= 0 Then
    Delete(Result);
End;

Function TCurrencyList.Remove(Const IntlSymbol: String): Integer;
Begin
  Result := IndexOf(IntlSymbol);
  If Result >= 0 Then
    Delete(Result);
End;

Procedure TCurrencyList.Delete(Index: Integer);
Begin
  Changing;
  Try
    RemoveLookupEntries(Items[Index]);
    List.Delete(Index);
  Finally
    Change;
  End;
End;

{ World }

Function LocaleEnumCallbackEx(LpLocaleString: PChar; DwFlags: DWORD; LParam: LPARAM): Integer; Stdcall;
Begin
  World.CultureList.Add(TCultureInfo.Create(LpLocaleString));
  Result := 1;
End;

Function LocaleEnumCallback(LpLocaleString: PChar): Integer; Stdcall;
Var
  LocaleID: LCID;
Begin
  LocaleID := StrToIntDef('$' + String(LpLocaleString), 0);
  If LocaleID <> 0 Then
    World.CultureList.Add(TCultureInfo.Create(LocaleID));
  Result := 1;
End;

Function GeoEnumCallBack(GeoID: GEOID): Integer; Stdcall;
Begin
  If Not World.TerritoryList.Exists(GeoID) Then
  Begin
{$IFDEF EXCLUDE_TERRITORIES_WITH_X_CODE}
    If Pos('X', GetGeoStr(GeoID, GEO_ISO3, 'X')) <> 1 Then
{$ENDIF}
      World.TerritoryList.Add(TTerritoryInfo.Create(GeoID));
  End;
  Result := 1;
End;

Class Procedure World.Startup;
Begin
  CultureList := TCultureList.Create;
  TerritoryList := TTerritoryList.Create;
  CurrencyList := TCurrencyList.Create;
  CultureList.Capacity := 300;
  TerritoryList.Capacity := 300;
  CurrencyList.Capacity := 300;
  Changing := True;
  Try
    EnumSystemLocales(@LocaleEnumCallback, LCID_SUPPORTED);
{$IFNDEF EXCLUDE_TERRITORIES_WITH_NO_CULTURE_INFO}
    EnumSystemGeoID(GEOCLASS_NATION, 0, @GeoEnumCallBack);
{$ENDIF}
  Finally
    Changing := False;
  End;
  CultureList.List.Capacity := CultureList.Count;
  TerritoryList.List.Capacity := TerritoryList.Count;
  CurrencyList.List.Capacity := CurrencyList.Count;
End;

Class Procedure World.Cleanup;
Var
  I: Integer;
Begin
  Changing := True;
  Try
    If Assigned(CultureList) Then
    Begin
      For I := 0 To CultureList.Count - 1 Do
        CultureList[I].Free;
      FreeAndNil(CultureList);
    End;
    If Assigned(TerritoryList) Then
    Begin
      For I := 0 To TerritoryList.Count - 1 Do
        TerritoryList[I].Free;
      FreeAndNil(TerritoryList);
    End;
    If Assigned(CurrencyList) Then
    Begin
      For I := 0 To CurrencyList.Count - 1 Do
        CurrencyList[I].Free;
      FreeAndNil(CurrencyList);
    End;
  Finally
    Changing := False;
  End;
End;

Class Function World.Cultures: TReadonlyCultureList;
Begin
  If Not Assigned(CultureList) Then
    Startup;
  Result := CultureList;
End;

Class Function World.Territories: TReadonlyTerritoryList;
Begin
  If Not Assigned(TerritoryList) Then
    Startup;
  Result := TerritoryList;
End;

Class Function World.Currencies: TReadonlyCurrencyList;
Begin
  If Not Assigned(CultureList) Then
    Startup;
  Result := CurrencyList;
End;

Initialization

Finalization

If Assigned(TheDefaultCalendar) Then
  FreeAndNil(TheDefaultCalendar);
World.Cleanup;

End.
