{ ------------------------------------------------------------------------------ }
{ }
{ i18n Package }
{ by Kambiz R. Khojasteh }
{ }
{ kambiz@delphiarea.com }
{ http://www.delphiarea.com }
{ }
{ ------------------------------------------------------------------------------ }

/// This unit implements MD5 hashing algorithm.
Unit i18nMD5;

{$I DELPHIAREA.INC}

Interface

Uses
  Classes, SysUtils, Types;

Type

{$REGION 'xmldoc'}
  /// <summary>
  /// EMD5Error is the exception class of the errors that occur in the instances of
  /// <see cref="TMD5"/> class.</summary>
  /// <remarks>
  /// EMD5Error is raised when an error is made in a <see cref="TMD5"/> object. This
  /// exception commonly occurs when an application appends new data to a <see cref="TMD5"/>
  /// object before reseting it.</remarks>
  /// <seealso cref="TMD5"/>
{$ENDREGION}
  EMD5Error = Class(Exception);

{$REGION 'xmldoc'}
  /// <summary>
  /// TMD5 calculates MD5 hash value of data from varius data sources.</summary>
{$ENDREGION}

  TMD5 = Class(TObject)
  Private
    Count: Cardinal; { message length in bytes }
    Digest: Array [0 .. 3] Of Cardinal; { digest buffer }
    Buffer: Array [0 .. 63] Of Byte; { accumulate block }
    Appending: Boolean;
    Function GetValue: String;
    Procedure Process(Const Block);
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Create an instance of TMD5 class.</summary>
{$ENDREGION}
    Constructor Create;
{$REGION 'xmldoc'}
    /// <summary>
    /// Resets the object for appending new set of data.</summary>
{$ENDREGION}
    Procedure Reset;
{$REGION 'xmldoc'}
    /// <summary>
    /// Appends a specified number of bytes to the MD5 buffer.</summary>
    /// <param name="Data">
    /// The data to append.</param>
    /// <param name="Size">
    /// The number of bytes of <paramref name="Data"/> to append.</param>
    /// <seealso cref="AppendString"/>
    /// <seealso cref="AppendStream"/>
    /// <seealso cref="AppendFile"/>
{$ENDREGION}
    Procedure Append(Const Data; Size: Cardinal);
{$REGION 'xmldoc'}
    /// <summary>
    /// Appends characters of a specified string to the MD5 buffer.</summary>
    /// <param name="Str">
    /// The string to append.</param>
    /// <seealso cref="Append"/>
    /// <seealso cref="AppendStream"/>
    /// <seealso cref="AppendFile"/>
{$ENDREGION}
    Procedure AppendString(Const Str: String); Overload; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Appends characters of a specified ansi string to the MD5 buffer.</summary>
    /// <param name="Str">
    /// The string to append.</param>
    /// <seealso cref="Append"/>
    /// <seealso cref="AppendStream"/>
    /// <seealso cref="AppendFile"/>
{$ENDREGION}
    Procedure AppendString(Const Str: AnsiString); Overload; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Appends characters of a specified wide string to the MD5 buffer.</summary>
    /// <param name="Str">
    /// The string to append.</param>
    /// <seealso cref="Append"/>
    /// <seealso cref="AppendStream"/>
    /// <seealso cref="AppendFile"/>
{$ENDREGION}
    Procedure AppendString(Const Str: WideString); Overload; Inline;
{$REGION 'xmldoc'}
    /// <summary>
    /// Appends content of a specified stream to the MD5 buffer.</summary>
    /// <param name="Stream">
    /// The stream to append.</param>
    /// <seealso cref="Append"/>
    /// <seealso cref="AppendString"/>
    /// <seealso cref="AppendFile"/>
{$ENDREGION}
    Procedure AppendStream(Stream: TStream);
{$REGION 'xmldoc'}
    /// <summary>
    /// Appends content of a specified file to the MD5 buffer.</summary>
    /// <param name="FileName">
    /// The path to the file to append.</param>
    /// <seealso cref="Append"/>
    /// <seealso cref="AppendString"/>
    /// <seealso cref="AppendStream"/>
{$ENDREGION}
    Procedure AppendFile(Const FileName: String);
{$REGION 'xmldoc'}
    /// <summary>
    /// Gets the 128-bit MD5 hash value of the previously appended data as a
    /// hexadecimal string.</summary>
{$ENDREGION}
    Property Value: String Read GetValue;
  End;

{$REGION 'xmldoc'}
  /// <summary>
  /// Calculates MD5 hash value of a specified number of bytes.</summary>
  /// <param name="Data">
  /// The data to calculate its hash value.</param>
  /// <param name="Size">
  /// The number of bytes of <paramref name="Data"/>.</param>
  /// <returns>
  /// The hash value as a hexadecimal string.</returns>
  /// <seealso cref="FileMD5"/>
{$ENDREGION}

Function MD5(Const Data; Size: Integer): String; Overload;

{$REGION 'xmldoc'}
/// <summary>
/// Calculates MD5 hash value of a specified string.</summary>
/// <param name="Str">
/// The string to calculate its hash value.</param>
/// <returns>
/// The hash value as a hexadecimal string.</returns>
/// <seealso cref="FileMD5"/>
{$ENDREGION}
Function MD5(Const Str: String): String; Overload;

{$REGION 'xmldoc'}
/// <summary>
/// Calculates MD5 hash value of a specified ansi string.</summary>
/// <param name="Str">
/// The string to calculate its hash value.</param>
/// <returns>
/// The hash value as a hexadecimal string.</returns>
/// <seealso cref="FileMD5"/>
{$ENDREGION}
Function MD5(Const Str: AnsiString): String; Overload;

{$REGION 'xmldoc'}
/// <summary>
/// Calculates MD5 hash value of a specified wide string.</summary>
/// <param name="Str">
/// The string to calculate its hash value.</param>
/// <returns>
/// The hash value as a hexadecimal string.</returns>
/// <seealso cref="FileMD5"/>
{$ENDREGION}
Function MD5(Const Str: WideString): String; Overload;

{$REGION 'xmldoc'}
/// <summary>
/// Calculates MD5 hash value of a specified stream.</summary>
/// <param name="Stream">
/// The stream to calculate hash value of its content.</param>
/// <returns>
/// The hash value as a hexadecimal string.</returns>
/// <seealso cref="FileMD5"/>
{$ENDREGION}
Function MD5(Stream: TStream): String; Overload;

{$REGION 'xmldoc'}
/// <summary>
/// Calculates MD5 hash value of a specified file.</summary>
/// <param name="FileName">
/// The path to the file that hash value of its content should be calculated.</param>
/// <returns>
/// The hash value as a hexadecimal string.</returns>
/// <seealso cref="MD5"/>
{$ENDREGION}
Function FileMD5(Const FileName: String): String;

Implementation

Resourcestring
  SResetRequired = 'You must reset the MD5 object before appending new data';

  { Helper Functions }

Function MD5(Const Data; Size: Integer): String;
Begin
  With TMD5.Create Do
    Try
      Append(Data, Size);
      Result := Value;
    Finally
      Free;
    End;
End;

Function MD5(Const Str: String): String;
Begin
  With TMD5.Create Do
    Try
      AppendString(Str);
      Result := Value;
    Finally
      Free;
    End;
End;

Function MD5(Const Str: AnsiString): String;
Begin
  With TMD5.Create Do
    Try
      AppendString(Str);
      Result := Value;
    Finally
      Free;
    End;
End;

Function MD5(Const Str: WideString): String;
Begin
  With TMD5.Create Do
    Try
      AppendString(Str);
      Result := Value;
    Finally
      Free;
    End;
End;

Function MD5(Stream: TStream): String;
Begin
  With TMD5.Create Do
    Try
      AppendStream(Stream);
      Result := Value;
    Finally
      Free;
    End;
End;

Function FileMD5(Const FileName: String): String;
Begin
  With TMD5.Create Do
    Try
      AppendFile(FileName);
      Result := Value;
    Finally
      Free;
    End;
End;

{ TMD5 }

Const
  T: Array [1 .. 64] Of Cardinal = ($D76AA478, $E8C7B756, $242070DB, $C1BDCEEE, $F57C0FAF, $4787C62A, $A8304613,
    $FD469501, $698098D8, $8B44F7AF, $FFFF5BB1, $895CD7BE, $6B901122, $FD987193, $A679438E, $49B40821,
    $F61E2562, $C040B340, $265E5A51, $E9B6C7AA, $D62F105D, $02441453, $D8A1E681, $E7D3FBC8, $21E1CDE6,
    $C33707D6, $F4D50D87, $455A14ED, $A9E3E905, $FCEFA3F8, $676F02D9, $8D2A4C8A, $FFFA3942, $8771F681,
    $6D9D6122, $FDE5380C, $A4BEEA44, $4BDECFA9, $F6BB4B60, $BEBFBC70, $289B7EC6, $EAA127FA, $D4EF3085,
    $04881D05, $D9D4D039, $E6DB99E5, $1FA27CF8, $C4AC5665, $F4292244, $432AFF97, $AB9423A7, $FC93A039,
    $655B59C3, $8F0CCC92, $FFEFF47D, $85845DD1, $6FA87E4F, $FE2CE6E0, $A3014314, $4E0811A1, $F7537E82,
    $BD3AF235, $2AD7D2BB, $EB86D391);

Type
  TBlock = Array [0 .. 15] Of Cardinal;

Procedure TMD5.Process(Const Block);

  Procedure Set1(Const X: TBlock; Var A, B, C, D: Cardinal; K, S, I: Cardinal); Inline;
  Var
    N: Cardinal;
  Begin
    N := A + X[K] + T[I] + ((B And C) Or (Not B And D));
    A := ((N Shl S) Or (N Shr (32 - S))) + B;
  End;

  Procedure Set2(Const X: TBlock; Var A, B, C, D: Cardinal; K, S, I: Cardinal); Inline;
  Var
    N: Cardinal;
  Begin
    N := A + X[K] + T[I] + ((B And D) Or (C And Not D));
    A := ((N Shl S) Or (N Shr (32 - S))) + B;
  End;

  Procedure Set3(Const X: TBlock; Var A, B, C, D: Cardinal; K, S, I: Cardinal); Inline;
  Var
    N: Cardinal;
  Begin
    N := A + X[K] + T[I] + (B Xor C Xor D);
    A := ((N Shl S) Or (N Shr (32 - S))) + B;
  End;

  Procedure Set4(Const X: TBlock; Var A, B, C, D: Cardinal; K, S, I: Cardinal); Inline;
  Var
    N: Cardinal;
  Begin
    N := A + X[K] + T[I] + (C Xor (B Or Not D));
    A := ((N Shl S) Or (N Shr (32 - S))) + B;
  End;

Var
  X: TBlock Absolute Block;
  A, B, C, D: Cardinal;
Begin
  A := Digest[0];
  B := Digest[1];
  C := Digest[2];
  D := Digest[3];
  Set1(X, A, B, C, D, 0, 7, 1);
  Set1(X, D, A, B, C, 1, 12, 2);
  Set1(X, C, D, A, B, 2, 17, 3);
  Set1(X, B, C, D, A, 3, 22, 4);
  Set1(X, A, B, C, D, 4, 7, 5);
  Set1(X, D, A, B, C, 5, 12, 6);
  Set1(X, C, D, A, B, 6, 17, 7);
  Set1(X, B, C, D, A, 7, 22, 8);
  Set1(X, A, B, C, D, 8, 7, 9);
  Set1(X, D, A, B, C, 9, 12, 10);
  Set1(X, C, D, A, B, 10, 17, 11);
  Set1(X, B, C, D, A, 11, 22, 12);
  Set1(X, A, B, C, D, 12, 7, 13);
  Set1(X, D, A, B, C, 13, 12, 14);
  Set1(X, C, D, A, B, 14, 17, 15);
  Set1(X, B, C, D, A, 15, 22, 16);
  Set2(X, A, B, C, D, 1, 5, 17);
  Set2(X, D, A, B, C, 6, 9, 18);
  Set2(X, C, D, A, B, 11, 14, 19);
  Set2(X, B, C, D, A, 0, 20, 20);
  Set2(X, A, B, C, D, 5, 5, 21);
  Set2(X, D, A, B, C, 10, 9, 22);
  Set2(X, C, D, A, B, 15, 14, 23);
  Set2(X, B, C, D, A, 4, 20, 24);
  Set2(X, A, B, C, D, 9, 5, 25);
  Set2(X, D, A, B, C, 14, 9, 26);
  Set2(X, C, D, A, B, 3, 14, 27);
  Set2(X, B, C, D, A, 8, 20, 28);
  Set2(X, A, B, C, D, 13, 5, 29);
  Set2(X, D, A, B, C, 2, 9, 30);
  Set2(X, C, D, A, B, 7, 14, 31);
  Set2(X, B, C, D, A, 12, 20, 32);
  Set3(X, A, B, C, D, 5, 4, 33);
  Set3(X, D, A, B, C, 8, 11, 34);
  Set3(X, C, D, A, B, 11, 16, 35);
  Set3(X, B, C, D, A, 14, 23, 36);
  Set3(X, A, B, C, D, 1, 4, 37);
  Set3(X, D, A, B, C, 4, 11, 38);
  Set3(X, C, D, A, B, 7, 16, 39);
  Set3(X, B, C, D, A, 10, 23, 40);
  Set3(X, A, B, C, D, 13, 4, 41);
  Set3(X, D, A, B, C, 0, 11, 42);
  Set3(X, C, D, A, B, 3, 16, 43);
  Set3(X, B, C, D, A, 6, 23, 44);
  Set3(X, A, B, C, D, 9, 4, 45);
  Set3(X, D, A, B, C, 12, 11, 46);
  Set3(X, C, D, A, B, 15, 16, 47);
  Set3(X, B, C, D, A, 2, 23, 48);
  Set4(X, A, B, C, D, 0, 6, 49);
  Set4(X, D, A, B, C, 7, 10, 50);
  Set4(X, C, D, A, B, 14, 15, 51);
  Set4(X, B, C, D, A, 5, 21, 52);
  Set4(X, A, B, C, D, 12, 6, 53);
  Set4(X, D, A, B, C, 3, 10, 54);
  Set4(X, C, D, A, B, 10, 15, 55);
  Set4(X, B, C, D, A, 1, 21, 56);
  Set4(X, A, B, C, D, 8, 6, 57);
  Set4(X, D, A, B, C, 15, 10, 58);
  Set4(X, C, D, A, B, 6, 15, 59);
  Set4(X, B, C, D, A, 13, 21, 60);
  Set4(X, A, B, C, D, 4, 6, 61);
  Set4(X, D, A, B, C, 11, 10, 62);
  Set4(X, C, D, A, B, 2, 15, 63);
  Set4(X, B, C, D, A, 9, 21, 64);
  Inc(Digest[0], A);
  Inc(Digest[1], B);
  Inc(Digest[2], C);
  Inc(Digest[3], D);
End;

Procedure TMD5.Append(Const Data; Size: Cardinal);
Var
  P: PByte;
  Offset: Cardinal;
  LeftBytes: Cardinal;
  CopyBytes: Cardinal;
Begin
  If Not Appending Then
    Raise EMD5Error.Create(SResetRequired);
  If Size = 0 Then
    Exit;
  P := @Data;
  LeftBytes := Size;
  Offset := Count And 63;
  // Update the message length
  Inc(Count, Size);
  // Process an initial partial block
  If Offset <> 0 Then
  Begin
    If Offset + Size > 64 Then
      CopyBytes := 64 - Offset
    Else
      CopyBytes := Size;
    Move(P^, Buffer[Offset], CopyBytes);
    If Offset + CopyBytes < 64 Then
      Exit;
    Inc(P, CopyBytes);
    Dec(LeftBytes, CopyBytes);
    Process(Buffer[0]);
  End;
  // Process full blocks
  While LeftBytes >= 64 Do
  Begin
    Process(P^);
    Inc(P, 64);
    Dec(LeftBytes, 64);
  End;
  // Save final partial block for future process
  If LeftBytes > 0 Then
    Move(P^, Buffer[0], LeftBytes);
End;

Procedure TMD5.AppendString(Const Str: String);
Begin
{$IFDEF UNICODE}
  AppendString(WideString(Str));
{$ELSE}
  AppendString(AnsiString(Str));
{$ENDIF}
End;

Procedure TMD5.AppendString(Const Str: AnsiString);
Begin
  Append(PAnsiChar(Str)^, Length(Str) * SizeOf(AnsiChar));
End;

Procedure TMD5.AppendString(Const Str: WideString);
Var
  Utf8Str: AnsiString;
Begin
  Utf8Str := UTF8Encode(Str);
  AppendString(Utf8Str);
End;

Procedure TMD5.AppendStream(Stream: TStream);
Var
  Bytes: Array [1 .. $FF00] Of Byte;
  ByteCount: Integer;
Begin
  ByteCount := Stream.Read(Bytes, SizeOf(Bytes));
  While ByteCount > 0 Do
  Begin
    Append(Bytes, ByteCount);
    ByteCount := Stream.Read(Bytes, SizeOf(Bytes));
  End;
End;

Procedure TMD5.AppendFile(Const FileName: String);
Var
  Stream: TStream;
Begin
  Stream := TFileStream.Create(FileName, FmOpenRead Or FmShareDenyWrite);
  Try
    AppendStream(Stream);
  Finally
    Stream.Free;
  End;
End;

Procedure TMD5.Reset;
Begin
  Count := 0;
  Digest[0] := $67452301;
  Digest[1] := $EFCDAB89;
  Digest[2] := $98BADCFE;
  Digest[3] := $10325476;
  Appending := True;
End;

Function TMD5.GetValue: String;
Const
  HexDigits: Array [0 .. 15] Of Char = '0123456789ABCDEF';
  Pad: Array [0 .. 63] Of Byte = ($80, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00,
    $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00,
    $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00,
    $00, $00, $00, $00, $00, $00, $00);
Var
  BitCount: Int64;
  S: PChar;
  B: PByte;
  I: Integer;
Begin
  If Appending Then
  Begin
    // Save the length before padding
    BitCount := Int64(Count) Shl 3;
    // Pad to 56 bytes mod 64
    Append(Pad, (55 - (Count And 63)) + 1);
    // Append the length
    Append(BitCount, 8);
    Appending := False;
  End;
  // convert 128-bit digest to string
  SetString(Result, Nil, 32);
  S := PChar(Result);
  B := Addr(Digest);
  For I := 0 To 15 Do
  Begin
    S^ := HexDigits[(B^ Shr 4) And 15];
    Inc(S);
    S^ := HexDigits[B^ And 15];
    Inc(S);
    Inc(B);
  End;
End;

Constructor TMD5.Create;
Begin
  Reset;
End;

End.
