{ ------------------------------------------------------------------------------ }
{ }
{ i18n Package }
{ by Kambiz R. Khojasteh }
{ }
{ kambiz@delphiarea.com }
{ http://www.delphiarea.com }
{ }
{ ------------------------------------------------------------------------------ }

/// This unit implements a class to deal with null character separated strings.
Unit i18nZStrList;

Interface

{$I DELPHIAREA.INC}

Uses
  SysUtils, Types;

Type

{$REGION 'xmldoc'}
  /// <summary>
  /// ZStrings provides a set of class methods to work with null character (#0)
  /// separated strings.</summary>
{$ENDREGION}
  ZStrings = Class
  Public
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates a list of null separated strings from a specified array of strings.</summary>
    /// <param name="StrArray">
    /// The source array of strings.</param>
    /// <returns>
    /// The list of null character separated strings.</returns>
{$ENDREGION}
    Class Function Construct(Const StrArray: Array Of String): String; Overload; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Creates a list of null separated strings from a specified string. The
    /// sub-strings of the source string are delimited by a specified delimiter
    /// string.</summary>
    /// <param name="Str">
    /// The source string.</param>
    /// <param name="Delimiter">
    /// The delimiter string that separates sub-strings of the source string.</param>
    /// <returns>
    /// The list of null character separated strings.</returns>
{$ENDREGION}
    Class Function Construct(Const Str: String; Const Delimiter: String): String; Overload; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Adds a sub-string at the end of a specified list of null character
    /// separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="SubStr">
    /// The value for the new sub-string.</param>
    /// <seealso cref="Insert"/>
{$ENDREGION}
    Class Procedure Add(Var ZStrs: String; Const SubStr: String); Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Inserts a sub-string at a specified index of a specified list of null character
    /// separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Index">
    /// The index where the sub-string must be inserted before.</param>
    /// <param name="SubStr">
    /// The value for the new sub-string.</param>
    /// <seealso cref="Add"/>
{$ENDREGION}
    Class Procedure Insert(Var ZStrs: String; Index: Integer; Const SubStr: String); Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Deletes a sub-string at a specified index of a specified list of null character
    /// separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Index">
    /// The index of the sub-string to delete.</param>
{$ENDREGION}
    Class Procedure Delete(Var ZStrs: String; Index: Integer); Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Replaces a sub-string at a specified index of a specified list of null character
    /// separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Index">
    /// The index of the sub-string to replace.</param>
    /// <param name="SubStr">
    /// The new value for the sub-string at the specified index.</param>
    /// <seealso cref="GetSubStrAt"/>
{$ENDREGION}
    Class Procedure SetSubStrAt(Var ZStrs: String; Index: Integer; Const SubStr: String); Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the sub-string at a specified index of a specified list of null
    /// character separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Index">
    /// The index of the sub-string to retrieve.</param>
    /// <returns>
    /// The sub-string at the specified index.</returns>
    /// <seealso cref="SetSubStrAt"/>
{$ENDREGION}
    Class Function GetSubStrAt(Const ZStrs: String; Index: Integer): String; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the number of sub-strings in a specified list of null character
    /// separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <returns>
    /// The number of sub-strings.</returns>
{$ENDREGION}
    Class Function Count(Const ZStrs: String): Integer; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Converts a specified list of null character separated strings to a list
    /// of delimited sub-strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Delimiter">
    /// The string that separates sub-strings.</param>
    /// <returns>
    /// The string of sub-strings delimited by given <paramref name="Delimiter"/>
    /// string.</returns>
{$ENDREGION}
    Class Function Split(Const ZStrs: String; Const Delimiter: String): String; Overload; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Converts a specified list of null character separated strings to an array
    /// of its sub-strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <returns>
    /// The array of sub-strings.</returns>
{$ENDREGION}
    Class Function Split(Const ZStrs: String): TStringDynArray; Overload; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the index of sub-string at a specified character position of a
    /// specified list of null character separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Pos">
    /// The character position withing the null character separated strings.</param>
    /// <returns>
    /// The zero-based index of sub-strings at the specified character position.
    /// If position given by <paramref name="Pos"/> is less than or equal to zero
    /// or <paramref name="ZStrs"/> is empty, the function returns zero.</returns>
{$ENDREGION}
    Class Function IndexFromCharPos(Const ZStrs: String; Pos: Integer): Integer; Static;
{$REGION 'xmldoc'}
    /// <summary>
    /// Returns the start position of a sub-string specified with its index within
    /// a specified list of null character separated strings.</summary>
    /// <param name="ZStrs">
    /// The null character separated strings.</param>
    /// <param name="Index">
    /// The index of the sub-string to find its start position.</param>
    /// <returns>
    /// The character position of the specified sub-strings within the specified
    /// null character separated strings.</returns>
{$ENDREGION}
    Class Function CharPosFromIndex(Const ZStrs: String; Index: Integer): Integer; Static;
  End;

Implementation

{ ZStrings }

Class Function ZStrings.Construct(Const StrArray: Array Of String): String;
Var
  I: Integer;
  Len: Integer;
  S: PChar;
Begin
  Len := Length(StrArray);
  For I := 0 To Length(StrArray) - 1 Do
    Inc(Len, Length(StrArray[I]));
  SetLength(Result, Len);
  S := PChar(Result);
  For I := 0 To Length(StrArray) - 1 Do
  Begin
    StrCopy(S, PChar(StrArray[I]));
    Inc(S, Length(StrArray[I]) + 1);
  End;
End;

Class Function ZStrings.Construct(Const Str, Delimiter: String): String;
Begin
  If Str = '' Then
    Result := ''
  Else
    Result := StringReplace(Str, Delimiter, #0, [RfReplaceAll]) + #0;
End;

Class Procedure ZStrings.Add(Var ZStrs: String; Const SubStr: String);
Begin
  ZStrs := ZStrs + SubStr + #0;
End;

Class Procedure ZStrings.Insert(Var ZStrs: String; Index: Integer; Const SubStr: String);
Var
  EndPtr, S: PChar;
Begin
  If Index <= 0 Then
    ZStrs := SubStr + #0 + ZStrs
  Else
  Begin
    S := PChar(ZStrs);
    EndPtr := S + Length(ZStrs);
    While (Index > 0) And (S < EndPtr) Do
    Begin
      S := StrEnd(S) + 1;
      Dec(Index);
      If Index = 0 Then
      Begin
        System.Insert(#0 + SubStr, ZStrs, S - PChar(ZStrs));
        Exit;
      End;
    End;
    ZStrs := ZStrs + SubStr + #0;
  End;
End;

Class Procedure ZStrings.Delete(Var ZStrs: String; Index: Integer);
Var
  EndPtr, S: PChar;
Begin
  If ZStrs <> '' Then
  Begin
    S := PChar(ZStrs);
    EndPtr := S + Length(ZStrs);
    While (Index > 0) And (S < EndPtr) Do
    Begin
      Dec(Index);
      S := StrEnd(S) + 1;
    End;
    If Index = 0 Then
      System.Delete(ZStrs, S - PChar(ZStrs) + 1, StrLen(S) + 1);
  End;
End;

Class Procedure ZStrings.SetSubStrAt(Var ZStrs: String; Index: Integer; Const SubStr: String);
Begin
  ZStrings.Delete(ZStrs, Index);
  ZStrings.Insert(ZStrs, Index, SubStr);
End;

Class Function ZStrings.GetSubStrAt(Const ZStrs: String; Index: Integer): String;
Var
  EndPtr, S: PChar;
Begin
  S := PChar(ZStrs);
  EndPtr := S + Length(ZStrs);
  While (Index > 0) And (S < EndPtr) Do
  Begin
    Dec(Index);
    S := StrEnd(S) + 1;
  End;
  SetString(Result, S, StrLen(S));
End;

Class Function ZStrings.Count(Const ZStrs: String): Integer;
Var
  EndPtr, S: PChar;
Begin
  Result := 0;
  S := PChar(ZStrs);
  EndPtr := S + Length(ZStrs);
  While S < EndPtr Do
  Begin
    Inc(Result);
    S := StrEnd(S) + 1;
  End;
End;

Class Function ZStrings.Split(Const ZStrs: String; Const Delimiter: String): String;
Var
  EndPtr, S: PChar;
Begin
  Result := '';
  S := PChar(ZStrs);
  EndPtr := S + Length(ZStrs);
  While S < EndPtr Do
  Begin
    If Result = '' Then
      Result := String(S)
    Else
      Result := Result + Delimiter + String(S);
    S := StrEnd(S) + 1;
  End;
End;

Class Function ZStrings.Split(Const ZStrs: String): TStringDynArray;
Var
  S, E: PChar;
  I: Integer;
Begin
  S := PChar(ZStrs);
  SetLength(Result, ZStrings.Count(ZStrs));
  For I := 0 To Length(Result) - 1 Do
  Begin
    E := StrEnd(S);
    SetString(Result[I], S, E - S);
    S := E + 1;
  End;
End;

Class Function ZStrings.IndexFromCharPos(Const ZStrs: String; Pos: Integer): Integer;
Var
  EndPtr, S, E: PChar;
Begin
  Result := -1;
  S := PChar(ZStrs);
  EndPtr := S + Length(ZStrs);
  While (S < EndPtr) And (Pos > 0) Do
  Begin
    E := StrEnd(S);
    Dec(Pos, E - S + 1);
    S := E + 1;
    Inc(Result);
  End;
End;

Class Function ZStrings.CharPosFromIndex(Const ZStrs: String; Index: Integer): Integer;
Var
  EndPtr, S: PChar;
Begin
  S := PChar(ZStrs);
  EndPtr := S + Length(ZStrs);
  While (Index > 0) And (S < EndPtr) Do
  Begin
    Dec(Index);
    S := StrEnd(S) + 1;
  End;
  Result := S - PChar(ZStrs) + 1;
End;

End.
