﻿Unit NCalGregorian;

{$I DELPHIAREA.INC}

Interface

Uses
  SysUtils, NCalendar;

Type
  TGregorianCalendar = Class(TCalendar)
  Protected Const
    GREGORIAN_EPOCH = 1721425.5;
    Function ToJulianDay(Year, Month, Day: Integer): Extended; Override;
    Function FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean; Override;
  Public Const
    CommonEra = 1;
    Class Function CalendarID: Cardinal; Override;
    Class Function CalendarKind: TCalendarKind; Override;
    Class Function MinSupportedDateTime: TDateTime; Override;
    Class Function MaxSupportedDateTime: TDateTime; Override;
    Class Function SettingsClass: TCalendarSettingsClass; Override;
    Function IsLeapYear(Era, Year: Integer): Boolean; Override;
    Function DaysInYear(Era, Year: Integer): Integer; Override;
    Function DaysInMonth(Era, Year, Month: Integer): Integer; Override;
    Function DaysToMonth(Era, Year, Month: Integer): Integer; Override;
  End;

  TGregorianCalendarSettings = Class(TCalendarSettings)
  Protected
    Procedure PrepareSettings(Const Locale: String; CalendarID: Cardinal); Override;
  Public
  Public
    Class Function CalendarClass: TCalendarClass; Override;
  End;

Implementation

Uses
  Math, Windows, Types, NWinNLS;

{ TGregorianCalendar }

Class Function TGregorianCalendar.CalendarID: Cardinal;
Begin
  Result := CAL_GREGORIAN;
End;

Class Function TGregorianCalendar.CalendarKind: TCalendarKind;
Begin
  Result := CkSolar;
End;

Class Function TGregorianCalendar.MinSupportedDateTime: TDateTime;
Begin
  Result := Max( { 0001/01/01 } -693593, Inherited);
End;

Class Function TGregorianCalendar.MaxSupportedDateTime: TDateTime;
Begin
  Result := Min( { 9999/12/31 } 2958465.99999, Inherited);
End;

Class Function TGregorianCalendar.SettingsClass: TCalendarSettingsClass;
Begin
  Result := TGregorianCalendarSettings;
End;

Function TGregorianCalendar.IsLeapYear(Era, Year: Integer): Boolean;
Begin
  Year := ConvertYear(Year, Era, CommonEra);
  Year := ToZeroBase(CommonEra, Year);
  Result := ((Year Mod 4) = 0) And (((Year Mod 100) <> 0) Or ((Year Mod 400) = 0));
End;

Function TGregorianCalendar.DaysInYear(Era, Year: Integer): Integer;
Begin
  If IsLeapYear(Era, Year) Then
    Result := 366
  Else
    Result := 365;
End;

Function TGregorianCalendar.DaysInMonth(Era, Year, Month: Integer): Integer;
Const
  Days: Array [Boolean] Of TDayTable = ((31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
    (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31));
Begin
  If (Month < 1) Or (Month > 12) Then
    MonthError(Era, Year, Month);
  Result := Days[IsLeapYear(Era, Year)][Month];
End;

Function TGregorianCalendar.DaysToMonth(Era, Year, Month: Integer): Integer;
Const
  Days: Array [Boolean] Of TDayTable = ((0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334),
    (0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335));
Begin
  If (Month < 1) Or (Month > 12) Then
    MonthError(Era, Year, Month);
  Result := Days[IsLeapYear(Era, Year)][Month];
End;

Function TGregorianCalendar.ToJulianDay(Year, Month, Day: Integer): Extended;
Const
  AdjustDays: Array [Boolean] Of Integer = (2, 1);
Var
  Adjust: Integer;
Begin
  If Month > 2 Then
    Adjust := AdjustDays[IsLeapYear(CommonEra, Year)]
  Else
    Adjust := 0;
  Year := ToZeroBase(CommonEra, Year);
  Result := GREGORIAN_EPOCH + (Year - 1) * 365 + Floor((Year - 1) / 4) - Floor((Year - 1) / 100) +
    Floor((Year - 1) / 400) + Floor(((367 * Month) - 362) / 12) + (Day - 1) - Adjust;
End;

Function TGregorianCalendar.FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean;
Var
  Period, QuadriCent, Cent, Quad, YearOfQuad, DayOfYear: Integer;
Begin
  JD := Trunc(JD - 0.5) + 0.5;
  Period := Trunc(JD - GREGORIAN_EPOCH);
  Divide(Period, 146097, QuadriCent, Period);
  Divide(Period, 36524, Cent, Period);
  Divide(Period, 1461, Quad, Period);
  YearOfQuad := Floor(Period / 365);
  Year := (QuadriCent * 400) + (Cent * 100) + (Quad * 4) + YearOfQuad;
  If (Cent <> 4) And (YearOfQuad <> 4) Then
    Inc(Year);
  Year := FromZeroBase(CommonEra, Year);
  DayOfYear := Trunc(JD - ToJulianDay(Year, 1, 1)) + 1;
  Result := DayOfYearToDayOfMonth(CommonEra, Year, DayOfYear, Month, Day);
End;

{ TGregorianCalendarSettings }

Class Function TGregorianCalendarSettings.CalendarClass: TCalendarClass;
Begin
  Result := TGregorianCalendar;
End;

Procedure TGregorianCalendarSettings.PrepareSettings(Const Locale: String; CalendarID: Cardinal);
Begin
  Inherited PrepareSettings(Locale, CalendarID);
  ShortDateFormat := 'yyyy/MM/dd';
  LongDateFormat := 'dddd d MMMM yyyy';
  ShortTimeFormat := 'hh:mm:ss ampm';
  LongTimeFormat := 'hh:nn:ss';
End;

Initialization

CalendarTypes.Register(TGregorianCalendar);

End.
