﻿Unit NCalHijri;

{$I DELPHIAREA.INC}

Interface

Uses
  SysUtils, Classes, NCalendar;

Type

  THijriCalendarAdjustmentEvent = Procedure(Sender: TObject; Const JulianDay: Extended;
    Var NumberOfDays: Integer) Of Object;

  THijriCalendar = Class(TCalendar)
  Private
    FAdjustDays: Integer;
    FOnAdjustment: THijriCalendarAdjustmentEvent;
    Procedure SetHijriAdjust(Value: Integer);
  Protected Const
    HIJRI_EPOCH = 1948439.5;
    Function ToJulianDay(Year, Month, Day: Integer): Extended; Override;
    Function FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean; Override;
    Function AdjustJulianDay(JD: Extended; Backward: Boolean): Extended; Override;
  Public Const
    HijriEra = 1;
    Class Function CalendarID: Cardinal; Override;
    Class Function CalendarKind: TCalendarKind; Override;
    Class Function MinSupportedDateTime: TDateTime; Override;
    Class Function MaxSupportedDateTime: TDateTime; Override;
    Class Function SettingsClass: TCalendarSettingsClass; Override;
    Function IsLeapYear(Era, Year: Integer): Boolean; Override;
    Function DaysInYear(Era, Year: Integer): Integer; Override;
    Function DaysInMonth(Era, Year, Month: Integer): Integer; Override;
    Procedure Assign(Source: TPersistent); Override;
    Property AdjustDays: Integer Read FAdjustDays Write SetHijriAdjust Default 0;
    Property OnAdjustment: THijriCalendarAdjustmentEvent Read FOnAdjustment Write FOnAdjustment;
  End;

  THijriCalendarSettings = Class(TCalendarSettings)
  Protected
    Procedure PrepareMonthNames(Const Locale: String; CalendarID: Cardinal); Override;
    Procedure PrepareSettings(Const Locale: String; CalendarID: Cardinal); Override;
  Public
    Class Function CalendarClass: TCalendarClass; Override;
  End;

Implementation

Uses
  Math, Types, NWinNLS;

{ THijriCalendar }

Class Function THijriCalendar.CalendarID: Cardinal;
Begin
  Result := CAL_HIJRI;
End;

Class Function THijriCalendar.CalendarKind: TCalendarKind;
Begin
  Result := CkLunar;
End;

Class Function THijriCalendar.MinSupportedDateTime: TDateTime;
Begin
  Result := Max( { 0001/01/01 Hijri } -466579.0, Inherited);
End;

Class Function THijriCalendar.MaxSupportedDateTime: TDateTime;
Begin
  Result := Min( { 9999/12/29 Hijri } 3076732.99999, Inherited);
End;

Class Function THijriCalendar.SettingsClass: TCalendarSettingsClass;
Begin
  Result := THijriCalendarSettings;
End;

Function THijriCalendar.IsLeapYear(Era, Year: Integer): Boolean;
Begin
  Year := ConvertYear(Year, Era, HijriEra);
  Year := ToZeroBase(HijriEra, Year);
  Result := Modulus(((Year * 11) + 14), 30) < 11;
End;

Function THijriCalendar.DaysInYear(Era: Integer; Year: Integer): Integer;
Begin
  If IsLeapYear(Era, Year) Then
    Result := 355
  Else
    Result := 354;
End;

Function THijriCalendar.DaysInMonth(Era: Integer; Year: Integer; Month: Integer): Integer;
Begin
  If (Month < 1) Or (Month > 12) Then
    MonthError(Era, Year, Month);
  If Odd(Month) Or ((Month = 12) And IsLeapYear(Era, Year)) Then
    Result := 30
  Else
    Result := 29;
End;

Function THijriCalendar.ToJulianDay(Year, Month, Day: Integer): Extended;
Begin
  Year := ToZeroBase(HijriEra, Year);
  Result := HIJRI_EPOCH + +(Year - 1) * 354 + Floor((3 + (11 * Year)) / 30) + (29 * (Month - 1)) + (Month Div 2)
    + (Day - 1);
End;

Function THijriCalendar.FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean;
Var
  YearDay: Integer;
Begin
  JD := Trunc(JD - 0.5) + 0.5;
  Year := Floor((30 * (JD - HIJRI_EPOCH) + 10646) / 10631);
  Year := FromZeroBase(HijriEra, Year);
  YearDay := Trunc(JD - ToJulianDay(Year, 1, 1)) + 1;
  Result := DayOfYearToDayOfMonth(HijriEra, Year, YearDay, Month, Day);
End;

Function THijriCalendar.AdjustJulianDay(JD: Extended; Backward: Boolean): Extended;
Var
  NumberOfDays: Integer;
Begin
  NumberOfDays := AdjustDays;
  If Assigned(OnAdjustment) Then
    OnAdjustment(Self, JD, NumberOfDays);
  If Backward Then
    Result := JD - NumberOfDays
  Else
    Result := JD + NumberOfDays;
End;

Procedure THijriCalendar.SetHijriAdjust(Value: Integer);
Begin
  If AdjustDays <> Value Then
  Begin
    FAdjustDays := Value;
    DoChange;
  End;
End;

Procedure THijriCalendar.Assign(Source: TPersistent);
Begin
  If Source Is THijriCalendar Then
    AdjustDays := THijriCalendar(Source).AdjustDays;
  Inherited Assign(Source);
End;

{ THijriCalendarSettings }

Class Function THijriCalendarSettings.CalendarClass: TCalendarClass;
Begin
  Result := THijriCalendar;
End;

Procedure THijriCalendarSettings.PrepareMonthNames(Const Locale: String; CalendarID: Cardinal);
Const
  EnglishMonthNames: Array [1 .. 12] Of String = ('Muharram', 'Safar', 'Rabi`al-Awwal', 'Rabi`ath-Thani',
    'Jumada l-Ula', 'Jumada t-Tania', 'Rajab', 'Sha`ban', 'Ramadan', 'Shawwal', 'Dhu l-Qa`da', 'Dhu l-Hijja');
Var
  M: Integer;
  LocaleScripts: String;
Begin
  LocaleScripts := GetLocaleScripts(Locale);
  If (Pos('Arab;', LocaleScripts) <> 0) Or (Pos('Thaa;', LocaleScripts) <> 0) Then
    Inherited PrepareMonthNames(Locale, CalendarID)
  Else
  Begin
    For M := 1 To 12 Do
    Begin
      MonthNames[M] := EnglishMonthNames[M];
      ShortMonthNames[M] := EnglishMonthNames[M];
      GenitiveMonthNames[M] := EnglishMonthNames[M];
    End;
  End;
End;

Procedure THijriCalendarSettings.PrepareSettings(Const Locale: String; CalendarID: Cardinal);
Begin
  Inherited PrepareSettings(Locale, CalendarID);
  ShortDateFormat := 'yyyy/MM/dd';
  LongDateFormat := 'dddd d MMMM yyyy';
  ShortTimeFormat := 'hh:mm:ss ampm';
  LongTimeFormat := 'hh:nn:ss';
End;

Initialization

CalendarTypes.Register(THijriCalendar, ['ar', 'ur-PK', 'dv-MV']);

End.
