﻿Unit NCalJalali;
{$I DELPHIAREA.INC}

Interface

Uses
  SysUtils, NCalendar;

Type

  TJalaliCalendar = Class(TCalendar)
  Protected Const
    JALALI_EPOCH = 1948320.5;
    Function ToJulianDay(Year, Month, Day: Integer): Extended; Override;
    Function FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean; Override;
  Public Const
    HijriEra = 1;
    Constructor Create; Overload;
    Class Function CalendarID: Cardinal; Override;
    Class Function CalendarKind: TCalendarKind; Override;
    Class Function MinSupportedDateTime: TDateTime; Override;
    Class Function MaxSupportedDateTime: TDateTime; Override;
    Class Function SettingsClass: TCalendarSettingsClass; Override;
    Function IsLeapYear(Era, Year: Integer): Boolean; Override;
    Function DaysInYear(Era, Year: Integer): Integer; Override;
    Function DaysInMonth(Era, Year, Month: Integer): Integer; Override;
    Function DaysToMonth(Era, Year, Month: Integer): Integer; Override;
    Function DayOfYearToDayOfMonth(Era, Year, YearDay: Integer; Var Month, Day: Integer): Boolean; Override;
  End;

  TJalaliCalendarSettings = Class(TCalendarSettings)
  Protected
    Procedure PrepareEraName(Era: Integer; Const Locale: String; CalendarID: Cardinal); Override;
    Procedure PrepareMonthNames(Const Locale: String; CalendarID: Cardinal); Override;
    Procedure PrepareDayNames(Const Locale: String; CalendarID: Cardinal); Override;
    Procedure PrepareSettings(Const Locale: String; CalendarID: Cardinal); Override;
  Public
    Class Function CalendarClass: TCalendarClass; Override;
  End;

Implementation

Uses
  Math, Types, NWinNLS;

{ TJalaliCalendar }

Class Function TJalaliCalendar.CalendarID: Cardinal;
Begin
  Result := CAL_PERSIAN;
End;

Class Function TJalaliCalendar.CalendarKind: TCalendarKind;
Begin
  Result := CkSolar;
End;

Constructor TJalaliCalendar.Create;
Begin
  Inherited Create;
  Settings.DateSeparator := '/';
End;

Class Function TJalaliCalendar.MinSupportedDateTime: TDateTime;
Begin
  Result := Max( { 0955/01/01 Jalali } -118257, Inherited);
End;

Class Function TJalaliCalendar.MaxSupportedDateTime: TDateTime;
Begin
  Result := Min( { 1403/12/29 Jalali } 45735.99999, Inherited);
End;

Class Function TJalaliCalendar.SettingsClass: TCalendarSettingsClass;
Begin
  Result := TJalaliCalendarSettings;
End;

Function TJalaliCalendar.IsLeapYear(Era, Year: Integer): Boolean;
Begin
  Year := ConvertYear(Year, Era, HijriEra);
  Year := ToZeroBase(HijriEra, Year);
  Result := ((((Modulus(Year - 474, 2820) + 474) + 38) * 682) Mod 2816) < 682;
End;

Function TJalaliCalendar.DaysInYear(Era, Year: Integer): Integer;
Begin
  If IsLeapYear(Era, Year) Then
    Result := 366
  Else
    Result := 365;
End;

Function TJalaliCalendar.DaysInMonth(Era, Year, Month: Integer): Integer;
Begin
  If (Month < 1) Or (Month > 12) Then
    MonthError(Era, Year, Month);
  If (Month = 12) And Not IsLeapYear(Era, Year) Then
    Result := 29
  Else If Month > 6 Then
    Result := 30
  Else
    Result := 31;
End;

Function TJalaliCalendar.DaysToMonth(Era, Year, Month: Integer): Integer;
Begin
  If (Month < 1) Or (Month > 12) Then
    MonthError(Era, Year, Month);
  Result := Min(Month - 1, 6) + ((Month - 1) * 30);
End;

Function TJalaliCalendar.DayOfYearToDayOfMonth(Era, Year, YearDay: Integer; Var Month, Day: Integer): Boolean;
Begin
  If YearDay <= 6 * 31 Then
  Begin
    Month := ((YearDay - 1) Div 31) + 1;
    Day := YearDay - (Month - 1) * 31;
    Result := True;
  End
  Else If YearDay <= DaysInYear(Era, Year) Then
  Begin
    Dec(YearDay, 6 * 31);
    Month := ((YearDay - 1) Div 30) + 7;
    Day := YearDay - (Month - 7) * 30;
    Result := True;
  End
  Else
    Result := False;
End;

Function TJalaliCalendar.ToJulianDay(Year, Month, Day: Integer): Extended;
Var
  EpBase, EpYear: Integer;
Begin
  Year := ToZeroBase(HijriEra, Year);
  EpBase := Year - 474;
  EpYear := 474 + Modulus(EpBase, 2820);
  Result := Jalali_EPOCH + (EpYear - 1) * 365 + Floor(((EpYear * 682) - 110) / 2816) + Floor(EpBase / 2820) *
    1029983 + DaysToMonth(HijriEra, Year, Month) + (Day - 1);
End;

Function TJalaliCalendar.FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean;
Var
  Period, Cycle, CYear, YCycle, Aux1, Aux2, YearDay: Integer;
Begin
  JD := Trunc(JD - 0.5) + 0.5;
  Period := Trunc(JD - ToJulianDay(475, 1, 1));
  Divide(Period, 1029983, Cycle, CYear);
  If CYear = 1029982 Then
    YCycle := 2820
  Else
  Begin
    Divide(CYear, 366, Aux1, Aux2);
    YCycle := Floor(((2134 * Aux1) + (2816 * Aux2) + 2815) / 1028522) + Aux1 + 1;
  End;
  Year := YCycle + (2820 * Cycle) + 474;
  Year := FromZeroBase(HijriEra, Year);
  YearDay := Trunc(JD - ToJulianDay(Year, 1, 1)) + 1;
  Result := DayOfYearToDayOfMonth(HijriEra, Year, YearDay, Month, Day);
End;

{ TJalaliCalendarSettings }

Class Function TJalaliCalendarSettings.CalendarClass: TCalendarClass;
Begin
  Result := TJalaliCalendar;
End;

Procedure TJalaliCalendarSettings.PrepareDayNames(Const Locale: String; CalendarID: Cardinal);
Const
  PersianDayNames: Array [1 .. 7] Of String = ('دوشنبه', 'سه شنبه', 'چهارشنبه', 'پنج شنبه', 'جمعه', 'شنبه',
    'يکشنبه');
  ArabicDayNames: Array [1 .. 7] Of String = ('الإثنين', 'الثلاثاء', 'الأربعاء', 'الخميس', 'الجمعة',
    'السبت', 'الأحد');
  EnglishDayNames: Array [1 .. 7] Of String = ('Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
    'Saturday', 'Sunday');
Var
  D: Integer;
Begin
  Case CalendarLang Of
    ClPersian:
      Begin
        For D := 1 To 7 Do
        Begin
          DayNames[TDayOfWeek(D)] := PersianDayNames[D];
          ShortDayNames[TDayOfWeek(D)] := Copy(PersianDayNames[D], 1, 1);
          ShortestDayNames[TDayOfWeek(D)] := Copy(PersianDayNames[D], 1, 1);
        End;
      End;
    ClArabic:
      Begin
        For D := 1 To 7 Do
        Begin
          DayNames[TDayOfWeek(D)] := ArabicDayNames[D];
          ShortDayNames[TDayOfWeek(D)] := ArabicDayNames[D];
          ShortestDayNames[TDayOfWeek(D)] := ArabicDayNames[D];
        End;
      End;
    ClEnglish:
      Begin
        For D := 1 To 7 Do
        Begin
          DayNames[TDayOfWeek(D)] := EnglishDayNames[D];
          ShortDayNames[TDayOfWeek(D)] := Copy(EnglishDayNames[D], 1, 3);
          ShortestDayNames[TDayOfWeek(D)] := Copy(EnglishDayNames[D], 1, 3);
        End;
      End;
  End;
End;

Procedure TJalaliCalendarSettings.PrepareEraName(Era: Integer; Const Locale: String; CalendarID: Cardinal);
Begin
  If (Era = TJalaliCalendar.HijriEra) And (Pos('Arab;', GetLocaleScripts(Locale)) <> 0) Then
  Begin
    EraNames[TJalaliCalendar.HijriEra] := 'هجری خورشیدی';
    ShortEraNames[TJalaliCalendar.HijriEra] := 'ﻫ';
  End
  Else
    Inherited PrepareEraName(Era, Locale, CalendarID);
End;

Procedure TJalaliCalendarSettings.PrepareMonthNames(Const Locale: String; CalendarID: Cardinal);
Const
  PersianMonthNames: Array [1 .. 12] Of String = ('فروردین', 'اردیبهشت', 'خرداد', 'تیر', 'مرداد', 'شهریور',
    'مهر', 'آبان', 'آذر', 'دی', 'بهمن', 'اسفند');
  ArabicMonthNames: Array [1 .. 12] Of String = ('حمل', 'ثور', 'جوزا', 'سرطان', 'اسد', 'سنبله', 'میزان', 'عقرب',
    'قوس', 'جدی', 'دلو', 'حوت');
  EnglishMonthNames: Array [1 .. 12] Of String = ('Farvardin', 'Ordibehesht', 'Khordad', 'Tir', 'Mordad',
    'Shahrivar', 'Mehr', 'Aban', 'Azar', 'Dey', 'Bahman', 'Esfand');
Var
  M: Integer;
Begin
  Case CalendarLang Of
    ClPersian:
      Begin
        For M := 1 To 12 Do
        Begin
          MonthNames[M] := PersianMonthNames[M];
          ShortMonthNames[M] := PersianMonthNames[M];
          GenitiveMonthNames[M] := PersianMonthNames[M];
        End;
      End;
    ClArabic:
      Begin
        For M := 1 To 12 Do
        Begin
          MonthNames[M] := ArabicMonthNames[M];
          ShortMonthNames[M] := ArabicMonthNames[M];
          GenitiveMonthNames[M] := ArabicMonthNames[M];
        End;
      End;
    ClEnglish:
      Begin
        For M := 1 To 12 Do
        Begin
          MonthNames[M] := EnglishMonthNames[M];
          ShortMonthNames[M] := Copy(EnglishMonthNames[M], 1, 3);
          GenitiveMonthNames[M] := EnglishMonthNames[M];
        End;
      End;
  End;
End;

Procedure TJalaliCalendarSettings.PrepareSettings(Const Locale: String; CalendarID: Cardinal);
Begin
  Inherited PrepareSettings(Locale, CalendarID);
  TwoDigitYearMax := 1350;
  CalendarName := 'گاهشمار جلالی';
  FirstDayOfWeek := Saturday;
  ShortDateFormat := 'yyyy/MM/dd';
  LongDateFormat := 'dddd d MMMM yyyy';
  ShortTimeFormat := 'hh:mm:ss ampm';
  LongTimeFormat := 'hh:nn:ss';
  DateSeparator := '/';
  YearMonthFormat := 'MMMM , yyyy';
End;

Initialization

CalendarTypes.Register(TJalaliCalendar, ['fa-IR', 'prs-AF', 'ps-AF']);

End.
